// SSM_Adaptation.h: interface for the SSM_Adaptation class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SSM_ADAPTATION_H__C635E2D7_92CA_4CC6_9792_08EE247C6131__INCLUDED_)
#define AFX_SSM_ADAPTATION_H__C635E2D7_92CA_4CC6_9792_08EE247C6131__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "sctp.h"
#include "globals.h"
#include "ssm_distribution.h"
//#include "distribution.h"


#define WspiapiMalloc(tSize)    calloc(1, (tSize))
#define WspiapiFree(p)          free(p)
#define WspiapiSwap(a, b, c)    { (c) = (a); (a) = (b); (b) = (c); }
#define getaddrinfo             WspiapiGetAddrInfo
#define getnameinfo             WspiapiGetNameInfo
#define freeaddrinfo            WspiapiFreeAddrInfo
#define socklen_t          unsigned int
#define size_t             unsigned int
#define NI_MAXHOST  1025  /* Max size of a fully-qualified domain name */
#define NI_MAXSERV    32  /* Max size of a service name */

#define NI_NOFQDN       0x01  /* Only return nodename portion for local hosts */
#define NI_NUMERICHOST  0x02  /* Return numeric form of the host's address */
#define NI_NAMEREQD     0x04  /* Error if the host's name not in DNS */
#define NI_NUMERICSERV  0x08  /* Return numeric form of the service (port #) */
#define NI_DGRAM        0x10  /* Service is a datagram service */

#define EAI_AGAIN       WSATRY_AGAIN
#define EAI_BADFLAGS    WSAEINVAL
#define EAI_FAIL        WSANO_RECOVERY
#define EAI_FAMILY      WSAEAFNOSUPPORT
#define EAI_MEMORY      WSA_NOT_ENOUGH_MEMORY

#define EAI_NONAME      WSAHOST_NOT_FOUND
#define EAI_SERVICE     WSATYPE_NOT_FOUND
#define EAI_SOCKTYPE    WSAESOCKTNOSUPPORT
#define EAI_NODATA      EAI_NONAME

#define AI_PASSIVE     0x1  /* Socket address will be used in bind() call */
#define AI_CANONNAME   0x2  /* Return canonical name in first ai_canonname */
#define AI_NUMERICHOST 0x4  /* Nodename must be a numeric address string */

class SSM_Adaptation  
{
public:
	static int init_poll_fds(void);
	static int	adl_register_fd_cb(int sfd, int eventcb_type, int event_mask,
					   void (*action) (void *, void *) , void* userData);
	static int adl_receive_message(int sfd, void *dest, int maxlen, union sockunion *from, union sockunion *to);
	static int adl_get_message(int sfd, void *dest, int maxlen, union sockunion *from, int *from_len);
	static void dispatch_event(int num_of_events);
	static void dispatch_timer(void);

	static int extendedPoll(struct extendedpollfd* fdlist,
                 int*                   count,
                 int                    time,
                 void                   (*lock)(void* data),
                 void                   (*unlock)(void* data),
                 void*                  data);

	static void assign_poll_fd(int fd_index, int sfd, int event_mask);

	static unsigned int adl_random(void);

	static boolean adl_equal_address(union sockunion *one, union sockunion *two);


	/**
	 *  converts address-string (hex for ipv6, dotted decimal for ipv4
	 *  to a sockunion structure
	 *  @return 0 for success, else -1.
	 */
	static int adl_str2sockunion(guchar * str, union sockunion *su);


	static int adl_sockunion2str(union sockunion *su, guchar * buf, size_t len);


	/*------------------------------------------------------------------------------------------------------------------*/
	/*------------------------------------------------------------------------------------------------------------------*/

	/**
	 * This function binds a local socket for incoming requests
	 * @return socket file descriptor for the newly opened and bound socket
	 * @param address (local) port to bind to
	 */
	static gint adl_open_sctp_socket(int af, int* myRwnd);

	static int adl_setReceiveBufferSize(int sfd, int new_size);

	static gint adl_get_sctpv4_socket(void);



	/**
	 * function to be called when we get a message from a peer sctp instance in the poll loop
	 * @param  sfd the socket file descriptor where data can be read...
	 * @param  buf pointer to a buffer, where we data is stored
	 * @param  len number of bytes to be sent, including the ip header !
	 * @param  address, where data goes from
	 * @param	dest_len size of the address
	 * @return returns number of bytes actually sent, or error
	 */
	static int adl_send_message(int sfd, void *buf, int len, union sockunion *dest, unsigned char tos);


	/**
	 * this function initializes the data of this module. It opens raw sockets for
	 * capturing SCTP packets, and also opens ICMP sockets, so we can get ICMP events,
	 * e.g.  for Path-MTU discovery !
	 */
	static int adl_init_adaptation_layer(int * myRwnd);



	/**
	 * function add a sfd to the list of sfds we want to wait for with the poll()
	 * @param sfd	        the socket file descriptor of the socket to react upon
	 * @param scf           function pointer holding the callback funtion for normal events.
	 * @return              the current number of sockets that are polled.
	 */
	static int adl_register_socket_cb(gint sfd, sctp_socketCallback scf);


	/**
	 *	function to close a bound socket from our list of socket descriptors
	 *	@param	sfd	socket file descriptor to be closed
	 *	@return  0 on success, -1 for error, 1 if socket was not bound
	 *    @author  ajung
	 */
	static int adl_remove_cb(gint sfd);

	/**
	 * remove a sfd from the poll_list, and shift that list to the left
	 * @return number of sfd's removed...
	 */
	static int adl_remove_poll_fd(gint sfd);


	/**
	 * function is to return difference in msecs between time a and b (i.e. a-b)
	 * @param a later time (e.g. current time)
	 * @param b earlier time
	 * @return -1 if a is earlier than b, else msecs that passed from b to a
	 */
	static int adl_timediff_to_msecs(struct timeval *a, struct timeval *b);


	static void adl_add_msecs_totime(struct timeval *t, unsigned int msecs);

	static int adl_gettime(struct timeval *tv);

	static int adl_extendedGetEvents(void (*lock)(void* data), void (*unlock)(void* data), void* data);




	static int adl_sendUdpData(int sfd, unsigned char* buf, int length,
						 unsigned char destination[], unsigned short dest_port);


	static int adl_registerStdinCallback(sctp_StdinCallback sdf, char* buffer, int length);
	static int adl_unregisterStdinCallback();

	static int adl_registerUserCallback(int fd, sctp_userCallback sdf, void* userData, short int eventMask);

	static int adl_unregisterUserCallback(int fd);

	static unsigned int adl_startMicroTimer(unsigned int seconds, unsigned int microseconds,
								sctp_timerCallback timer_cb, int ttype, void *param1, void *param2);

	static unsigned int adl_startTimer(unsigned int milliseconds, sctp_timerCallback timer_cb,
									 int ttype, void *param1, void *param2);

	static int adl_stopTimer(unsigned int tid);

	static unsigned int adl_restartTimer(unsigned int timer_id, unsigned int milliseconds);

	static unsigned int adl_restartMicroTimer(unsigned int timer_id, unsigned int seconds, unsigned int microseconds);

	static int adl_getEvents(void);

	static int adl_eventLoop();

	static UINT threadFunc ( LPVOID p ) ;

	static int adl_extendedEventLoop(void (*lock)(void* data), void (*unlock)(void* data), void* data);

	static gboolean adl_filterInetAddress(union sockunion* newAddress, AddressScopingFlags  flags);

	/*
	 * this is an ugly part to code, so it was taken an adapted from the
	 * SCTP reference implementation by Randy Stewart
	 * see http://www.sctp.org
	 * maybe I should rewrite it to use the Linux Netlink socket also
	 * returns TRUE is successful, else FALSE
	 */
	static gboolean adl_gatherLocalAddresses(union sockunion **localAddresses,
		 int *numberOfNets,
		 int sctp_fd,
		 gboolean with_ipv6,
		 int *max_mtu,
		 const AddressScopingFlags  flags);
	
	

	static	char *
	WspiapiStrdup (
	  const char *                    pszString);

	static BOOL
	WspiapiParseV4Address (
		  const char *                    pszAddress,
		 PDWORD                          pdwAddress);


	static struct addrinfo *
	WspiapiNewAddrInfo (
		  int                             iSocketType,
		  int                             iProtocol,
		  WORD                            wPort,
		  DWORD                           dwAddress);


	static int
	WspiapiQueryDNS(
		  const char                      *pszNodeName,
		  int                             iSocketType,
		  int                             iProtocol,
		  WORD                            wPort,
		 char                            pszAlias[NI_MAXHOST],
		struct addrinfo                 **pptResult);


	static int
	WspiapiLookupNode(
		  const char                      *pszNodeName,
		  int                             iSocketType,
		  int                             iProtocol,
		  WORD                            wPort,
		  BOOL                            bAI_CANONNAME,
		 struct addrinfo                 **pptResult);

	static int
	WspiapiClone (
		  WORD                            wPort,
		  struct addrinfo                 *ptResult);


	static void
	WspiapiLegacyFreeAddrInfo (
		IN  struct addrinfo                 *ptHead);


	static int
	WspiapiLegacyGetAddrInfo(
		 const char                       *pszNodeName,
		 const char                       *pszServiceName,
		 const struct addrinfo            *ptHints,
		 struct addrinfo                 **pptResult);


	static int
	WspiapiLegacyGetNameInfo(
		  const struct sockaddr           *ptSocketAddress,
		  socklen_t                       tSocketLength,
		 char                            *pszNodeName,
		  size_t                          tNodeLength,
		 char                            *pszServiceName,
		  size_t                          tServiceLength,
		  int                             iFlags);


	static FARPROC
	WspiapiLoad(
		  WORD                            wFunction);


	static int
	WspiapiGetAddrInfo(
		 const char                       *nodename,
		 const char                       *servname,
		 const struct addrinfo            *hints,
		 struct addrinfo                 **res);


	static int
	WspiapiGetNameInfo (
		  const struct sockaddr           *sa,
		  socklen_t                       salen,
		 char                            *host,
		  size_t                          hostlen,
		 char                            *serv,
		  size_t                          servlen,
		  int                             flags);


	static void
	WspiapiFreeAddrInfo (
		 struct addrinfo                 *ai);














	
	SSM_Adaptation();

	
	virtual ~SSM_Adaptation();

};

#endif // !defined(AFX_SSM_ADAPTATION_H__C635E2D7_92CA_4CC6_9792_08EE247C6131__INCLUDED_)
