// SSM_Bundling.h: interface for the SSM_Bundling class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SSM_BUNDLING_H__2381ABA3_23A8_49FC_8CB6_CFDDBD734E12__INCLUDED_)
#define AFX_SSM_BUNDLING_H__2381ABA3_23A8_49FC_8CB6_CFDDBD734E12__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000



#include "globals.h"            /* for chunk struct definition */
#include "SSM_SCTP_control.h"
#include "SSM_pathmanagement.h"
//#include "pathmanagement.h"

class SSM_Bundling  
{
public:
	SSM_Bundling();
	virtual ~SSM_Bundling();


	static void bu_init_bundling(void);
	static gint bu_put_Ctrl_Chunk(SCTP_simple_chunk * chunk,unsigned int * dest_index);
	static gint bu_put_Data_Chunk(SCTP_simple_chunk * chunk,unsigned int * dest_index);


	/*
	 * bu_new: Creates a new bundling instance and returns a pointer to its data. 
	 */
	static gpointer bu_new(void);

	/* 
	 * bu_delete: Deletes a bundling instance
	 * 
	 * Params: Pointer/handle which was returned by bu_new()
	 */
	static void bu_delete(gpointer instancePtr);

	/**
	 * returns a value indicating which chunks are in the packet.
	 */
	static unsigned int rbu_scanPDU(guchar * pdu, guint len);

	/*
	 * rbu_datagramContains: looks for chunk_type in a newly received datagram
	 * Should be called after rbu_scanPDU().
	 * The chunkArray parameter is inspected. This only really checks for chunks
	 * with an ID <= 30. For all other chunks, it just guesses...
	 * @return true is chunk_type exists in chunkArray, false if it is not in there
	 */
	static gboolean rbu_datagramContains(gushort chunk_type, unsigned int chunkArray);

	static guchar* rbu_scanInitChunkForParameter(guchar * chunk, gushort paramType);

	/*
	 * rbu_findChunk: looks for chunk_type in a newly received datagram
	 *
	 * All chunks within the datagram are looked at, until one is found
	 * that equals the parameter chunk_type.
	 * @return pointer to first chunk of chunk_type in SCTP datagram, else NULL
	 */
	static guchar* rbu_findChunk(guchar * datagram, guint len, gushort chunk_type);

	/*
	 * rbu_scanDatagramForError : looks for Error chunk_type in a newly received datagram
	 * that contains a special error cause code
	 *
	 * All chunks within the datagram are lookes at, until one is found
	 * that equals the parameter chunk_type.
	 * @return true is chunk_type exists in SCTP datagram, false if it is not in there
	 */
	static gboolean rbu_scanDatagramForError(guchar * datagram, guint len, gushort error_cause);

	/*
	 * rbu_findAddress: looks for address type parameters in INIT or INIT-ACKs
	 * All parameters within the chunk are looked at, and the n-th supported address is
	 * copied into the provided buffer pointed to by the foundAddress parameter.
	 * If there are less than n addresses, an appropriate error is
	 * returned. n should be at least 1, of course.
	 * @param  chunk            pointer to an INIT or INIT ACK chunk
	 * @param  n                get the n-th address
	 * @param  foundAddress     pointer to a buffer where an address, if found, will be copied
	 * @return -1  for parameter problem, 0 for success (i.e. address found), 1 if there are not
	 *             that many addresses in the chunk.
	 */
	static gint rbu_findAddress(guchar * chunk, guint n, union sockunion* foundAddress, int supportedAddressTypes);

	/*
	 * rbu_rcvDatagram: Hands a lower layer datagram over to bundling (for de-bundling)
	 * 
	 * All chunks within the datagram are dispatched and sent to the appropriate
	 * module, i.e.: control chunks are sent to sctp_control/pathmanagement, 
	 * SACK chunks to reliable_transfer, and data_chunks to RX_control.
	 * Those modules must get a pointer to the start of a chunk and 
	 * information about its size (without padding).
	 */
	static gint rbu_rcvDatagram(guint address_index, guchar * datagram, guint len);


	static void bu_lock_sender(void);
	static void bu_unlock_sender(guint* ad_idx);
	static gboolean bu_userDataOutbound(void);

	/**
	 * bu_put_SACK_Chunk: Called by recvcontrol, when a SACK must be piggy-backed
	 *
	 * @param chunk pointer to chunk, that is to be put in the bundling buffer
	 * @return error value, 0 on success, -1 on error
	 */
	static gint bu_put_SACK_Chunk(SCTP_sack_chunk * chunk, unsigned int* dest_index);


	/*
	 * bu_sendAllChunks: Trigger to send all chunks previously entered with putChunk.
	 *
	 * Return value: error value
	 * Chunks sent are deleted afterwards.
	 */
	static gint bu_sendAllChunks(guint * ad_idx);

	static void bu_request_sack(void);


};

#endif // !defined(AFX_SSM_BUNDLING_H__2381ABA3_23A8_49FC_8CB6_CFDDBD734E12__INCLUDED_)
