// SSM_Errorhandler.h: interface for the SSM_Errorhandler class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SSM_ERRORHANDLER_H__C85003BD_72DD_4CC1_8F63_CB480475C6E9__INCLUDED_)
#define AFX_SSM_ERRORHANDLER_H__C85003BD_72DD_4CC1_8F63_CB480475C6E9__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "globals.h"            /* for chunk struct definition */

class SSM_Errorhandler  
{
public:
	SSM_Errorhandler();
	virtual ~SSM_Errorhandler();


	
	static void eh_init_errorhandler(void);

	/*
	 * eh_new: Create a new instance and returns a pointer to its data.
	 */
	static void *eh_new(void);

	/* 
	 * eh_delete: Deletes a bundling instance
	 * 
	 * Params: Pointer/handle which was returned by eh_new()
	 */
	static void eh_delete(void *instancePtr);

	/*
	 *  eh_recv_chunk gets a pointer to an error chunk and decodes it
	 *  accordingly....
	 *  @return  error code, 0 for success, less than one for error
	 */
	static int eh_recv_chunk(SCTP_simple_chunk * errchunk);

	/**
	 * function to trigger sending of error chunk, after receiving an invalid stream id
	 * @return error value, 0 on success, -1 on error
	 */
	static int eh_make_invalid_streamid_error(unsigned short streamid);


	/**
	 * function sends the unknown chunk back
	 */
	static int eh_send_unrecognized_chunktype(unsigned char* faulty_chunk, unsigned short length);

	/**
	 * function to trigger sending of error chunk, after mandatory parameter(s) was(were) missing
	 * @return error value, 0 on success, -1 on error
	 */
	static int eh_make_missing_mandatory_param(unsigned int number, unsigned short *param_types);

	/**
	 * function to trigger sending of error chunk, after receiving an invalid stream id
	 * @param number number of pointers passed as second argument
	 * @param addresses pointers (or array of pointers) to unrecognized addresses
	 * @return error value, 0 on success, -1 on error
	 */
	static int eh_send_unresolvable_address(unsigned int number, unsigned char *addresses);



	/**
	 * function to add an error chunk, after empty data chunk was received
	 * @return error value, 0 on success, -1 on error
	 */
	static int eh_make_empty_data_chunk_error(unsigned int tsn);


};

#endif // !defined(AFX_SSM_ERRORHANDLER_H__C85003BD_72DD_4CC1_8F63_CB480475C6E9__INCLUDED_)
