// SSM_SCTP_Control.cpp: implementation of the SSM_SCTP_Control class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SSM_SCTP_Control.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#include "ssm_distribution.h"
//#include "distribution.h"
#include "SSM_bundling.h"
//#include "bundling.h"
#include "ssm_adaptation.h"
//#include "adaptation.h"
#include "SSM_pathmanagement.h"
//#include "pathmanagement.h"
#include "ssm_reltransfer.h"
//#include "reltransfer.h"
#include "ssm_recvctrl.h"
//#include "recvctrl.h"
#include "SSM_ChunkHandler.h"
//#include "chunkHandler.h"
#include "ssm_flowcontrol.h"
//#include "flowcontrol.h"
#include "ssm_streamengine.h"
//#include "streamengine.h"

#ifdef HAVE_STRINGS_H
    #include <strings.h>
#endif
#include <stdio.h>
#include <errno.h>

#include "sctp.h"


/**
  SCTP-control structure. Stores also the current state of the state-machine.
 */
typedef struct SCTP_CONTROLDATA
{
    /*@{ */
    /** the state of this state machine */
    guint32 association_state;
    /** stores timer-ID of init/cookie-timer, used to stop this timer */
    TimerID initTimer;
    /** */
    unsigned int initTimerDuration;
    /**  stores the association id (==tag) of this association */
    unsigned int associationID;
    /** Counter for init and cookie retransmissions */
    short initRetransCounter;
    /** pointer to the init chunk data structure (for retransmissions) */
    SCTP_init *initChunk;
    /** pointer to the cookie chunk data structure (for retransmissions) */
    SCTP_cookie_echo *cookieChunk;
    /** my tie tag for cross initialization and other sick cases */
    guint32 local_tie_tag;
    /** peer's tie tag for cross initialization and other sick cases */
    guint32 peer_tie_tag;
    /** we store these here, too. Maybe better be stored with StreamEngine ? */
    unsigned short NumberOfOutStreams;
    /** we store these here, too. Maybe better be stored with StreamEngine ? */
    unsigned short NumberOfInStreams;
    /** value for maximum retransmissions per association */
    int assocMaxRetransmissions;
    /** value for maximum initial retransmissions per association */
    int assocMaxInitRetransmissions;
    /** value for the current cookie lifetime */
    int cookieLifeTime;
    /** the sctp instance */
    void * instance;
    /*@} */
} SCTP_controlData;

/* -------------------- Declarations -------------------------------------------------------------*/

/*
pointer to the current controller structure. Only set when association exists.
*/
static SCTP_controlData *localData;



SSM_SCTP_Control::SSM_SCTP_Control()
{

}

SSM_SCTP_Control::~SSM_SCTP_Control()
{

}



static void sci_timer_expired(TimerID timerID, void *associationIDvoid, void *unused)
{
    unsigned int state;
    boolean initFailed;
    ChunkID shutdownCID;
    ChunkID shutdownAckCID;
    guint primary;

    if (SSM_Distribution::mdi_setAssociationData(*((unsigned int *) associationIDvoid))) {
        /* error log: expired timer refers to a non existent association. */
        error_logi(ERROR_MAJOR,
                   "init timer expired but association %u does not exist",
                   (*(unsigned int *) associationIDvoid));
        return;
    }


    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log: association exist, but has no SCTP-control ? */
        error_log(ERROR_MAJOR, "Association without SCTP-control");
        return;
    }

    state = localData->association_state;
    initFailed = FALSE;

    primary = SSM_Pathmanagement::pm_readPrimaryPath();

    event_logiii(VERBOSE, "sci_timer_expired(AssocID=%u,  state=%u, Primary=%u",
        (*(unsigned int *) associationIDvoid), state, primary);

    switch (state) {
    case COOKIE_WAIT:

        event_log(EXTERNAL_EVENT, "init timer expired in state COOKIE_WAIT");

        if (localData->initRetransCounter < localData->assocMaxInitRetransmissions) {
            /* increase retransmissission-counter, resend init and restart init-timer */
            localData->initRetransCounter++;
            SSM_Bundling::bu_put_Ctrl_Chunk((SCTP_simple_chunk *) localData->initChunk,NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            /* restart init timer after timer backoff */
            localData->initTimerDuration = min(localData->initTimerDuration * 2, (unsigned int)SSM_Pathmanagement::pm_getRtoMax());
            event_logi(INTERNAL_EVENT_0, "init timer backedoff %d msecs",
                       localData->initTimerDuration);
            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_INIT,
                                (void *) &localData->associationID, NULL);
        } else {
            /* log error to log-file */
            event_log(EXTERNAL_EVENT,
                      "init retransmission counter exeeded threshold in state COOKIE_WAIT");
            /* report error to ULP tbd: status */
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_EXCEEDED_RETRANSMISSIONS);
            /* free memory for initChunk */
            free(localData->initChunk);
            localData->initTimer = 0;
            localData->initChunk = NULL;
            /* delete this association */
            initFailed = TRUE;
        }
        break;

    case COOKIE_ECHOED:

        event_log(EXTERNAL_EVENT, "cookie timer expired in state COOKIE_ECHOED");

        if (localData->initRetransCounter < localData->assocMaxInitRetransmissions) {
            /* increase retransmissission-counter, resend init and restart init-timer */
            localData->initRetransCounter++;
            SSM_Bundling::bu_put_Ctrl_Chunk((SCTP_simple_chunk *) localData->cookieChunk,NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            /* restart cookie timer after timer backoff */
            localData->initTimerDuration = min(localData->initTimerDuration * 2, (unsigned int)SSM_Pathmanagement::pm_getRtoMax());
            event_logi(INTERNAL_EVENT_0, "cookie timer backedoff %d msecs",
                       localData->initTimerDuration);

            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_INIT,
                                (void *) &localData->associationID, NULL);
        } else {
            /* log error to log-file */
            event_log(EXTERNAL_EVENT,
                      "init retransmission counter exeeded threshold; state: COOKIE_ECHOED");
            /* report error to ULP tbd: status */
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_EXCEEDED_RETRANSMISSIONS);
            /* free memory for cookieChunk */
            free(localData->cookieChunk);
            localData->initTimer = 0;
            localData->cookieChunk = NULL;
            /* delete this association */
            initFailed = TRUE;
        }
        break;

    case SHUTDOWNSENT:

        /* some of the variable names are missleading, because they where only used
           for init, but are reused for shutdown after the shutdown timer was introduced
           in the draft. */

        if (localData->initRetransCounter < localData->assocMaxRetransmissions) {
            /* increase retransmissission-counter */
            localData->initRetransCounter++;

            /* make and send shutdown again, with updated TSN (section 9.2)     */
            shutdownCID = SSM_ChunkHandler::ch_makeShutdown(SSM_Recvctrl::rxc_read_cummulativeTSNacked());
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownCID),&primary);
            SSM_Bundling::bu_sendAllChunks(&primary);
            SSM_ChunkHandler::ch_deleteChunk(shutdownCID);

            /* restart shutdown timer after timer backoff */
            localData->initTimerDuration = min(localData->initTimerDuration * 2, (unsigned int)SSM_Pathmanagement::pm_getRtoMax());
            event_logi(INTERNAL_EVENT_0, "shutdown timer backed off %d msecs",
                       localData->initTimerDuration);

            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired, TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);
        } else {
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_EXCEEDED_RETRANSMISSIONS);
            localData->initTimer = 0;
            /* shut down failed, delete current association. */
            initFailed = TRUE;
        }
        break;

    case SHUTDOWNACKSENT:

        /* some of the variable names are missleading, because they where only used
           for init, but are reused for shutdown */

        if (localData->initRetransCounter < localData->assocMaxRetransmissions) {
            /* increase retransmissission-counter */
            localData->initRetransCounter++;

            shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),&primary);
            SSM_Bundling::bu_sendAllChunks(&primary);
            SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);

            /* COMMENTED OUT BECAUSE PROBABLY VERY WRONG............. */
            /* make and send shutdown_complete again */
            /* shutdown_complete_CID = ch_makeSimpleChunk(CHUNK_SHUTDOWN_COMPLETE, FLAG_NONE); */
            /* bu_put_Ctrl_Chunk(ch_chunkString(shutdown_complete_CID)); */
            /* bu_sendAllChunks(&primary); */
            /* ch_deleteChunk(shutdown_complete_CID); */

            /* restart shutdown timer after timer backoff */
            localData->initTimerDuration = min(localData->initTimerDuration * 2, (unsigned int)SSM_Pathmanagement::pm_getRtoMax());
            event_logi(INTERNAL_EVENT_0, "shutdown timer backed off %d msecs",
                       localData->initTimerDuration);
            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired, TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);
        } else {
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_EXCEEDED_RETRANSMISSIONS);
            localData->initTimer = 0;
            /* shut down failed, delete current association. */
            initFailed = TRUE;
        }
        break;

    default:
        /* error log */
        error_logi(ERROR_MAJOR, "unexpected event: timer expired in state %02d", state);
        localData->initTimer = 0;
        break;
    }

    localData = NULL;
    if (initFailed == TRUE) {
        SSM_Distribution::mdi_deleteCurrentAssociation();
    }

    SSM_Distribution::mdi_clearAssociationData();
}


/*------------------- Functions called by the ULP via message-distribution -----------------------*/

/**
 * This function is called to initiate the setup an association.
 *
 * The local tag and the initial TSN are randomly generated.
 * Together with the parameters of the function, they are used to create the init-message.
 * This data are also stored in a newly created association-record.
 *
 * @param noOfOutStreams        number of send streams.
 * @param noOfInStreams         number of receive streams.
 */
void SSM_SCTP_Control::scu_associate(unsigned short noOfOutStreams,
                   unsigned short noOfInStreams,
                   union sockunion* destinationList,
                   unsigned int numDestAddresses,
                   gboolean withPRSCTP)
{
    guint32 state;
    guint16 nlAddresses;
    union sockunion lAddresses[MAX_NUM_ADDRESSES];
    ChunkID initCID;
    unsigned int supportedTypes = 0, count;

    /* ULP has called sctp_associate at distribution.
       Distribution has allready allocated the association data and partially initialized */

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        error_log(ERROR_MAJOR, "read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case CLOSED:
        event_log(EXTERNAL_EVENT, "event: scu_assocatiate in state CLOSED");
        /* create init chunk and write data to it -- take AssocID as tag !!! */
        initCID = SSM_ChunkHandler::ch_makeInit(SSM_Distribution::mdi_readLocalTag(),
                              SSM_Distribution::mdi_getDefaultMyRwnd(),
                              noOfOutStreams, noOfInStreams, SSM_Distribution::mdi_generateStartTSN());

        /* store the number of streams */
        localData->NumberOfOutStreams = noOfOutStreams;
        localData->NumberOfInStreams = noOfInStreams;

        supportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();

        /* enter enter local addresses to message. I send an Init here, so
         * I will include all of my addresses !
         */
        SSM_Distribution::mdi_readLocalAddresses(lAddresses,
                               &nlAddresses,
                               destinationList,
                               numDestAddresses,
                               supportedTypes,
                               FALSE);

        event_logi(VERBOSE, "1: supportedTypes : %u", supportedTypes);

        if (withPRSCTP) {
           SSM_ChunkHandler:: ch_addParameterToInitChunk(initCID, VLPARAM_PRSCTP, 0, NULL);
        }

#ifdef BAKEOFF
         ch_addParameterToInitChunk(initCID, 0x8123, 17, (unsigned char*)localData);
         ch_addParameterToInitChunk(initCID, 0x8343, 23, (unsigned char*)localData);
         ch_addParameterToInitChunk(initCID, 0x8324, 1, (unsigned char*)localData);
         ch_addParameterToInitChunk(initCID, 0xC123, 31, (unsigned char*)localData);
#endif


        SSM_ChunkHandler::ch_enterSupportedAddressTypes(initCID, TRUE, FALSE, FALSE);


        event_logi(VERBOSE, "2: supportedTypes : %u", supportedTypes);

        if (nlAddresses > 1)
            SSM_ChunkHandler::ch_enterIPaddresses(initCID, lAddresses, nlAddresses);


        localData->initChunk = (SCTP_init *) SSM_ChunkHandler::ch_chunkString(initCID);
        SSM_ChunkHandler::ch_forgetChunk(initCID);

        /* send init chunk */
        for (count = 0; count < numDestAddresses; count++) {
            SSM_Bundling::bu_put_Ctrl_Chunk((SCTP_simple_chunk *) localData->initChunk, &count);
            SSM_Bundling::bu_sendAllChunks(&count);
        }

        localData->cookieChunk = NULL;
        localData->local_tie_tag = 0;
        localData->peer_tie_tag = 0;

        /* start init timer */
        localData->initTimerDuration = SSM_Pathmanagement::pm_readRTO(SSM_Pathmanagement::pm_readPrimaryPath());

        if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);

        localData->initTimer = SSM_Adaptation::adl_startTimer(localData->initTimerDuration,
                                               &sci_timer_expired,
                                               TIMER_TYPE_INIT,
                                               (void *) &localData->associationID, NULL);

        state = COOKIE_WAIT;
        break;
    default:
        error_logi(EXTERNAL_EVENT_X, "Erroneous Event : scu_associate called in state %u", state);
        break;
    }

    localData->association_state = state;
    localData = NULL;
}



/**
 * function initiates the shutdown of this association.
 */
void SSM_SCTP_Control::scu_shutdown()
{
    guint32 state;
    ChunkID shutdownCID;
    boolean readyForShutdown;

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case ESTABLISHED:
        event_log(EXTERNAL_EVENT, "event: scu_shutdown in state ESTABLISHED");

        /* disable heartbeat */
        SSM_Pathmanagement::pm_disableAllHB();

        /* stop reliable transfer and read its state */
        readyForShutdown = (SSM_Reltransfer::rtx_readNumberOfUnackedChunks() == 0) &&
            (SSM_Flowcontrol::fc_readNumberOfQueuedChunks() == 0);

        if (readyForShutdown) {
            /* make and send shutdown */
            shutdownCID = SSM_ChunkHandler::ch_makeShutdown(SSM_Recvctrl::rxc_read_cummulativeTSNacked());
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownCID),NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            SSM_ChunkHandler::ch_deleteChunk(shutdownCID);

            /* start shutdown timer */
            localData->initTimerDuration = SSM_Pathmanagement::pm_readRTO(SSM_Pathmanagement::pm_readPrimaryPath());

            if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);

            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);

            localData->initRetransCounter = 0;

            /* receive control must acknoweledge every datachunk at once after the shutdown
               was sent. */
            SSM_Recvctrl::rxc_send_sack_everytime();

            state = SHUTDOWNSENT;
        } else {
            /* shutdown in progress info to reliable transfer, this stopps data transmission */
            SSM_Reltransfer::rtx_shutdown();
            /* wait for sci_allChunksAcked from reliable transfer */
            state = SHUTDOWNPENDING;
        }
        localData->association_state = state;
        localData = NULL;

        break;
    case CLOSED:
    case COOKIE_WAIT:
    case COOKIE_ECHOED:        /* Siemens convention: ULP can not send datachunks
                                   until it has received the communication up. */
        event_logi(EXTERNAL_EVENT, "event: scu_shutdown in state %02d --> aborting", state);
        scu_abort(ECC_USER_INITIATED_ABORT, 0, NULL);
        break;
    case SHUTDOWNSENT:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNPENDING:
    case SHUTDOWNACKSENT:
        /* ignore, keep on waiting for completion of the running shutdown */
        event_logi(EXTERNAL_EVENT, "event: scu_shutdown in state %", state);
        localData = NULL;
        break;
    default:
        /* error logging */
        event_log(EXTERNAL_EVENT_X, "unexpected event: scu_shutdown");
        localData = NULL;
        break;
    }
}



static void sci_add_abort_error_cause(ChunkID abortChunk,
                                unsigned short etype,
                                unsigned short eplen,
                                unsigned char* epdata)
{

    switch (etype) {

        case ECC_INVALID_STREAM_ID:
        case ECC_MISSING_MANDATORY_PARAM:
        case ECC_STALE_COOKIE_ERROR:
        case ECC_OUT_OF_RESOURCE_ERROR:
        case ECC_UNRESOLVABLE_ADDRESS:
        case ECC_UNRECOGNIZED_CHUNKTYPE:
        case ECC_INVALID_MANDATORY_PARAM:
        case ECC_UNRECOGNIZED_PARAMS:
        case ECC_NO_USER_DATA:
        case ECC_COOKIE_RECEIVED_DURING_SHUTDWN:
        case ECC_RESTART_WITH_NEW_ADDRESSES:
        case ECC_USER_INITIATED_ABORT:
            SSM_ChunkHandler::ch_enterErrorCauseData(abortChunk, etype, eplen, epdata);
            break;
        default:
            break;
    }
    return;
}



/**
 * this function aborts this association. And optionally adds an error parameter
 * to the ABORT chunk that is sent out.
 */
void SSM_SCTP_Control::scu_abort(short error_type, unsigned short error_param_length, unsigned char* error_param_data)
{
    guint32 state;
    ChunkID abortCID;
    gboolean removed = FALSE;

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case CLOSED:
        event_log(EXTERNAL_EVENT, "event: scu_abort in state CLOSED");
        /* delete all data of this association */
        SSM_Distribution::mdi_deleteCurrentAssociation();
        SSM_Distribution::mdi_clearAssociationData();

        break;

    case COOKIE_WAIT:
    case COOKIE_ECHOED:
    case SHUTDOWNSENT:
    case SHUTDOWNACKSENT:
        event_logi(EXTERNAL_EVENT, "event: scu_abort in state %2d --> send abort", state);

        /* make and send abort message */
        abortCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NONE);

        if (error_type >= 0) {
            sci_add_abort_error_cause(abortCID,  (unsigned short)error_type, error_param_length, error_param_data);
        }
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(abortCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_Bundling::bu_unlock_sender(NULL);
        /* free abort chunk */
        SSM_ChunkHandler::ch_deleteChunk(abortCID);
        /* stop init timer */
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        }
        /* delete all data of this association */
        SSM_Distribution::mdi_deleteCurrentAssociation();
        removed = TRUE;
        break;

    case ESTABLISHED:
    case SHUTDOWNPENDING:
    case SHUTDOWNRECEIVED:

        event_logi(EXTERNAL_EVENT, "event: scu_abort in state %02d --> send abort", state);

        /* make and send abort message */
        abortCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NONE);

        if (error_type >= 0) {
            sci_add_abort_error_cause(abortCID,  (unsigned short)error_type, error_param_length,error_param_data);
        }

        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(abortCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_Bundling::bu_unlock_sender(NULL);
        /* free abort chunk */
        SSM_ChunkHandler::ch_deleteChunk(abortCID);
        /* delete all data of this association */
        SSM_Distribution::mdi_deleteCurrentAssociation();
        removed = TRUE;

        break;
    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "scu_abort in state %02d: unexpected event", state);
        break;
    }

    if (removed == TRUE) {
        SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_ABORTED);
        SSM_Distribution::mdi_clearAssociationData();
    }
}


/*------------------- Functions called by the (de-)bundling for received control chunks ----------*/

/**
 * sctlr_init is called by bundling when a init message is received from the peer.
 * an InitAck may be returned, alongside with a cookie chunk variable parameter.
 * The following data are created and included in the init acknowledgement:
 * a COOKIE parameter.
 * @param init  pointer to the received init-chunk (including optional parameters)
 */
int SSM_SCTP_Control::sctlr_init(SCTP_init * init)
{
    /*  this function does not expect any data allocated for the new association,
       but if there are, implementation will act according to section 5.2.1 (simultaneous
       initialization) and section 5.2.2 (duplicate initialization)
     */

    unsigned int state;
    guint16 nlAddresses;
    union sockunion lAddresses[MAX_NUM_ADDRESSES];
    guint16 nrAddresses;
    union sockunion rAddresses[MAX_NUM_ADDRESSES];
    union sockunion last_source;

    ChunkID initCID;
    ChunkID initCID_local;
    ChunkID initAckCID;
    ChunkID abortCID;
    ChunkID shutdownAckCID;
    unsigned short inbound_streams, outbound_streams;
    unsigned int supportedTypes=0, peerSupportedTypes=0;
    int process_further, result;
    int return_state = STATE_OK;

    guint primary;

    event_log(EXTERNAL_EVENT, "sctlr_init() is executed");

    initCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) init);

    if (SSM_ChunkHandler::ch_chunkType(initCID) != CHUNK_INIT) {
        /* error logging */
        SSM_ChunkHandler::ch_forgetChunk(initCID);
        error_log(ERROR_MAJOR, "sctlr_init: wrong chunk type");
        return return_state;
    }

    if (SSM_ChunkHandler::ch_noOutStreams(initCID) == 0 || SSM_ChunkHandler::ch_noInStreams(initCID) == 0 || SSM_ChunkHandler::ch_initiateTag(initCID) == 0) {
        event_log(EXTERNAL_EVENT, "event: received init with zero number of streams, or zero TAG");

        /* make and send abort message */
        abortCID =SSM_ChunkHandler:: ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NONE);
        SSM_ChunkHandler::ch_enterErrorCauseData(abortCID, ECC_INVALID_MANDATORY_PARAM, 0, NULL);

        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(abortCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        /* free abort chunk */
        SSM_ChunkHandler::ch_deleteChunk(abortCID);
        /* delete all data of this association */
        if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) != NULL) {
            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_Distribution::mdi_deleteCurrentAssociation();
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_INVALID_PARAMETER);
            SSM_Distribution::mdi_clearAssociationData();

            return_state = STATE_STOP_PARSING_REMOVED;
        }
        return return_state;
    }

    result = SSM_Distribution::mdi_readLastFromAddress(&last_source);
    if (result != 0) {
        if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
            SSM_Distribution::mdi_clearAssociationData();
            return_state = STATE_STOP_PARSING_REMOVED;
            return return_state;
        }
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        }
        SSM_Bundling::bu_unlock_sender(NULL);
        SSM_Distribution::mdi_deleteCurrentAssociation();
        SSM_Distribution::mdi_communicationLostNotif(0);
        SSM_Distribution::mdi_clearAssociationData();
        return_state = STATE_STOP_PARSING_REMOVED;
        return return_state;
    }

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        event_log(VERBOSE, " DO_5_1_B_INIT: Normal init case ");
        /* DO_5_1_B_INIT : Normal case, no association exists yet */
        /* save a-sides init-tag from init-chunk to be used as a verification tag of the sctp-
           message carrying the initAck (required since no association is created). */
        SSM_Distribution::mdi_writeLastInitiateTag(SSM_ChunkHandler::ch_initiateTag(initCID));

        /* Limit the number of sendstreams a-side requests to the max. number of input streams
           this z-side is willing to accept.
         */
        inbound_streams = min(SSM_ChunkHandler::ch_noOutStreams(initCID), SSM_Distribution::mdi_readLocalInStreams());
        outbound_streams = min(SSM_ChunkHandler::ch_noInStreams(initCID), SSM_Distribution::mdi_readLocalOutStreams());
        /* fire back an InitAck with a Cookie */
        initAckCID = SSM_ChunkHandler::ch_makeInitAck(SSM_Distribution::mdi_generateTag(),
                                    SSM_Distribution::mdi_getDefaultMyRwnd(),
                                    outbound_streams,
                                    inbound_streams, SSM_Distribution::mdi_generateStartTSN());

        /* retreive a-side source addresses from message */
        supportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();

        nrAddresses = SSM_ChunkHandler::ch_IPaddresses(initCID, supportedTypes, rAddresses, &peerSupportedTypes, &last_source);


        if ((supportedTypes & peerSupportedTypes) == 0)
            error_log(ERROR_FATAL, "BAKEOFF: Program error, no common address types in sctlr_init()");

        /* enter variable length params initAck */
        SSM_Distribution::mdi_readLocalAddresses(lAddresses, &nlAddresses, &last_source, 1, peerSupportedTypes,TRUE);
        /* enter local addresses into initAck */
        if (nlAddresses > 1)
            SSM_ChunkHandler::ch_enterIPaddresses(initAckCID, lAddresses, nlAddresses);

        /* append cookie to InitAck Chunk */
        SSM_ChunkHandler::ch_enterCookieVLP(initCID, initAckCID,
                         SSM_ChunkHandler:: ch_initFixed(initCID), SSM_ChunkHandler::ch_initFixed(initAckCID),
                          SSM_ChunkHandler::ch_cookieLifeTime(initCID), 0, /* tie tags are both zero */
                          0, lAddresses, nlAddresses, rAddresses, nrAddresses);


        process_further = SSM_ChunkHandler::ch_enterUnrecognizedParameters(initCID, initAckCID, supportedTypes);

        if (process_further == -1) {
         /*   ch_deleteChunk(initAckCID);
            ch_forgetChunk(initCID); */
            return_state = STATE_STOP_PARSING; /* to stop parsing without actually removing it */
            /* return return_state; */
        } else {
            if (process_further == 1) {
                return_state = STATE_STOP_PARSING; /* to stop parsing without actually removing it */
            }
            /* send initAck */
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(initAckCID),NULL);
        }
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_Bundling::bu_unlock_sender(NULL);
        SSM_ChunkHandler::ch_deleteChunk(initAckCID);
        event_log(INTERNAL_EVENT_1, "event: initAck sent");
    } else {
        /* save a-sides init-tag from init-chunk to be used as a verification tag of the sctp-
           message carrying the initAck (required since peer may have changed the verification
           tag).
           mdi_writeLastInitiateTag(ch_initiateTag(initCID)); */

        state = localData->association_state;
        event_logi(EXTERNAL_EVENT, "sctlr_init: received INIT chunk in state %02u", state);
        primary = SSM_Pathmanagement::pm_readPrimaryPath();
        supportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();

        switch (state) {
            /* see section 5.2.1 */
        case COOKIE_WAIT:
            if ((localData->local_tie_tag != 0) || (localData->peer_tie_tag != 0)) {
                error_logii(ERROR_FATAL, "Tie tags NOT zero in COOKIE_WAIT, but %u and %u",
                            localData->local_tie_tag, localData->peer_tie_tag);
            }
            localData->local_tie_tag = 0;
            localData->peer_tie_tag = 0;

        case COOKIE_ECHOED:
            if ((state == COOKIE_ECHOED) &&
                ((localData->local_tie_tag == 0) || (localData->peer_tie_tag == 0))) {
                error_logii(ERROR_FATAL, "Tie tags zero in COOKIE_ECHOED, local: %u, peer: %u",
                            localData->local_tie_tag, localData->peer_tie_tag);
            }

            if (state == COOKIE_ECHOED) {
                /*
                 * For an endpoint that is in the COOKIE-ECHOED state it MUST populate
                 * its Tie-Tags with random values so that possible attackers cannot guess
                 * real tag values of the association (see Implementer's Guide > version 10)
                 */
                localData->local_tie_tag = SSM_Distribution::mdi_generateTag();
                localData->peer_tie_tag = SSM_Distribution::mdi_generateTag();
            }

            /* save remote  tag ?
               mdi_writeLastInitiateTag(ch_initiateTag(initCID)); */
            inbound_streams = min(SSM_ChunkHandler::ch_noOutStreams(initCID), SSM_Distribution::mdi_readLocalInStreams( ));

            /* Set length of chunk to HBO !! */
            initCID_local = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) localData->initChunk);
            /* section 5.2.1 : take original parameters from first INIT chunk */
            initAckCID = SSM_ChunkHandler::ch_makeInitAck(SSM_ChunkHandler::ch_initiateTag(initCID_local),
                                        SSM_ChunkHandler::ch_receiverWindow(initCID_local),
                                        SSM_ChunkHandler::ch_noInStreams(initCID), /* peers inbound are MY outbound */
                                        inbound_streams,
                                        SSM_ChunkHandler::ch_initialTSN(initCID_local));

            /* reset length field again to NBO...and remove reference */
            SSM_ChunkHandler::ch_chunkString(initCID_local);
            SSM_ChunkHandler::ch_forgetChunk(initCID_local);

            /* retreive a-side source addresses from message */
            nrAddresses =SSM_ChunkHandler:: ch_IPaddresses(initCID, supportedTypes, rAddresses, &peerSupportedTypes, &last_source);

            /* the initAck (and consequently the Cookie) will contain my assocID as my local
               tag, and the peers tag from the init we got here */
            SSM_Distribution::mdi_readLocalAddresses(lAddresses, &nlAddresses, &last_source, 1, peerSupportedTypes, TRUE);
            /* enter local addresses into initAck */
            if (nlAddresses > 1)
                SSM_ChunkHandler::ch_enterIPaddresses(initAckCID, lAddresses, nlAddresses);

            SSM_ChunkHandler::ch_enterCookieVLP(initCID, initAckCID,
                              SSM_ChunkHandler::ch_initFixed(initCID),
                              SSM_ChunkHandler::ch_initFixed(initAckCID),
                              SSM_ChunkHandler::ch_cookieLifeTime(initCID),
                              localData->local_tie_tag, /* tie tags may be zero OR populated here */
                              localData->peer_tie_tag,
                              lAddresses, nlAddresses,
                              rAddresses, nrAddresses);

            process_further = SSM_ChunkHandler::ch_enterUnrecognizedParameters(initCID, initAckCID, supportedTypes);

            if (process_further == -1) {
                SSM_ChunkHandler::ch_deleteChunk(initAckCID);
                SSM_ChunkHandler::ch_forgetChunk(initCID);
                return_state =STATE_STOP_PARSING ; /* to stop parsing without actually removing it */
                return return_state;
            } else if (process_further == 1) {
                return_state = STATE_STOP_PARSING; /* to stop parsing without actually removing it */
            }

            /* send initAck */
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(initAckCID),NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_ChunkHandler::ch_deleteChunk(initAckCID);
            event_logi(INTERNAL_EVENT_1, "event: initAck sent in state %u", state);
            break;

            /* see section 5.2.2 */
        case ESTABLISHED:
        case SHUTDOWNPENDING:
        case SHUTDOWNRECEIVED:
        case SHUTDOWNSENT:
            if ((localData->local_tie_tag == 0) || (localData->peer_tie_tag == 0)) {
                error_logiii(ERROR_MINOR, "Tie tags zero in state %u, local: %u, peer: %u --> Restart ?",
                             state, localData->local_tie_tag, localData->peer_tie_tag);
            }

            inbound_streams = min(SSM_ChunkHandler::ch_noOutStreams(initCID), SSM_Distribution::mdi_readLocalInStreams());

            initAckCID = SSM_ChunkHandler::ch_makeInitAck(SSM_Distribution::mdi_generateTag(),
                                        SSM_Recvctrl::rxc_get_local_receiver_window(),
                                        SSM_Streamengine::se_numOfSendStreams(), SSM_Streamengine::se_numOfRecvStreams(),
                                        /* TODO : check whether we take NEW TSN or leave an old one */
                                        SSM_Distribution::mdi_generateStartTSN());

            /*
               localData->local_tie_tag = mdi_generateTag();
               localData->peer_tie_tag = mdi_generateTag();
             */

            /* retreive remote source addresses from message */
            nrAddresses = SSM_ChunkHandler::ch_IPaddresses(initCID, supportedTypes, rAddresses, &peerSupportedTypes, &last_source);

            SSM_Distribution::mdi_readLocalAddresses(lAddresses, &nlAddresses, &last_source, 1, peerSupportedTypes, TRUE);

            /* enter local addresses into initAck */
            if (nlAddresses > 1)
                SSM_ChunkHandler::ch_enterIPaddresses(initAckCID, lAddresses, nlAddresses);

            SSM_ChunkHandler::ch_enterCookieVLP(initCID, initAckCID,
                              SSM_ChunkHandler::ch_initFixed(initCID),
                              SSM_ChunkHandler::ch_initFixed(initAckCID),
                              SSM_ChunkHandler::ch_cookieLifeTime(initCID),
                              localData->local_tie_tag, /* this should be different from that in Init_Ack now */
                              localData->peer_tie_tag,
                              lAddresses, nlAddresses,
                              rAddresses, nrAddresses);

            process_further = SSM_ChunkHandler::ch_enterUnrecognizedParameters(initCID, initAckCID, supportedTypes);

            if (process_further == -1) {
                SSM_ChunkHandler::ch_deleteChunk(initAckCID);
                SSM_ChunkHandler::ch_forgetChunk(initCID);
                return_state = STATE_STOP_PARSING;
                return return_state;
            } else if (process_further == 1) {
                return_state = STATE_STOP_PARSING; /* to stop parsing without actually removing it */
            }

            /* send initAck */
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(initAckCID),NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_ChunkHandler::ch_deleteChunk(initAckCID);
            event_logi(INTERNAL_EVENT_1, "event: initAck sent in state %u", state);
            break;
        case SHUTDOWNACKSENT:
            /* We are supposed to discard the Init, and retransmit SHUTDOWN_ACK (9.2) */
            shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);
            break;
        default:
            error_logi(ERROR_MAJOR, "Unexpected State %02u - Program Error ???", state);
            break;
        }
    }

    /* was only treated with ch_makeChunk -- it is enough to "FORGET" it */
    SSM_ChunkHandler::ch_forgetChunk(initCID);
    return return_state;
}


/**
 * sctlr_initAck is called by bundling when a init acknowledgement was received from the peer.
 * The following data are retrieved from the init-data and saved for this association:
 * \begin{itemize}
 * \item remote tag from the initiate tag field
 * \item receiver window credit of the peer
 * \item number of send streams of the peer, must be lower or equal the number of receive streams
 *   this host has announced with the init-chunk
 * \item number of receive streams the peer allows the receiver of this initAck to use
 * \end{itemize}
 * The initAck must contain a cookie which is returned to the peer with the cookie acknowledgement.
 * @param initAck  pointer to received initAck-chunk including optional parameters without chunk header
 */
gboolean SSM_SCTP_Control::sctlr_initAck(SCTP_init * initAck)
{
    guint32 state;
    int result;
    unsigned int index=0;
    union sockunion destAddress;
    union sockunion dAddresses[MAX_NUM_ADDRESSES];
    unsigned int ndAddresses;
    unsigned short inbound_streams;
    unsigned short outbound_streams;

    unsigned int peerSupportedTypes=0, supportedTypes=0;
    int process_further = 0;
    ChunkID cookieCID;
    ChunkID initCID;
    ChunkID initAckCID;
    ChunkID errorCID, abortCID;
    SCTP_MissingParams missing_params;
    int return_state = STATE_OK;

    union sockunion preferredPrimary;
    gboolean preferredSet      = FALSE;
    gboolean peerSupportsPRSCTP = FALSE;
    gboolean peerSupportsADDIP = FALSE;
    gboolean peerSupportsIPV4 = FALSE;
    gboolean peerSupportsIPV6 = FALSE;
    short preferredPath;

    initAckCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) initAck);

    if (SSM_ChunkHandler::ch_chunkType(initAckCID) != CHUNK_INIT_ACK) {
        /* error logging */
        SSM_ChunkHandler::ch_forgetChunk(initAckCID);
        error_log(ERROR_MAJOR, "sctlr_initAck: wrong chunk type");
        return return_state;
    }

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        SSM_ChunkHandler::ch_forgetChunk(initAckCID);
        error_log(ERROR_MAJOR, "sctlr_initAck: read SCTP-control failed");
        return return_state;
    }

    state = localData->association_state;

    switch (state) {
    case COOKIE_WAIT:

        event_log(EXTERNAL_EVENT, "event: initAck in state COOKIE_WAIT");

        /* Set length of chunk to HBO !! */
        initCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) localData->initChunk);

        /* FIXME: check also the noPeerOutStreams <= noLocalInStreams */
        if (SSM_ChunkHandler::ch_noOutStreams(initAckCID) == 0 || SSM_ChunkHandler::ch_noInStreams(initAckCID) == 0 || SSM_ChunkHandler::ch_initiateTag(initAckCID) == 0) {
            if (localData->initTimer != 0) {
                sctp_stopTimer(localData->initTimer);
                localData->initTimer = 0;
            }
           /* make and send abort message */
            abortCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NONE);
            SSM_ChunkHandler::ch_enterErrorCauseData(abortCID, ECC_INVALID_MANDATORY_PARAM, 0, NULL);
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(abortCID),NULL);
            SSM_ChunkHandler::ch_deleteChunk(abortCID);

            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            /* delete all data of this association */
            SSM_Distribution::mdi_deleteCurrentAssociation();
            SSM_Distribution::mdi_communicationLostNotif(0);
            SSM_Distribution::mdi_clearAssociationData();
            return_state = STATE_STOP_PARSING_REMOVED;
            return return_state;
        }

        result = SSM_Distribution::mdi_readLastFromAddress(&destAddress);
        if (result != 0) {
            if (localData->initTimer != 0) {
                sctp_stopTimer(localData->initTimer);
                localData->initTimer = 0;
            }
            SSM_Bundling::bu_unlock_sender(NULL);
            SSM_Distribution::mdi_deleteCurrentAssociation();
            SSM_Distribution::mdi_communicationLostNotif(0);
            SSM_Distribution::mdi_clearAssociationData();
            return_state = STATE_STOP_PARSING_REMOVED;
            return return_state;
        }

        supportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();
        /* retrieve addresses from initAck */
        ndAddresses =SSM_ChunkHandler:: ch_IPaddresses(initAckCID, supportedTypes, dAddresses, &peerSupportedTypes, &destAddress);

       SSM_Distribution:: mdi_writeDestinationAddresses(dAddresses, ndAddresses);

        /* initialize rest of association with data received from peer */

        inbound_streams = min(SSM_ChunkHandler::ch_noOutStreams(initAckCID), localData->NumberOfInStreams);
        outbound_streams = min(SSM_ChunkHandler::ch_noInStreams(initAckCID), localData->NumberOfOutStreams);

        peerSupportsPRSCTP = SSM_ChunkHandler::ch_getPRSCTPfromInitAck(initAckCID);


        SSM_Distribution::mdi_initAssociation(SSM_ChunkHandler::ch_receiverWindow(initAckCID), /* remotes side initial rwnd */
                            inbound_streams, /* # of remote output/local input streams */
                            outbound_streams, /* # of remote input/local output streams */
                            SSM_ChunkHandler::ch_initialTSN(initAckCID), /* remote initial TSN */
                            SSM_ChunkHandler::ch_initiateTag(initAckCID), /* remote init tag */
                            SSM_ChunkHandler::ch_initialTSN(initCID), /* local initial TSN for sending */
                            peerSupportsPRSCTP,
                            FALSE);

       event_logii(VERBOSE, "sctlr_InitAck(): called mdi_initAssociation(in-streams=%u, out-streams=%u)",
                    inbound_streams,outbound_streams);


        /* reset length field again to NBO... */
        SSM_ChunkHandler::ch_chunkString(initCID),
        /* free initChunk memory */
        SSM_ChunkHandler::ch_forgetChunk(initCID);

        cookieCID = SSM_ChunkHandler::ch_makeCookie(SSM_ChunkHandler::ch_cookieParam(initAckCID));

        if (cookieCID < 0) {
            event_log(EXTERNAL_EVENT, "received a initAck without cookie");

            /* stop shutdown timer */
            if (localData->initTimer != 0) {
                sctp_stopTimer(localData->initTimer);
                localData->initTimer = 0;
            }
            missing_params.numberOfParams = htonl(1);
            missing_params.params[0] = htons(VLPARAM_COOKIE);

            scu_abort(ECC_MISSING_MANDATORY_PARAM, 6, (unsigned char*)&missing_params);
            SSM_Bundling::bu_unlock_sender(NULL);
            /* delete this association */

            return_state = STATE_STOP_PARSING_REMOVED;
            localData->association_state = CLOSED;
            localData = NULL;
            return return_state;
        }


        process_further = SSM_ChunkHandler::ch_enterUnrecognizedErrors(initAckCID,
                                                     supportedTypes,
                                                     &errorCID,
                                                     &preferredPrimary,
                                                     &preferredSet,
                                                     &peerSupportsIPV4,
                                                     &peerSupportsIPV6,
                                                     &peerSupportsPRSCTP,
                                                     &peerSupportsADDIP);


        if (process_further == -1) {
            SSM_ChunkHandler::ch_forgetChunk(initAckCID);
            SSM_ChunkHandler::ch_deleteChunk(cookieCID);
            if (errorCID != 0) SSM_ChunkHandler::ch_deleteChunk(errorCID);
            SSM_Bundling::bu_unlock_sender(NULL);
            if (localData->initTimer != 0) {
                sctp_stopTimer(localData->initTimer);
                localData->initTimer = 0;
            }
            SSM_Distribution::mdi_deleteCurrentAssociation();
            SSM_Distribution:: mdi_communicationLostNotif(SCTP_COMM_LOST_FAILURE);
            SSM_Distribution::mdi_clearAssociationData();
            localData->association_state = CLOSED;
            localData = NULL;
            return_state = STATE_STOP_PARSING_REMOVED;
            return return_state;
        } else if (process_further == 1) {
            return_state = STATE_STOP_PARSING;
        }

        localData->cookieChunk = (SCTP_cookie_echo *) SSM_ChunkHandler::ch_chunkString(cookieCID);
        /* populate tie tags -> section 5.2.1/5.2.2 */
        localData->local_tie_tag = SSM_Distribution::mdi_readLocalTag();
        localData->peer_tie_tag = SSM_ChunkHandler::ch_initiateTag(initAckCID);


        localData->NumberOfOutStreams = outbound_streams;
        localData->NumberOfInStreams =  inbound_streams;


        SSM_ChunkHandler::ch_forgetChunk(cookieCID);
        SSM_ChunkHandler::ch_forgetChunk(initAckCID);

        /* send cookie back to the address where we got it from     */
        for (index = 0; index < ndAddresses; index++)
            if (SSM_Adaptation::adl_equal_address(&(dAddresses[index]),&destAddress)) break;

        /* send cookie */
        SSM_Bundling::bu_put_Ctrl_Chunk((SCTP_simple_chunk *) localData->cookieChunk, &index);
        if (errorCID != 0) {
            SSM_Bundling::bu_put_Ctrl_Chunk((SCTP_simple_chunk *)SSM_ChunkHandler::ch_chunkString(errorCID), &index);
            SSM_ChunkHandler::ch_deleteChunk(errorCID);
        }

        SSM_Bundling::bu_sendAllChunks(&index);
        SSM_Bundling::bu_unlock_sender(&index);
        event_logi(INTERNAL_EVENT_1, "event: sent cookie echo to PATH %u", index);

        if (preferredSet == TRUE) {
            preferredPath = SSM_Distribution::mdi_getIndexForAddress(&preferredPrimary);
            if (preferredPath != -1)
                SSM_Pathmanagement::pm_setPrimaryPath(preferredPath);
        }

        state = COOKIE_ECHOED;

        if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);
        /* start cookie timer */
        localData->initTimer = SSM_Adaptation::adl_startTimer(localData->initTimerDuration,
                                               &sci_timer_expired, TIMER_TYPE_INIT,
                                               (void *) &localData->associationID, NULL);
        break;

    case COOKIE_ECHOED:
        /* Duplicated initAck, ignore */
        event_log(EXTERNAL_EVENT, "event: duplicatied sctlr_initAck in state COOKIE_ECHOED");
        break;
    case CLOSED:
    case ESTABLISHED:
    case SHUTDOWNPENDING:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNSENT:
        /* In this states the initAck is unexpected event. */
        event_logi(EXTERNAL_EVENT, "discarding event: sctlr_initAck in state %02d", state);
        break;
    default:
        /* error logging: unknown event */
        event_logi(EXTERNAL_EVENT, "sctlr_initAck: unknown state %02d", state);
        break;
    }

    localData->association_state = state;
    localData = NULL;
    return return_state;
}


/**
  sctlr_cookie_echo is called by bundling when a cookie echo chunk was received from  the peer.
  The following data is retrieved from the cookie and saved for this association:
    \begin{itemize}
    \item  from the init chunk:
        \begin{itemize}
        \item peers tag
        \item peers receiver window credit
        \item peers initial TSN
        \item peers network address list if multihoming is used
        \end{itemize}
    \item local tag generated before the initAck was sent
    \item my initial TSN generated before the initAck was sent
    \item number of send streams I use, must be lower or equal to peers number of receive streams from init chunk
    \item number of receive streams I use (can be lower than number of send streams the peer requested in
     the init chunk
    \end{itemiz}
   @param  cookie_echo pointer to the received cookie echo chunk
 */
void SSM_SCTP_Control::sctlr_cookie_echo(SCTP_cookie_echo * cookie_echo)
{
    union sockunion destAddress;
    union sockunion dAddresses[MAX_NUM_ADDRESSES];
    int ndAddresses, result;
    guint32 state, new_state = 0xFFFFFFFF;
    unsigned int cookieLifetime;
    unsigned int mySupportedTypes;
    ChunkID cookieCID;
    ChunkID cookieAckCID;
    ChunkID initCID;
    ChunkID initAckCID;
    ChunkID shutdownAckCID;
    ChunkID errorCID;
    guint32 cookie_local_tag, cookie_remote_tag;
    guint32 cookie_local_tietag, cookie_remote_tietag;
    guint32 local_tag, remote_tag;
    short primaryDestinationAddress;
    short noOfDestinationAddresses;
    gboolean peerSupportsPRSCTP;

    unsigned short noSuccess, restart_result;
    unsigned int peerAddressTypes;

    int SendCommUpNotification = -1;

    event_log(INTERNAL_EVENT_0, "sctlr_cookie_echo() is being executed");

    cookieCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) cookie_echo);

    if (SSM_ChunkHandler::ch_chunkType(cookieCID) != CHUNK_COOKIE_ECHO) {
        /* error logging */
       SSM_ChunkHandler:: ch_forgetChunk(cookieCID);
        error_log(ERROR_MAJOR, "sctlr_cookie_echo: wrong chunk type");
        return;
    }
    /* section 5.2.4. 1) and 2.) */
    if (SSM_ChunkHandler::ch_goodCookie(cookieCID)) {
        SSM_ChunkHandler::ch_forgetChunk(cookieCID);
        event_log(EXTERNAL_EVENT, "event: invalidCookie received");
        return;
    }
    initCID    = SSM_ChunkHandler::ch_cookieInitFixed(cookieCID);
    initAckCID = SSM_ChunkHandler::ch_cookieInitAckFixed(cookieCID);

    cookie_remote_tag = SSM_ChunkHandler::ch_initiateTag(initCID);
    cookie_local_tag  = SSM_ChunkHandler::ch_initiateTag(initAckCID);

    /* these two will be zero, if association is not up yet */
    local_tag  = SSM_Distribution::mdi_readLocalTag();
    remote_tag = SSM_Distribution::mdi_readTagRemote();

    if ((SSM_Distribution::mdi_readLastInitiateTag()   != cookie_local_tag) &&
        (SSM_Distribution::mdi_readLastFromPort()      != SSM_ChunkHandler::ch_CookieSrcPort(cookieCID)) &&
        (SSM_Distribution::mdi_readLastDestPort()      != SSM_ChunkHandler::ch_CookieDestPort(cookieCID)))  {

        SSM_ChunkHandler::ch_forgetChunk(cookieCID);
        SSM_ChunkHandler::ch_deleteChunk(initCID);
        SSM_ChunkHandler::ch_deleteChunk(initAckCID);
        event_log(EXTERNAL_EVENT, "event: good cookie echo received, but with incorrect verification tag");

        /* for test purposes */
        exit(-1);
        /* for test purposes */

        return;
    }

    /* section 5.2.4. 3.) */
    if ((cookieLifetime = SSM_ChunkHandler::ch_staleCookie(cookieCID)) > 0) {
        event_logi(EXTERNAL_EVENT, "event: staleCookie received, lifetime = %d", cookieLifetime);

        if ((cookie_local_tag != local_tag) || (cookie_remote_tag != remote_tag)) {

            SSM_Distribution::mdi_writeLastInitiateTag(cookie_remote_tag);
            /* make and send stale cookie error */
            errorCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_ERROR, FLAG_NONE);
            SSM_ChunkHandler::ch_enterStaleCookieError(errorCID, (unsigned int) (1.2 * cookieLifetime));
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(errorCID),NULL);
            SSM_Bundling::bu_sendAllChunks(NULL);
            SSM_ChunkHandler::ch_forgetChunk(cookieCID);
            SSM_ChunkHandler::ch_deleteChunk(initCID);
            SSM_ChunkHandler::ch_deleteChunk(initAckCID);
            SSM_ChunkHandler::ch_deleteChunk(errorCID);
            return;
        }                       /* ELSE : Case 5.2.4.E. Valid Cookie, unpack into a TCB */
    }


    result = SSM_Distribution::mdi_readLastFromAddress(&destAddress);
    if (result != 0) {
       error_log(ERROR_MAJOR, "sctlr_cookie_echo: mdi_readLastFromAddress failed !");
       SSM_ChunkHandler::ch_deleteChunk(initCID);
       SSM_ChunkHandler::ch_deleteChunk(initAckCID);
       SSM_ChunkHandler::ch_forgetChunk(cookieCID);
       return;
    }

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        noOfDestinationAddresses = 1;
        primaryDestinationAddress = 0;

        /* why is INSTANCE_NAME here zero ? */
        noSuccess = SSM_Distribution::mdi_newAssociation(NULL, SSM_Distribution::mdi_readLastDestPort(),
                                        SSM_Distribution::mdi_readLastFromPort(),
                                        cookie_local_tag, /* this is MY tag */
                                        primaryDestinationAddress,
                                        noOfDestinationAddresses, &destAddress);

        if (noSuccess) {
            /* new association could not be entered in the list of associations */
            error_log(ERROR_MAJOR, "sctlr_cookie_echo: Creation of association failed");
            SSM_ChunkHandler::ch_deleteChunk(initCID);
            SSM_ChunkHandler::ch_deleteChunk(initAckCID);
            SSM_ChunkHandler::ch_forgetChunk(cookieCID);
            return;
        }
    }

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        error_log(ERROR_MAJOR, "sctlr_cookie-echo: program error: SCTP-control NULL");
        SSM_ChunkHandler::ch_deleteChunk(initCID);
        SSM_ChunkHandler::ch_deleteChunk(initAckCID);
        SSM_ChunkHandler::ch_forgetChunk(cookieCID);
        return;
    }

    state = localData->association_state;

    event_logiii(VERBOSE, "State : %u, cookie_remote_tag : %x , cookie_local_tag : %x ",
                            state, cookie_remote_tag, cookie_local_tag);
    event_logii(VERBOSE, "remote_tag ; %x , local_tag : %x ", remote_tag, local_tag);

    switch (state) {
    case CLOSED:
        /*----------------- Normal association setup -----------------------------------------*/
        event_log(EXTERNAL_EVENT, "event: sctlr_cookie_echo in state CLOSED");
        mySupportedTypes =SSM_Distribution:: mdi_getSupportedAddressTypes();
        /* retrieve destination addresses from cookie */
        ndAddresses = SSM_ChunkHandler::ch_cookieIPDestAddresses(cookieCID, mySupportedTypes, dAddresses,&peerAddressTypes, &destAddress);

        if (ndAddresses > 0) {
            /* save addresses if initAck contained more then zero, otherwise the source address
               of the IP-message carrying the cookie-chunk will be used for this association. */
            event_logi(VERBOSE, "Storing %d destination addresses as paths", ndAddresses);
            SSM_Distribution::mdi_writeDestinationAddresses(dAddresses, ndAddresses);
        }

        peerSupportsPRSCTP = SSM_ChunkHandler::ch_getPRSCTPfromCookie(cookieCID);

        /* initialize new association from cookie data */
        SSM_Distribution::mdi_initAssociation(SSM_ChunkHandler::ch_receiverWindow(initCID),
                            SSM_ChunkHandler::ch_noInStreams(initAckCID),
                            SSM_ChunkHandler::ch_noOutStreams(initAckCID),
                            SSM_ChunkHandler::ch_initialTSN(initCID), cookie_remote_tag, SSM_ChunkHandler::ch_initialTSN(initAckCID),
                            peerSupportsPRSCTP, FALSE);


        localData->NumberOfOutStreams = SSM_ChunkHandler::ch_noOutStreams(initAckCID);
        localData->NumberOfInStreams = SSM_ChunkHandler::ch_noInStreams(initAckCID);
        event_logii(VERBOSE, "Set Outbound Stream to %u, Inbound Streams to %u",
            localData->NumberOfOutStreams, localData->NumberOfInStreams);


        /* make cookie acknowledgement */
        cookieAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_COOKIE_ACK, FLAG_NONE);

        /* send cookie acknowledgement */
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(cookieAckCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_Bundling::bu_unlock_sender(NULL);
        SSM_ChunkHandler::ch_deleteChunk(cookieAckCID);

        /* notification to ULP */
        SendCommUpNotification = SCTP_COMM_UP_RECEIVED_VALID_COOKIE;

        /* mdi_communicationUpNotif(SCTP_COMM_UP_RECEIVED_VALID_COOKIE); */

        new_state = ESTABLISHED;
        break;

        /* For the rest of these (pathological) cases, refer to section 5.2.4 as to what to do */

    case COOKIE_WAIT:
    case COOKIE_ECHOED:
    case ESTABLISHED:
    case SHUTDOWNPENDING:
    case SHUTDOWNSENT:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNACKSENT:
        cookie_local_tietag  = SSM_ChunkHandler::ch_CookieLocalTieTag(cookieCID);
        cookie_remote_tietag = SSM_ChunkHandler::ch_CookiePeerTieTag(cookieCID);

        event_logii(VERBOSE, "cookie_remote_tietag ; %x , cookie_local_tietag : %x ",
                    cookie_remote_tietag, cookie_local_tietag);
        /* cookie_local_tag, cookie_remote_tag are set */
        /* local_tag, remote_tag are also set from the TCB */

        if (cookie_local_tag == local_tag) {        /* cases B or D */
            if (cookie_remote_tag == remote_tag) {  /* case D */
                /*  the endpoint should always enter the ESTABLISHED state, if it has not
                    already done so. It should stop any init or cookie timers that may be
                    running and send a COOKIE ACK */
                event_log(VERBOSE, "Dupl. CookieEcho, case 5.2.4.D)");
                /* stop COOKIE timers */
                if (localData->initTimer != 0) {
                    sctp_stopTimer(localData->initTimer);
                    localData->initTimer = 0;
                }
                /* go to ESTABLISHED state */
                new_state = ESTABLISHED;
                if (state == COOKIE_WAIT || state==COOKIE_ECHOED) {
                    mySupportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();
                    ndAddresses = SSM_ChunkHandler::ch_cookieIPDestAddresses(cookieCID, mySupportedTypes, dAddresses,&peerAddressTypes, &destAddress);
                    if (ndAddresses > 0) {
                        /* save addresses if initAck contained more then zero, otherwise the source address
                           of the IP-message carrying the cookie-chunk will be used for this association. */
                        event_logi(VERBOSE, "Storing %d destination addresses as paths", ndAddresses);
                        SSM_Distribution::mdi_writeDestinationAddresses(dAddresses, ndAddresses);
                    }
                    peerSupportsPRSCTP = SSM_ChunkHandler::ch_getPRSCTPfromCookie(cookieCID);

                    /* initialize new association from cookie data */
                    SSM_Distribution::mdi_initAssociation(SSM_ChunkHandler::ch_receiverWindow(initCID),
                                        SSM_ChunkHandler::ch_noInStreams(initAckCID),
                                        SSM_ChunkHandler::ch_noOutStreams(initAckCID),
                                        SSM_ChunkHandler::ch_initialTSN(initCID),
                                        cookie_remote_tag,
                                        SSM_ChunkHandler::ch_initialTSN(initAckCID),
                                        peerSupportsPRSCTP, FALSE);

                    localData->NumberOfOutStreams = SSM_ChunkHandler::ch_noOutStreams(initAckCID);
                    localData->NumberOfInStreams = SSM_ChunkHandler::ch_noInStreams(initAckCID);
                    event_logii(VERBOSE, "Set Outbound Stream to %u, Inbound Streams to %u",
                    localData->NumberOfOutStreams, localData->NumberOfInStreams);

                    /* notification to ULP */
                    SendCommUpNotification = SCTP_COMM_UP_RECEIVED_VALID_COOKIE;
                }
                /* make cookie acknowledgement */
                cookieAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_COOKIE_ACK, FLAG_NONE);
                /* send cookie acknowledgement */
                SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(cookieAckCID),NULL);
                SSM_Bundling::bu_sendAllChunks(NULL);
                SSM_Bundling::bu_unlock_sender(NULL);
                SSM_ChunkHandler::ch_deleteChunk(cookieAckCID);
            } else {                                /* case B */
                /*  The endpoint should stay in or enter
                    the ESTABLISHED state but it MUST update its peer's Verification
                    Tag from the State Cookie, stop any init or cookie timers that may
                    running and send a COOKIE ACK. */
                event_log(VERBOSE, "Dupl. CookieEcho, case 5.2.4.B)");
                /* stop COOKIE timers */
                if (localData->initTimer != 0) {
                    sctp_stopTimer(localData->initTimer);
                    localData->initTimer = 0;
                }
                new_state = ESTABLISHED;

                if (state == COOKIE_WAIT || state==COOKIE_ECHOED) {
                    mySupportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();
                    ndAddresses = SSM_ChunkHandler::ch_cookieIPDestAddresses(cookieCID, mySupportedTypes, dAddresses,&peerAddressTypes, &destAddress);
                    if (ndAddresses > 0) {
                        /* save addresses if initAck contained more then zero, otherwise the source address
                           of the IP-message carrying the cookie-chunk will be used for this association. */
                        event_logi(VERBOSE, "Storing %d destination addresses as paths", ndAddresses);
                        SSM_Distribution::mdi_writeDestinationAddresses(dAddresses, ndAddresses);
                    }
                    peerSupportsPRSCTP = SSM_ChunkHandler::ch_getPRSCTPfromCookie(cookieCID);

                    /* initialize new association from cookie data */
                    SSM_Distribution::mdi_initAssociation(SSM_ChunkHandler::ch_receiverWindow(initCID),
                                        SSM_ChunkHandler::ch_noInStreams(initAckCID),
                                        SSM_ChunkHandler::ch_noOutStreams(initAckCID),
                                        SSM_ChunkHandler::ch_initialTSN(initCID),
                                        cookie_remote_tag,
                                        SSM_ChunkHandler::ch_initialTSN(initAckCID),
                                        peerSupportsPRSCTP, FALSE);

                    localData->NumberOfOutStreams = SSM_ChunkHandler::ch_noOutStreams(initAckCID);
                    localData->NumberOfInStreams = SSM_ChunkHandler::ch_noInStreams(initAckCID);
                    event_logii(VERBOSE, "Set Outbound Stream to %u, Inbound Streams to %u",
                    localData->NumberOfOutStreams, localData->NumberOfInStreams);

                    /* notification to ULP */
                    SendCommUpNotification = SCTP_COMM_UP_RECEIVED_VALID_COOKIE;
                }
                SSM_Distribution::mdi_rewriteTagRemote(cookie_remote_tag);
                SSM_Distribution::mdi_rewriteLocalTag(cookie_local_tag);
                /* make cookie acknowledgement */
                cookieAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_COOKIE_ACK, FLAG_NONE);
                /* send cookie acknowledgement */
                SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(cookieAckCID),NULL);
                SSM_Bundling::bu_sendAllChunks(NULL);
                SSM_Bundling::bu_unlock_sender(NULL);
                SSM_ChunkHandler::ch_deleteChunk(cookieAckCID);
            }
        } else {                                    /* cases A or C */
            if ((cookie_remote_tag      == remote_tag)  &&
                (cookie_local_tietag    == 0)           &&
                (cookie_remote_tietag   == 0)) {  /* is case C */
                    /* section 5.2.4. action C : silently discard cookie */
                    event_log(VERBOSE, "Dupl. CookieEcho, case 5.2.4.C) --> Silently discard !");
                    SSM_ChunkHandler::ch_forgetChunk(cookieCID);
                    SSM_ChunkHandler::ch_deleteChunk(initCID);
                    SSM_ChunkHandler::ch_deleteChunk(initAckCID);
                    localData = NULL;
                    return;         /* process data as usual ? */
            }  else if ((cookie_remote_tag != remote_tag) &&
                        (cookie_local_tietag == localData->local_tie_tag) &&
                        (cookie_remote_tietag == localData->peer_tie_tag)) {     /* case A */
                /* section 5.2.4. action A : Possible Peer Restart  */
                if (state != SHUTDOWNACKSENT) {
                    event_logi(VERBOSE, "Peer Restart, case 5.2.4.A, state == %u", state);

                    mySupportedTypes = SSM_Distribution::mdi_getSupportedAddressTypes();
                    ndAddresses = SSM_ChunkHandler::ch_cookieIPDestAddresses(cookieCID, mySupportedTypes, dAddresses, &peerAddressTypes, &destAddress);
                    peerSupportsPRSCTP = SSM_ChunkHandler::ch_getPRSCTPfromCookie(cookieCID);

                    restart_result = SSM_Distribution::mdi_restartAssociation(SSM_ChunkHandler::ch_noInStreams(initAckCID),
                                                            SSM_ChunkHandler::ch_noOutStreams(initAckCID),
                                                            SSM_ChunkHandler::ch_receiverWindow(initCID),
                                                            SSM_ChunkHandler::ch_initialTSN(initCID),
                                                            SSM_ChunkHandler::ch_initialTSN(initAckCID),
                                                            (short)ndAddresses, 0, dAddresses,
                                                            peerSupportsPRSCTP,FALSE); /* setting 0 as primary */
                    if (restart_result == 0) {
                        /* what happens to SCTP data chunks is implementation specific */
						SSM_Distribution::mdi_rewriteTagRemote(cookie_remote_tag);
					    SSM_Distribution::mdi_rewriteLocalTag(cookie_local_tag);
                        /* go to ESTABLISHED state */
                        new_state = ESTABLISHED;
                        /* make cookie acknowledgement */
                        cookieAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_COOKIE_ACK, FLAG_NONE);
                        /* send cookie acknowledgement */
                        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(cookieAckCID),NULL);
                        SSM_Bundling::bu_sendAllChunks(NULL);
                        SSM_Bundling::bu_unlock_sender(NULL);
                        SSM_ChunkHandler::ch_deleteChunk(cookieAckCID);

                        SendCommUpNotification = SCTP_COMM_UP_RECEIVED_COOKIE_RESTART;
                        /* mdi_restartNotif(); */

                    } else {  /* silently discard */
                        event_log(VERBOSE, "Restart not successful, silently discarding CookieEcho");
                        SSM_ChunkHandler::ch_forgetChunk(cookieCID);
                        SSM_ChunkHandler::ch_deleteChunk(initCID);
                        SSM_ChunkHandler::ch_deleteChunk(initAckCID);
                        localData = NULL;
                        return;             /* process data as usual ? */
                    }
                } else {
                    event_log(VERBOSE, "Peer Restart case, state == SHUTDOWN_ACK_SENT");
                    /* resend SHUTDOWN_ACK */
                    shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
                    /* add ERROR_CHUNK with Error Cause : "Cookie Received while shutting down" */
                    SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),NULL);
                    errorCID = SSM_ChunkHandler::ch_makeErrorChunk();
                    SSM_ChunkHandler:: ch_enterErrorCauseData(errorCID, ECC_COOKIE_RECEIVED_DURING_SHUTDWN, 0, NULL);
                    SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(errorCID),NULL);
                    /* send cookie acknowledgement */
                    SSM_Bundling::bu_sendAllChunks(NULL);
                    SSM_Bundling::bu_unlock_sender(NULL);
                    SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);
                    SSM_ChunkHandler::ch_deleteChunk(errorCID);
                }

            } else { /* silently discard */
                event_log(VERBOSE, "Dupl. CookieEcho, silently discarding CookieEcho");
                SSM_ChunkHandler::ch_forgetChunk(cookieCID);
                SSM_ChunkHandler::ch_deleteChunk(initCID);
                SSM_ChunkHandler::ch_deleteChunk(initAckCID);
                localData = NULL;
                return;             /* process data as usual ? */
            }
        }
        break;
    default:
        /* error logging: unknown event */
        error_logi(EXTERNAL_EVENT_X, "sctlr_cookie_echo : unknown state %02u", state);
        break;
    }

    SSM_ChunkHandler::ch_deleteChunk(initCID);
    SSM_ChunkHandler::ch_deleteChunk(initAckCID);
    SSM_ChunkHandler::ch_forgetChunk(cookieCID);

    if (new_state != 0xFFFFFFFF)
        localData->association_state = new_state;
    localData = NULL;

    if (SendCommUpNotification != -1) {
        if (SendCommUpNotification == SCTP_COMM_UP_RECEIVED_COOKIE_RESTART) SSM_Distribution::mdi_restartNotif();
        else if (SendCommUpNotification == SCTP_COMM_UP_RECEIVED_VALID_COOKIE)
             SSM_Distribution::mdi_communicationUpNotif(SCTP_COMM_UP_RECEIVED_VALID_COOKIE);
    }

}



/**
  sctlr_cookieAck is called by bundling when a cookieAck chunk was received from  the peer.
  The only purpose is to inform the active side that peer has received the cookie chunk.
  The association is in established state after this function is called.
  Communication up is signalled to the upper layer in this case.
  @param cookieAck pointer to the received cookie ack chunk
*/
void SSM_SCTP_Control::sctlr_cookieAck(SCTP_simple_chunk * cookieAck)
{
    guint32 state;
    ChunkID cookieAckCID;
    int SendCommUpNotif = -1;

    cookieAckCID = SSM_ChunkHandler::ch_makeChunk(cookieAck);

    if (SSM_ChunkHandler::ch_chunkType(cookieAckCID) != CHUNK_COOKIE_ACK) {
        /* error logging */
        error_log(ERROR_MAJOR, "sctlr_cookieAck: wrong chunk type");
        return;
    }
    SSM_ChunkHandler::ch_forgetChunk(cookieAckCID);


    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        error_log(ERROR_MAJOR, "sctlr_cookieAck: read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case COOKIE_ECHOED:

        event_logi(EXTERNAL_EVENT_X, "event: sctlr_cookieAck in state %02d", state);
        /* stop init timer */
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        }
        /* free  cookieChunk */
        free(localData->initChunk);
        free(localData->cookieChunk);
        localData->initChunk = NULL;
        localData->cookieChunk = NULL;
        SendCommUpNotif = SCTP_COMM_UP_RECEIVED_COOKIE_ACK;
        /* mdi_communicationUpNotif(SCTP_COMM_UP_RECEIVED_COOKIE_ACK); */

        state = ESTABLISHED;
        break;

    case ESTABLISHED:
        /* Duplicated cookie, ignore */
        break;
    case CLOSED:
    case COOKIE_WAIT:
    case SHUTDOWNPENDING:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNSENT:
        /* In this states the cookie is unexpected event.
           Do error logging  */
        event_logi(EXTERNAL_EVENT_X, "unexpected event: sctlr_cookieAck in state %02d", state);
        break;
    default:
        /* error logging: unknown event */
        break;
    }

    localData->association_state = state;
    localData = NULL;
    if (SendCommUpNotif == SCTP_COMM_UP_RECEIVED_COOKIE_ACK)
        SSM_Distribution::mdi_communicationUpNotif(SCTP_COMM_UP_RECEIVED_COOKIE_ACK);

}


/**
  sctlr_shutdown is called by bundling when a shutdown chunk was received from the peer.
  This function initiates a graceful shutdown of the association.
  @param  shutdown_chunk pointer to the received shutdown chunk
*/
int SSM_SCTP_Control::sctlr_shutdown(SCTP_simple_chunk * shutdown_chunk)
{
    guint32 state, new_state;
    boolean readyForShutdown;
    gboolean removed = FALSE, sendNotification = FALSE;
    unsigned int lastFromPath;
    int return_state = STATE_OK;
    ChunkID abortCID;
    ChunkID shutdownAckCID;
    ChunkID shutdownCID;

    shutdownCID = SSM_ChunkHandler::ch_makeChunk(shutdown_chunk);

    if (SSM_ChunkHandler::ch_chunkType(shutdownCID) != CHUNK_SHUTDOWN) {
        /* error logging */
        error_log(ERROR_MAJOR, "sctlr_cookieAck: wrong chunk type");
        SSM_ChunkHandler::ch_forgetChunk(shutdownCID);
        return return_state;
    }

    if ((localData = (SCTP_controlData *)SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sctlr_shutdown: read SCTP-control failed");
        SSM_ChunkHandler::ch_forgetChunk(shutdownCID);
        return return_state;
    }

    state = localData->association_state;
    new_state = state;

    lastFromPath = SSM_Distribution::mdi_readLastFromPath();

    switch (state) {
    case CLOSED:
        event_log(EXTERNAL_EVENT, "event: sctlr_shutdown in state CLOSED, send ABORT ! ");
        abortCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NO_TCB);
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(abortCID),&lastFromPath);
        SSM_Bundling::bu_sendAllChunks(&lastFromPath);
		SSM_Bundling::bu_unlock_sender(&lastFromPath);
        SSM_ChunkHandler::ch_deleteChunk(abortCID);
        /* delete all data of this association */
        SSM_Distribution::mdi_deleteCurrentAssociation();
        removed = TRUE;
        return_state = STATE_STOP_PARSING_REMOVED;
        break;

    case COOKIE_WAIT:
    case COOKIE_ECHOED:
    case SHUTDOWNPENDING:
        event_logi(EXTERNAL_EVENT, "event: sctlr_shutdown in state %2u -> discarding !", state);
        SSM_ChunkHandler::ch_forgetChunk(shutdownCID);
        break;

    case SHUTDOWNRECEIVED:
    case SHUTDOWNACKSENT:
        event_log(EXTERNAL_EVENT, "sctlr_shutdown in state SHUTDOWN_RECEIVED/SHUTDOWN_ACK_SENT -> acking CTSNA !");
        SSM_Reltransfer::rtx_rcv_shutdown_ctsna(SSM_ChunkHandler::ch_cummulativeTSNacked(shutdownCID));
        break;

    case ESTABLISHED:
        event_log(EXTERNAL_EVENT, "event: sctlr_shutdown in state ESTABLISHED");

        new_state = SHUTDOWNRECEIVED;

        SSM_Reltransfer::rtx_rcv_shutdown_ctsna(SSM_ChunkHandler::ch_cummulativeTSNacked(shutdownCID));

        readyForShutdown = (SSM_Reltransfer::rtx_readNumberOfUnackedChunks() == 0) &&
                           (SSM_Flowcontrol::fc_readNumberOfQueuedChunks() == 0);

        sendNotification = TRUE;

        if (readyForShutdown) {
            /* retransmissions are not necessary */
            /* send shutdownAck */
            event_log(VERBOSE, "We are ready for SHUTDOWN, sending SHUTDOWN_ACK !");
            shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),&lastFromPath);
            SSM_Bundling::bu_sendAllChunks(&lastFromPath);
            SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);
            if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);

            localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);
            new_state = SHUTDOWNACKSENT;
        } else {
            /* retrieve cummunalative TSN acked from shutdown chunk */
            SSM_Reltransfer::rtx_shutdown();
            /* retransmissions are necessary */
            /* call reliable transfer and wait for sci_allChunksAcked */
        }
        break;

    case SHUTDOWNSENT:
        /* check wether reliable transfer is ready for shutdown */
        readyForShutdown = (SSM_Reltransfer::rtx_readNumberOfUnackedChunks() == 0) &&
            (SSM_Flowcontrol::fc_readNumberOfQueuedChunks() == 0);

        sendNotification = TRUE;

        if (readyForShutdown) {
            /* retransmissions are not necessary */
            /* send shutdownAck */
            shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
            SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),&lastFromPath);
            SSM_Bundling::bu_sendAllChunks(&lastFromPath);
            SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);
            if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);
			localData->initTimer =
                SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);

            new_state = SHUTDOWNACKSENT;
        } else {
            error_log(ERROR_MAJOR, "Error in Program Logic !!!");
            error_log(ERROR_MAJOR,
                      "SHUTDOWN_SENT state may not be entered, if queues are not empty !!!!");

        }
        break;

    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "sctlr_shutdown in state %02d: unexpected event", state);
        break;
    }
    SSM_ChunkHandler::ch_forgetChunk(shutdownCID);
    if (sendNotification) {
       SSM_Distribution:: mdi_peerShutdownReceivedNotif();
    }

    localData->association_state = new_state;
    localData = NULL;
    if (removed == TRUE) {
        SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_NO_TCB);
        SSM_Distribution::mdi_clearAssociationData();
    }
    return return_state;

}



/**
  sctlr_shutdownAck is called by bundling when a shutdownAck chunk was received from the peer.
  Depending on the current state of the association, COMMUNICATION LOST is signaled to the
  Upper Layer Protocol, and the association marked for removal.
*/
int SSM_SCTP_Control::sctlr_shutdownAck()
{
    guint32 state, new_state;
    unsigned int lastFromPath, lastTag;
    boolean tagWasZero = FALSE;
    ChunkID shdcCID;
    int return_state = STATE_OK;
    int removed = 0;    /* i.e. meaning FALSE here ! */

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sctlr_shutdownAck: read SCTP-control failed");
        return return_state;
    }

    lastFromPath = SSM_Distribution::mdi_readLastFromPath();
    state = localData->association_state;
    new_state = state;

    switch (state) {
    case CLOSED:
        error_log(ERROR_FATAL,
                  "sctlr_shutdownAck in state CLOSED, should have been handled before ! ");
        break;
    case COOKIE_WAIT:
    case COOKIE_ECHOED:
        /* see also section 8.5.E.) treat this like OOTB packet, leave T1 timer run ! */
        event_logi(EXTERNAL_EVENT,
                   "event: sctlr_shutdownAck in state %u, send SHUTDOWN_COMPLETE ! ", state);
        shdcCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_COMPLETE, FLAG_NO_TCB);

        /* make sure the shutdown_complete is written to the peer with his tag */
        if (SSM_Distribution::mdi_readTagRemote() == 0) {
            tagWasZero = TRUE;
            lastTag = SSM_Distribution::mdi_readLastInitiateTag();
            SSM_Distribution::mdi_rewriteTagRemote(lastTag);
        }

        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shdcCID),&lastFromPath);
        SSM_Bundling::bu_sendAllChunks(&lastFromPath);
		SSM_Bundling::bu_unlock_sender(&lastFromPath);
        SSM_ChunkHandler::ch_deleteChunk(shdcCID);
        return_state = STATE_OK;

        if (tagWasZero == TRUE) {
            SSM_Distribution::mdi_rewriteTagRemote(0);
        }
        break;
    case ESTABLISHED:
        error_log(ERROR_FATAL,
                  "sctlr_shutdownAck in state ESTABLISHED, peer not standard conform ! ");
        break;
    case SHUTDOWNPENDING:
        error_log(ERROR_FATAL,
                  "sctlr_shutdownAck in state SHUTDOWNPENDING, peer not standard conform ! ");
        break;
    case SHUTDOWNRECEIVED:
        error_log(ERROR_FATAL,
                  "sctlr_shutdownAck in state SHUTDOWNRECEIVED, peer not standard conform ! ");
        break;

    case SHUTDOWNSENT:
    case SHUTDOWNACKSENT:

        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        } else {
            error_log(ERROR_FATAL, "Timer not running - Error in Program Logic");
        }

        shdcCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_COMPLETE, FLAG_NONE);
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shdcCID),&lastFromPath);

        SSM_Bundling::bu_sendAllChunks(&lastFromPath);
        SSM_ChunkHandler::ch_deleteChunk(shdcCID);

        SSM_Bundling::bu_unlock_sender(&lastFromPath);
        /* delete all data of this association */
        return_state = STATE_STOP_PARSING_REMOVED;
        SSM_Distribution::mdi_deleteCurrentAssociation();

        removed = SCTP_SHUTDOWN_COMPLETE;   /* i.e. meaning SHUTDOWN_COMPLETE here */

        new_state = CLOSED;
        break;

    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "sctlr_shutdownAck in state %02d: unexpected event", state);
        break;
    }

    localData->association_state = new_state;
    localData = NULL;
    if (removed != 0) {
        if (removed == SCTP_SHUTDOWN_COMPLETE){
            SSM_Distribution::mdi_shutdownCompleteNotif();
        } else {
            SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_NO_TCB);
        }
        SSM_Distribution::mdi_clearAssociationData();
    }
    return return_state;

}

/**
  sctlr_shutdownComplete is called by bundling when a SHUTDOWN COMPLETE chunk was received from the peer.
  COMMUNICATION LOST is signaled to the ULP, timers stopped, and the association is marked for removal.
*/
int SSM_SCTP_Control::sctlr_shutdownComplete()
{
    guint32 state, new_state;
    unsigned int lastFromPath;
    int return_state = STATE_OK;

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sctlr_shutdownComplete: read SCTP-control failed");
        return return_state;
    }

    lastFromPath = SSM_Distribution::mdi_readLastFromPath();

    state = localData->association_state;
    new_state = state;

    switch (state) {
    case CLOSED:
    case COOKIE_WAIT:
    case COOKIE_ECHOED:
    case ESTABLISHED:
    case SHUTDOWNPENDING:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNSENT:
        error_logi(EXTERNAL_EVENT, "sctlr_shutdownComplete in state %u -> discarding ! ", state);
        break;

    case SHUTDOWNACKSENT:
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        } else {
            error_log(ERROR_FATAL,
                      "sctlr_shutdownComplete : Timer not running - problem in Program Logic!");
        }
        SSM_Pathmanagement::pm_disableAllHB();

        SSM_Bundling::bu_unlock_sender(&lastFromPath);
        /* delete all data of this association */
        SSM_Distribution::mdi_deleteCurrentAssociation();

        localData->association_state = CLOSED;

        SSM_Distribution::mdi_shutdownCompleteNotif();
        SSM_Distribution::mdi_clearAssociationData();

        localData = NULL;

        return_state =STATE_STOP_PARSING_REMOVED;

        return return_state;
        break;

    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "sctlr_shutdownComplete in state %02d: unexpected event", state);
        break;
    }
    localData->association_state = new_state;
    localData = NULL;
    return return_state;
}

/**
  sctlr_abort is called by bundling when an ABORT chunk was received from  the peer.
  COMMUNICATION LOST is signalled to the ULP, timers are stopped, and the association
  is marked for removal.
 */
int SSM_SCTP_Control::sctlr_abort()
{
    guint32 state;
    unsigned int lastFromPath;
    int return_state = STATE_OK;

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sctlr_abort: read SCTP-control failed");
        return return_state;
    }

    lastFromPath = SSM_Distribution::mdi_readLastFromPath();

    state = localData->association_state;

    switch (state) {
    case CLOSED:
        event_log(EXTERNAL_EVENT, "event: sctlr_abort in state CLOSED -> discard chunk");
        /* discard chunk */
        break;
    case COOKIE_WAIT:
    case COOKIE_ECHOED:
    case SHUTDOWNSENT:
        event_logi(EXTERNAL_EVENT, "event: sctlr_abort in state %2d", state);

        /* stop possible timer */
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        }
        /* delete all data of this association */

        SSM_Bundling::bu_unlock_sender(&lastFromPath);

        return_state = STATE_STOP_PARSING_REMOVED;
        SSM_Distribution::mdi_deleteCurrentAssociation();

        SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_ABORTED);
        SSM_Distribution::mdi_clearAssociationData();

        break;
    case ESTABLISHED:
    case SHUTDOWNPENDING:
    case SHUTDOWNRECEIVED:
    case SHUTDOWNACKSENT:
        event_logi(EXTERNAL_EVENT, "event: sctlr_abort in state %02d", state);
        /* delete all data of this association */
        return_state = STATE_STOP_PARSING_REMOVED;

        /* stop init timer, just in case */
        if (localData->initTimer != 0) {
            sctp_stopTimer(localData->initTimer);
            localData->initTimer = 0;
        }

        SSM_Bundling::bu_unlock_sender(&lastFromPath);

        SSM_Distribution::mdi_deleteCurrentAssociation();

        SSM_Distribution::mdi_communicationLostNotif(SCTP_COMM_LOST_ABORTED);
        SSM_Distribution::mdi_clearAssociationData();

        break;
    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "sctlr_abort in state %02d: unexpected event", state);
        break;
    }
    localData = NULL;
    return return_state;
}


/**
   sctlr_staleCookie is called by bundling when a 'stale cookie' error chunk was received.
   @param error_chunk pointer to the received error chunk
*/
void SSM_SCTP_Control::sctlr_staleCookie(SCTP_simple_chunk * error_chunk)
{
    guint32 state;
    ChunkID errorCID;
    ChunkID initCID;

    errorCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) error_chunk);

    if (SSM_ChunkHandler::ch_chunkType(errorCID) != CHUNK_ERROR) {
        /* error logging */
        SSM_ChunkHandler::ch_forgetChunk(errorCID);
        error_log(ERROR_MAJOR, "sctlr_staleCookie: wrong chunk type");
        return;
    }

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sctlr_staleCookie: read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case COOKIE_ECHOED:

        /* make chunkHandler init chunk from stored init chunk string */
        initCID = SSM_ChunkHandler::ch_makeChunk((SCTP_simple_chunk *) localData->initChunk);

        /* read staleness from error chunk and enter it into the cookie preserv. */
        SSM_ChunkHandler::ch_enterCookiePreservative(initCID, SSM_ChunkHandler::ch_stalenessOfCookieError(errorCID));

        /* resend init */
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(initCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_ChunkHandler::ch_forgetChunk(initCID);

        state = COOKIE_WAIT;
        break;

    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "sctlr_staleCookie in state %02d: unexpected event", state);
        break;
    }
    localData->association_state = state;
    localData = NULL;
}

/**
   sci_getState is called by distribution to get the state of the current SCTP-control instance.
   This function also logs the state with log-level VVERBOSE.
   @return state value (0=CLOSED, 3=ESTABLISHED)
*/
guint32 SSM_SCTP_Control::sci_getState()
{
    guint32 state;

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_getState: read SCTP-control failed");
        return CLOSED;
    }

    state = localData->association_state;

    switch (state) {
    case CLOSED:
        event_log(VVERBOSE, "Current state : CLOSED");
        break;
    case COOKIE_WAIT:
        event_log(VVERBOSE, "Current state :COOKIE_WAIT ");
        break;
    case COOKIE_ECHOED:
        event_log(VVERBOSE, "Current state : COOKIE_ECHOED");
        break;
    case ESTABLISHED:
        event_log(VVERBOSE, "Current state : ESTABLISHED");
        break;
    case SHUTDOWNPENDING:
        event_log(VVERBOSE, "Current state : SHUTDOWNPENDING");
        break;
    case SHUTDOWNRECEIVED:
        event_log(VVERBOSE, "Current state : SHUTDOWNRECEIVED");
        break;
    case SHUTDOWNSENT:
        event_log(VVERBOSE, "Current state : SHUTDOWNSENT");
        break;
    case SHUTDOWNACKSENT:
        event_log(VVERBOSE, "Current state : SHUTDOWNACKSENT");
        break;
    default:
        event_log(VVERBOSE, "Unknown state : return closed");
        return CLOSED;
        break;
    }

    return state;
    localData = NULL;
}


/*------------------- Functions called by reliable transfer --------------------------------------*/

/**
  Called by reliable transfer if all (sent !) chunks in its retransmission queue have been acked.
  This function is used to move from state SHUTDOWNPENDING to  SHUTDOWNSENT (after having sent a
  shutdown chunk) or to move from  SHUTDOWNRECEIVED to SHUTDOWNACKSENT (after having sent a
  shutdown-ack chunk)
*/
void SSM_SCTP_Control::sci_allChunksAcked()
{
    guint32 state;
    ChunkID shutdownCID;
    ChunkID shutdownAckCID;
    SCTP_controlData* old_data = localData;

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_allChunksAcked: read SCTP-control failed");
        return;
    }

    state = localData->association_state;

    switch (state) {
    case SHUTDOWNPENDING:

        event_log(EXTERNAL_EVENT, "event: sci_allChunksAcked in state SHUTDOWNPENDING");

        /* make and send shutdown */
        shutdownCID = SSM_ChunkHandler::ch_makeShutdown(SSM_Recvctrl::rxc_read_cummulativeTSNacked());
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_ChunkHandler::ch_deleteChunk(shutdownCID);

        /* start shutdown timer */
        localData->initTimerDuration = SSM_Pathmanagement::pm_readRTO(SSM_Pathmanagement::pm_readPrimaryPath());

        if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);

        localData->initTimer = SSM_Adaptation::adl_startTimer(localData->initTimerDuration,
                                               &sci_timer_expired,TIMER_TYPE_SHUTDOWN,
                                               (void *) &localData->associationID, NULL);

        localData->initRetransCounter = 0;

        /* receive control must acknowledge every datachunk at once after the shutdown
           was sent. */
        SSM_Recvctrl::rxc_send_sack_everytime();

        state = SHUTDOWNSENT;

        break;

    case SHUTDOWNRECEIVED:

        event_log(EXTERNAL_EVENT, "event: sci_allChunksAcked in state SHUTDOWNRECEIVED");

        /* send shutdownAck */
        shutdownAckCID = SSM_ChunkHandler::ch_makeSimpleChunk(CHUNK_SHUTDOWN_ACK, FLAG_NONE);
        SSM_Bundling::bu_put_Ctrl_Chunk(SSM_ChunkHandler::ch_chunkString(shutdownAckCID),NULL);
        SSM_Bundling::bu_sendAllChunks(NULL);
        SSM_ChunkHandler::ch_deleteChunk(shutdownAckCID);

        /* ADDED : should probably be OK */
        if (localData->initTimer != 0) sctp_stopTimer(localData->initTimer);

        localData->initTimer =  SSM_Adaptation::adl_startTimer(localData->initTimerDuration, &sci_timer_expired,TIMER_TYPE_SHUTDOWN,
                                (void *) &localData->associationID, NULL);

        state = SHUTDOWNACKSENT;
        break;

    default:
        /* error logging */
        event_logi(EXTERNAL_EVENT_X, "unexpected event: sci_allChunksAcked in state %d", state);
        break;
    }

    localData->association_state = state;
    localData = old_data;
}



/*------------------- Functions called message by distribution to create and delete --------------*/

/**
    newSCTP_control allocates data for a new SCTP-Control instance
 */
void * SSM_SCTP_Control::sci_newSCTP_control(void* sctpInstance)
{
    SCTP_controlData*  tmp=NULL;

    event_logi(INTERNAL_EVENT_0, "Create SCTP-control for Instance %x", sctpInstance);

    tmp = (SCTP_controlData *) malloc(sizeof(SCTP_controlData));

    if (tmp == NULL) {
        error_log(ERROR_MAJOR," Malloc failed in sci_newSCTP_control()");
        return NULL;
    }

    tmp->association_state = CLOSED;
    tmp->initTimer = 0;
    tmp->initTimerDuration = RTO_INITIAL;
    tmp->initRetransCounter = 0;
    tmp->initChunk = NULL;
    tmp->cookieChunk = NULL;
    tmp->associationID = SSM_Distribution::mdi_readAssociationID();
    tmp->NumberOfOutStreams = SSM_Distribution::mdi_readLocalOutStreams();
    tmp->NumberOfInStreams = SSM_Distribution::mdi_readLocalInStreams();
    tmp->local_tie_tag = 0;
    tmp->peer_tie_tag = 0;

    tmp->assocMaxRetransmissions = SSM_Distribution::mdi_getDefaultAssocMaxRetransmits(sctpInstance);
    tmp->assocMaxInitRetransmissions = SSM_Distribution::mdi_getDefaultMaxInitRetransmits(sctpInstance);
    tmp->cookieLifeTime = SSM_Distribution::mdi_getDefaultValidCookieLife(sctpInstance);
    tmp->instance = sctpInstance;

    event_log(INTERNAL_EVENT_0, "event: created SCTP-control Instance");

    return (void *) tmp;
}



/**
  deleteSCTP_control frees memory allocated for a SCTP-Control instance
*/
void SSM_SCTP_Control::sci_deleteSCTP_control(void *sctpControlData)
{
    SCTP_controlData *sctpCD;

    event_log(INTERNAL_EVENT_0, "deleting SCTP-control");
    sctpCD = (SCTP_controlData *) sctpControlData;
    if (sctpCD->initTimer != 0) {
        sctp_stopTimer(sctpCD->initTimer);
    }
    if (sctpCD->initChunk != NULL)
        free(sctpCD->initChunk);
    if (sctpCD->cookieChunk != NULL)
        free(sctpCD->cookieChunk);
    free(sctpControlData);
}

/**
 * get current parameter value for assocMaxRetransmissions
 * @return current value, -1 on error
 */
int SSM_SCTP_Control::sci_getMaxAssocRetransmissions(void)
{
    SCTP_controlData* old_data = localData;
    int max;

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_getMaxAssocRetransmissions(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    max =   localData->assocMaxRetransmissions;
    localData = old_data;
    return max;
}

/**
 * get current parameter value for assocMaxInitRetransmissions
 * @return current value, -1 on error
 */
int SSM_SCTP_Control::sci_getMaxInitRetransmissions(void)
{
    SCTP_controlData* old_data = localData;
    int max;
    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_getMaxInitRetransmissions(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    max =   localData->assocMaxInitRetransmissions;
    localData = old_data;
    return max;
}

/**
 * get current parameter value for cookieLifeTime
 * @return current value, -1 on error
 */
int SSM_SCTP_Control:: sci_getCookieLifeTime(void)
{
    int max;
    SCTP_controlData* old_data = localData;

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MINOR, "sci_getCookieLifeTime(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    max =   localData->cookieLifeTime;
    localData = old_data;
    return max;
}

/**
 * set new parameter value for assocMaxRetransmissions
 * @param new_max  new parameter value for assocMaxRetransmissions
 * @return 0 for success, -1 on error
 */
int SSM_SCTP_Control::sci_setMaxAssocRetransmissions(int new_max)
{
    SCTP_controlData* old_data = localData;
    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_setMaxAssocRetransmissions(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    localData->assocMaxRetransmissions = new_max;
    localData = old_data;
    return 0;
}

/**
 * set new parameter value for assocMaxRetransmissions
 * @param new_max  new parameter value for assocMaxRetransmissions
 * @return 0 for success, -1 on error
 */
int SSM_SCTP_Control::sci_setMaxInitRetransmissions(int new_max)
{
    SCTP_controlData* old_data = localData;

    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_setMaxInitRetransmissions(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    localData->assocMaxInitRetransmissions = new_max;
    localData = old_data;
    return 0;
}

/**
 * set new parameter value for cookieLifeTime
 * @param new_max  new parameter value for cookieLifeTime
 * @return 0 for success, -1 on error
 */
int SSM_SCTP_Control::sci_setCookieLifeTime(int new_max)
{
    SCTP_controlData* old_data = localData;

    if ((localData = (SCTP_controlData *) SSM_Distribution::mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_setCookieLifeTime(): read SCTP-control failed");
        localData = old_data;
        return -1;
    }
    localData->cookieLifeTime= new_max;
    localData = old_data;
    return 0;
}


gboolean SSM_SCTP_Control::sci_shutdown_procedure_started()
{
    SCTP_controlData* old_data = localData;

    guint32 state;
    if ((localData = (SCTP_controlData *)SSM_Distribution:: mdi_readSCTP_control()) == NULL) {
        /* error log */
        error_log(ERROR_MAJOR, "sci_readState : read SCTP-control failed");
        localData = old_data;
        return FALSE;
    }
    state = localData->association_state;
    localData = old_data;

    if (state == SHUTDOWNPENDING || state == SHUTDOWNRECEIVED
        || state == SHUTDOWNSENT || state == SHUTDOWNACKSENT) return TRUE;
    else
        return FALSE;
}

/*@}*/
