// SSM_Streamengine.h: interface for the SSM_Streamengine class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SSM_STREAMENGINE_H__A429AD43_8B66_4914_A278_0DA86111997D__INCLUDED_)
#define AFX_SSM_STREAMENGINE_H__A429AD43_8B66_4914_A278_0DA86111997D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000



#include  "globals.h"           /* boolean, etc */
#include  "messages.h"



/**
* return values for se_ulpreceivefrom
*/
#define  RECEIVE_DATA           0 /* received data ok in streamengine */
#define  STREAM_ID_OVERFLOW     1 /* wrong stream Id from ulp */
#define  NO_DATA_AVAILABLE      2 /* no datagram in streamengine receive list */


class SSM_Streamengine  
{
public:
	SSM_Streamengine();
	virtual ~SSM_Streamengine();


	
	/******************** Function Definitions *****************************************/


	/* This function is called to instanciate one Stream Engine for an association.
	   It is called by Message Distribution.
	   called from MessageDisribution 
	   returns: the pointer to the Stream Engine 
	*/
	static void* se_new_stream_engine (unsigned int numberReceiveStreams,        /* max of streams to receive */
								unsigned int numberSendStreams,           /* max of streams to send */
								gboolean assocSupportsPRSCTP);

	/* 
	 * se_delete_stream_engine: Deletes a stream engine instance
	 * 
	 * Params: Pointer/handle which was returned by se_new()
	 * Return value: error value
	 */
	static void se_delete_stream_engine(void *instancePtr);



	/* returns the number of in- and out-streams */
	static int se_readNumberOfStreams(unsigned short *inStreams, unsigned short *outStreams);



	/**
	 * This function is called to send a chunk.
	 *  called from MessageDistribution
	 * @return 0 for success, -1 for error (e.g. data sent in shutdown state etc.)
	*/
	static int se_ulpsend(unsigned short streamId,
				   unsigned char *buffer,
				   unsigned int byteCount, unsigned int protocolId,
				   short destAddressIndex, void* context, unsigned int lifetime,
				   gboolean unorderedDelivery,  /* optional (=FALSE if none) */
				   gboolean dontBundle);         /* optional (=null if none)  */






	/* after all chunks in a SCTP pdu have been given to the
	   Stream Engine module, we start reassembly and notifications */
	static int se_doNotifications(void);


	/* This function is called from ULP to receive a chunk.
	*/
	static short se_ulpreceivefrom(unsigned char *buffer, unsigned int *byteCount,
							unsigned short streamId, unsigned short* streamSN,
							unsigned int * tsn, unsigned int* addressIndex, unsigned int flags);


	/*
	 * This function is called from RX_Control to receive a chunk.
	 */
	static int se_recvDataChunk(SCTP_data_chunk * dataChunk, unsigned int byteCount, unsigned int address_index);


	/**
	 * function to return the number of chunks that can be retrieved
	 * by the ULP - this function may need to be refined !!!!!!
	 */
	static guint32 se_numOfQueuedChunks(void);

	/**
	 * function to return the number of streams that we may
	 * send on
	 */
	static guint16 se_numOfSendStreams(void);

	/**
	 * function to return the number of streams that we are allowed to
	 * receive data on
	 */
	static guint16 se_numOfRecvStreams(void);


	static int se_deliver_unreliably(unsigned int up_to_tsn, SCTP_forward_tsn_chunk* chk);


	static int se_getQueuedBytes(void);



};

#endif // !defined(AFX_SSM_STREAMENGINE_H__A429AD43_8B66_4914_A278_0DA86111997D__INCLUDED_)
