// SSM_Timer_list.h: interface for the SSM_Timer_list class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SSM_TIMER_LIST_H__E9EE7CC6_DD29_4304_B10D_B2D06FE6092C__INCLUDED_)
#define AFX_SSM_TIMER_LIST_H__E9EE7CC6_DD29_4304_B10D_B2D06FE6092C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000




#ifdef WIN32
#include <winsock2.h>
#endif

#if (defined (SOLARIS) || defined (WIN32))
#define timeradd(a, b, result) \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec + (b)->tv_sec;                             \
    (result)->tv_usec = (a)->tv_usec + (b)->tv_usec;                          \
    if ((result)->tv_usec >= 1000000)                                         \
      {                                                                       \
        ++(result)->tv_sec;                                                   \
        (result)->tv_usec -= 1000000;                                         \
      }                                                                       \
  } while (0)

#define timersub(a, b, result)                                                \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                             \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                          \
    if ((result)->tv_usec < 0) {                                              \
      --(result)->tv_sec;                                                     \
      (result)->tv_usec += 1000000;                                           \
    }                                                                         \
  } while (0)

#endif

#include "globals.h"



typedef struct alarm_timer
{
    unsigned int timer_id;
    int timer_type;
/* the time when it is to go off */
    struct timeval action_time;
/* pointer to possible arguments */
    void *arg1;
    void *arg2;
/* the callback function 	*/
    void (*action) (TimerID, void *, void *);
/* arranged in a sorted, linked list	*/
}
AlarmTimer;
/**
 *	function to initialize a list. Creates a timer_list structure
 *	@param	new_list 	pointer to newly alloc'ed list structure
 *	@return	0 success, -1 for out of memory.
 */

class SSM_Timer_list  
{
public:
	SSM_Timer_list();
	virtual ~SSM_Timer_list();



	static void init_timer_list(void);

	/**
	 *	function to delete a list. Walks through the list and deallocates
	 *	all timer_item structs. Finally destroys the timer_list struct
	 *	@param	del_list	pointer to the timer_list struct to be deleted
	 *	@return	0 on success, -1 if pointer was NULL or other error
	 */
	static void del_timer_list(void);


	/**
	 *	this function inserts a timer_item into the list. Keeps it ordered,
	 *	and updates length, and possibly one timeval entry. Checks whether
	 *	we insert at beginning/end first. timer_item must have been alloc'ed
	 *	first by the application, this is not done by this function !
	 *	@param	tlist		pointer to the timer_list instance
	 *	@param	item	pointer to the event item that is to be added
	 *	@return	timer id  on success, 0 if a pointer was NULL or other error
	 */
	static unsigned int insert_item(AlarmTimer * item);

	/**
	 *	a function to remove a certain action item, first checks current_item,
	 *	then traverses the list from the start, updates length etc.
	 *	@param	tlist		pointer to the timer_list instance
	 *	@param	timer_id	id of the timer to be removed
	 *	@param	item	pointer to where deleted data is to be copied !
	 *	@return	0 on success, -1 if a pointer was NULL or other error, 1 if not found
	 */
	static int remove_item(unsigned int id);

	/**
	 * same function, but remove timer by its pointer to the AlarmTimer
	 * data structure. This is easier on the list
	 */
	static int remove_timer(AlarmTimer* item);

	/**
	 *      function to be called, when a timer is reset. Basically calls get_item(),
	 *    saves the function pointer, updates the execution time (msecs milliseconds
	 *      from now) and removes the item from the list. Then calls insert_item
	 *      with the updated timer_item struct.
	 *      @param  tlist           pointer to the timer_list instance
	 *      @param  timer_id        id of the timer to be updated
	 *      @param  msecs           action to be executed msecs ms from _now_
	 *      @return timer id
	 */
	static unsigned int update_item(unsigned int id, unsigned int msecs);
	static unsigned int micro_update_item(unsigned int id, unsigned int seconds, unsigned int microseconds);

	/*
	 * function prototype from function in adaptation.h/.c
	 * @return milliseconds up to the expiry of the next timer
	 */
	static int get_msecs_to_nexttimer(void);


	static void adl_add_msecs_totime(struct timeval *t, unsigned int msecs);

	static void print_debug_list(short event_log_level);

	/**
	 * @return 1 (true) if list empty, 0 if list not empty
	 */
	static int timer_list_empty(void);

	/**
	 * copies first event to where the pointer dest points to
	 */
	static int get_next_event(AlarmTimer ** dest);


};

#endif // !defined(AFX_SSM_TIMER_LIST_H__E9EE7CC6_DD29_4304_B10D_B2D06FE6092C__INCLUDED_)
