//---------------------------------------------------------------------------
#ifndef _AUDIOCODEC_H_
#define _AUDIOCODEC_H_
//---------------------------------------------------------------------------


/////////////////////////////////////////////////////////////////////////////
// CCODEC window
#define  False 0
#define  True  1

/* Coder global constants */
#define  Frame       240
#define  LpcFrame    180
#define  SubFrames   4
#define  SubFrLen    (Frame/SubFrames)

/* LPC constants */
#define  LpcOrder          10
#define  RidgeFact         10
#define  CosineTableSize   512
#define  PreCoef           ((float) -0.25)

#define  LspPred0          ((float)12.0/(float)32.0)
#define  LspPred1          ((float)23.0/(float)32.0)

#define  LspQntBands       3
#define  LspCbSize         256
#define  LspCbBits         8

/* LTP constants */
#define  PitchMin          18
#define  PitchMax          (PitchMin+127)
#define  PwRange           3
#define  ClPitchOrd        5
#define  Pstep             1
#define NbFilt085          85
#define NbFilt170          170

/* MP-MLQ constants */
#define  Sgrid             2
#define  MaxPulseNum       6
#define  MlqSteps          2

/* acelp constants */

#define SubFrLen2          (SubFrLen +4)
#define DIM_RR             416
#define NB_POS             8
#define STEP               8
#define MSIZE              64
#define threshold          ((float) 0.5)
#define max_time           120

/* Gain constant */
#define  NumOfGainLev      24

/* FER constant */
#define  ErrMaxNum         3

/* CNG constants  */
#define NbAvAcf            3  /* Nb of frames for Acf average               */
#define NbAvGain           3  /* Nb of frames for gain average              */
#define ThreshGain         3  /* Theshold for quantized gains               */
#define FracThreshP1       ((float) 1.2136)
#define MaxLev             22 /* Max. gain index                            */
#define NbPulsBlk          11 /* Nb of pulses in 2-subframes blocks         */

#define InvNbPulsBlk       ((float)1.0/(float)NbPulsBlk)
#define NbFilt             50 /* size of LT filters table                   */
#define LpcOrderP1         (LpcOrder+1)
#define SizAcf             ((NbAvAcf+1)*LpcOrderP1) /* size of array Acf    */
#define SubFrLenD          (2*SubFrLen)
#define Gexc_Max           ((float)5000.0) /* Max gain for fixed excitation */

/* Taming constants */
#define NbFilt085_min       51
#define NbFilt170_min       93
#define SizErr              5
#define Err0                ((float) 0.00000381464)
#define ThreshErr           ((float) 128.0)

#define MAXV                ((float) 256.0)


typedef  long  int   Word32;
typedef  short int   Word16;
typedef  short int   Flag;

/*   Used structures */

typedef  struct   {
/* High pass variables */
   float    HpfZdl;
   float    HpfPdl;
/* Lsp previos vector */
   float   PrevLsp[LpcOrder];

/* All pitch operation buffers */
   float    PrevWgt[PitchMax];
   float    PrevErr[PitchMax];
   float    PrevExc[PitchMax];

/* Requered memory for the delay */
   float   PrevDat[LpcFrame-SubFrLen];

/* Used delay lines */
   float    WghtFirDl[LpcOrder];
   float    WghtIirDl[LpcOrder];
   float    RingFirDl[LpcOrder];
   float    RingIirDl[LpcOrder];

/* For taming procedure */

   Word16   SinDet;
   float    Err[SizErr];

   } CODSTATDEF;

typedef  struct   {
   int     Ecount;
   float   InterGain;
   Word16  InterIndx;
   Word16  Rseed;
   float   Park;
   float   Gain;
/* Lsp previous vector */
   float   PrevLsp[LpcOrder];

/* All pitch operation buffers */
   float   PrevExc[PitchMax];

/* Used delay lines */
   float   SyntIirDl[LpcOrder];
   float   PostFirDl[LpcOrder];
   float   PostIirDl[LpcOrder];

   } DECSTATDEF;

/* subframe coded parameters */
typedef  struct   {
   int      AcLg;
   int      AcGn;
   int      Mamp;
   int      Grid;
   int      Tran;
   int      Pamp;
   Word32   Ppos;
   } SFSDEF;

/* frame coded parameters */
typedef  struct   {
   Word16   Crc;
   Word32   LspId;
   int      Olp[SubFrames/2];
   SFSDEF   Sfs[SubFrames];
   } LINEDEF;

/* harmonic noise shaping filter parameters */
typedef  struct   {
   int   Indx;
   float Gain;
   } PWDEF;

/* pitch postfilter parameters */
typedef  struct   {
   int     Indx;
   float   Gain;
   float   ScGn;
   } PFDEF;

/* best excitation vector parameters for the high rate */
typedef  struct {
   float    MaxErr;
   int      GridId;
   int      MampId;
   int      UseTrn;
   int      Ploc[MaxPulseNum];
   float    Pamp[MaxPulseNum];
   } BESTDEF;

    /* VAD static variables */
typedef struct {
    Word16  Hcnt;
    Word16  Vcnt;
    float   Penr;
    float   Nlev;
    Word16  Aen;
    Word16  Polp[4];
    float   NLpc[LpcOrder];
} VADSTATDEF;


/* CNG features */

/* Coder part */
typedef struct {
    float   CurGain;
    Word16  PastFtyp;
    float   Acf[SizAcf];
    float   LspSid[LpcOrder];
    float   SidLpc[LpcOrder];
    float   RC[LpcOrderP1];
    float   Ener[NbAvGain];
    Word16  NbEner;
    Word16  IRef;
    float   SidGain;
    Word16  RandSeed;
} CODCNGDEF;

/* Decoder part */
typedef struct {
    float   CurGain;
    Word16  PastFtyp;
    float   LspSid[LpcOrder];
    float   SidGain;
    Word16  RandSeed;
} DECCNGDEF;

class  CCODEC
{

public :

	CCODEC();//CWnd *parent);
	virtual ~CCODEC();


// Operations
private:  // code_cng
        void    Init_Cod_Cng(void);
        void    Cod_Cng(float *DataExc, Word16 *Ftyp, LINEDEF *Line, float *QntLpc);
        void    Update_Acf(float *Acfsf);

	//coder
        void    Init_Coder(void);
        Flag    Coder(float *DataBuff, char *Vout);

	//deccng
        void     Init_Dec_Cng(void);
        void     Dec_Cng(Word16 Ftyp, LINEDEF *Line, float *DataExc, float *QntLpc);

	//decod2
        void    Init_Decod(void);
        Flag    Decod(float *DataBuff, char *Vinp, Word16 Crc);

	//exc2
        int   Estim_Pitch(float *Dpnt, int Start);
        PWDEF Comp_Pw(float *Dpnt, int Start, int Olp);
        void  Filt_Pw(float *DataBuff, float *Dpnt, int Start, PWDEF Pw);
        void  Find_Fcbk(float *Dpnt, float *ImpResp, LINEDEF *Line, int Sfc);
        void  Find_Best(BESTDEF *Best, float *Tv, float *ImpResp, int Np, int Olp);
        void  Fcbk_Pack(float *Dpnt, SFSDEF *Sfs, BESTDEF *Best, int Np);
        void  Fcbk_Unpk(float *Tv, SFSDEF Sfs, int Olp, int Sfc);
        int   ACELP_LBC_code(float *X, float *h, int T0, float *code,
                     int *ind_gain, int *shift, int *sign, float gain_T0);
        void  Cor_h(float *H, float *rr);
        void  Cor_h_X(float h[],float X[],float D[]);
        int   D4i64_LBC(float Dn[], float rr[], float h[], float cod[],
                float y[], int *code_shift, int *sign);
        int   G_code(float X[], float Y[], float *gain_q);
        void  Gen_Trn(float *Dst, float *Src, int Olp);

        void  Find_Acbk(float *Tv, float *ImpResp, float *PrevExc, LINEDEF
                *Line, int Sfc);
        void  Get_Rez(float *Tv, float *PrevExc, int Lag);
        void  Decod_Acbk(float *Tv, float *PrevExc, int Olp, int Lid, int Gid);
        Word16 Comp_Info(float *Buff, int Olp, float *Gain);
        void  Regen(float *DataBuff, float *Buff, Word16 Lag, float Gain,
            int Ecount, Word16 *Sd);
        PFDEF Comp_Lpf(float *Buff, int Olp, int Sfc);
        int   Find_B(float *Buff, int Olp, int Sfc);
        int   Find_F(float *Buff, int Olp, int Sfc);
        PFDEF Get_Ind(int Ind, float Ten, float Ccr, float Enr);
        void  Filt_Lpf(float *Tv, float *Buff, PFDEF Pf, int Sfc);
        int   search_T0(int T0, int Gid, float *gain_T0);
        void  reset_max_time(void);

        void Update_Err(int Olp, int AcLg, int AcGn);
        int  Test_Err(int Lag1, int Lag2);

	// lpc2
        void   Comp_Lpc(float *UnqLpc, float *PrevDat, float *DataBuff);
        float  Durbin(float *Lpc, float *Corr, float Err, float *Pk2);
        void   Wght_Lpc(float *PerLpc, float *UnqLpc);
        void   Error_Wght(float *Dpnt, float *PerLpc);
        void   Comp_Ir(float *ImpResp, float *QntLpc, float *PerLpc, PWDEF Pw);
        void   Sub_Ring(float *Dpnt, float *QntLpc, float *PerLpc, float *PrevErr, PWDEF Pw);
        void   Upd_Ring(float *Dpnt, float *QntLpc, float *PerLpc, float *PrevErr);
        void   Synt(float *Dpnt, float *Lpc);
        float  Spf(float *Tv, float *Lpc);

	//lsp2
        float  Polynomial(float *Lpq, int CosPtr);
        void   AtoLsp(float *LspVect, float *Lpc, float *PrevLsp);
        Word32 Lsp_Qnt(float *CurrLsp, float *PrevLsp);
        Word32 Lsp_Svq(float *Lsp, float *Wvect);
        void   Lsp_Inq(float *Lsp, float *PrevLsp, Word32 LspId, Word16 Crc);
        void   Lsp_Int(float *QntLpc, float *CurrLsp, float *PrevLsp);
        void   LsptoA(float *Lsp);
	// util 2
        void     Rem_Dc(float *Dpnt);
        void     Mem_Shift(float *PrevDat, float *DataBuff);
        Word16   Rand_lbc(Word16 *p);
        void     Line_Pack(LINEDEF *Line, char *Vout, Word16 Ftyp);
        LINEDEF  Line_Unpk(char *Vinp, Word16 *Ftyp, Word16 Crc);
        Word16*  Par2Ser(Word32 Inp, Word16 *Pnt, int BitNum);
        Word32   Ser2Par(Word16 **Pnt, int Count);
        void     Scale(float *Tv,float Sen);
        float    DotProd(float *in1, float *in2, int len);

	//util cng
        void    Calc_Exc_Rand(float curGain, float *PrevExc, float *DataExc,
                                      Word16 *nRandom, LINEDEF *Line);
        Word16  Qua_SidGain(float *Ener, Word16 nq);
        float   Dec_SidGain(Word16 iGain);

	// vad2
        void    Init_Vad(void);
        Flag    Comp_Vad(float *Dpnt);

	void ComputePastAvFilter(float *Coeff);
	void CalcRC(float *Coeff, float *RC);
	Flag LpcDiff(float *RC, float *ptrAcf, float alpha);
	Word16 random_number(Word16 np1, Word16 *nRandom);

        float   *xDpnt;

        int   ReinitSize;

        /* Coder rate */
        int  WrkRate;


public:

	void Init_Encoding();
	void Init_Decoding();

	void Mem_Encoding(Word16 *Raw_Data, char *Encod_Data);
	void Mem_Decoding(Word16 *Raw_Data, char *Encod_Data);


private:
        HWND hWnd;

};

//=========================================//
extern CCODEC *m_coder;
extern CCODEC *m_decoder;
//=========================================//

#endif