//---------------------------------------------------------------------------
#include "stdafx.h"
#include "AudioControl.h"
#include "SCTPClientDlg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CAudioIO *AudioIO;
void CALLBACK waveInProc( HWAVEIN hwi, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 );

//---------------------------------------------------------------------------
CAudioIO::CAudioIO(CWnd *parent)
{
	m_pMain = parent;
	
	AudioIO = this;
	
	bRecoding = false;
	
	//========================================//
	m_coder = new CCODEC();
	m_decoder = new CCODEC();
	//========================================//

	InitialAudio(1);

}

//---------------------------------------------------------------------------
CAudioIO::~CAudioIO()
{
//	bRecoding=false;
//	RecordUnprepare();
//	FreeBuffers();
	//========================================//
	delete m_coder;
	delete m_decoder;
	//========================================//
}

//---------------------------------------------------------------------------
void CAudioIO::AllocBuffers()
{
	int i;
	
	m_line = new  char[LINE_LENG];
	rm_line = new char[LINE_LENG];
	
	for (i = 0; i < MAX_BUFFERS; i++)
	{
		whr[i] = (WAVEHDR *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(WAVEHDR) );
		if (whr[i])
		{
			whr[i]->lpData = (char *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,DATABLOCK_SIZE);
			whr[i]->dwBufferLength = DATABLOCK_SIZE;
		}
		
		whp[i] = (WAVEHDR *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(WAVEHDR) );
		if (whp[i])
		{
			whp[i]->lpData = (char *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,DATABLOCK_SIZE);
			whp[i]->dwBufferLength = DATABLOCK_SIZE;
		}
	}
	
	pRecBuffer = ( short *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,DATABLOCK_SIZE);
	pPlayBuffer	= ( short *) HeapAlloc( GetProcessHeap(),HEAP_ZERO_MEMORY,DATABLOCK_SIZE);
	
	
}

//---------------------------------------------------------------------------
void CAudioIO::InitialAudio(int parm_sample)
{
	WAVEFORMATEX wfx;
	
	AllocBuffers();
	
	//========================================//
	m_coder->Init_Encoding();
	m_decoder->Init_Decoding();
	//========================================//
	
	
	hwo = NULL;
	hwi = NULL;
	
	
	wfx.wFormatTag		= WAVE_FORMAT_PCM;
	wfx.nChannels		= 1;
	wfx.nSamplesPerSec	= 8000 * parm_sample;
	
	wfx.wBitsPerSample  = 16;
	wfx.nBlockAlign     = 2;
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
	wfx.cbSize          = 0;
	
	waveOutOpen(&hwo, WAVE_MAPPER, &wfx, (DWORD)0, 0, CALLBACK_NULL);
	waveInOpen(&hwi, WAVE_MAPPER, &wfx, (DWORD) waveInProc, (DWORD) 0, CALLBACK_FUNCTION);
	
	if (hwi == NULL || hwo == NULL)
		return;
}

//---------------------------------------------------------------------------
void CAudioIO::FreeBuffers()
{
	int i;
	for (i = 0; i < MAX_BUFFERS; i++)
		waveOutUnprepareHeader(hwo, whp[i], sizeof(WAVEHDR));
	waveOutClose(hwo);
	
	for (i = 0; i < MAX_BUFFERS; i++)
	{
		if (whr[i] != NULL)
		{
			HeapFree(GetProcessHeap(), 0, whr[i]->lpData);
			HeapFree(GetProcessHeap(), 0, whr[i]);
			whr[i] = NULL;
		}
		
		if (whp[i] != NULL)
		{
			HeapFree(GetProcessHeap(), 0, whp[i]->lpData);
			HeapFree(GetProcessHeap(), 0, whp[i]);
			whp[i] = NULL;
		}
	}
	HeapFree(GetProcessHeap(), 0, pRecBuffer);
	HeapFree(GetProcessHeap(), 0, pPlayBuffer);
	
	delete[] m_line;
	delete[] rm_line;
	
	waveInClose(hwi);
}

//---------------------------------------------------------------------------
void CAudioIO::StartRecording()
{
	
	if (bRecoding)
		return;
	
	for (int i = 0; i < MAX_BUFFERS; i++)
	{
		waveInPrepareHeader(hwi, whr[i], sizeof(WAVEHDR));
		waveInAddBuffer(hwi, whr[i], sizeof(WAVEHDR));
	}
	
	waveInStart(hwi);
	
	
	bRecoding = true;
	
}

//---------------------------------------------------------------------------
void CAudioIO::StopRecoding()
{
	bRecoding = false;
	RecordUnprepare();
}

//---------------------------------------------------------------------------
void CAudioIO::Recoding(LPARAM lParam)
{
	LPWAVEHDR lpwh = (LPWAVEHDR)lParam;
	
	if (bRecoding)
	{
		memcpy( (( char*)pRecBuffer), lpwh->lpData, DATABLOCK_SIZE);
		
		waveInPrepareHeader(hwi, lpwh, sizeof(WAVEHDR));
		waveInAddBuffer(hwi, lpwh, sizeof(WAVEHDR));
		
		Transmission();
	}
//	else
//	{
//		RecordUnprepare();
//	}
}

//---------------------------------------------------------------------------
void CAudioIO::DecodingPlay()
{
	
	for (int i=0; i < LINE_COUNT; i++)
	{
		m_decoder->Mem_Decoding(pPlayBuffer + i*240 , rm_line+24*i);
	}
	
	Playing();
}

//---------------------------------------------------------------------------
void CAudioIO::Playing()
{
	static int pi = 0;
	
	memcpy(whp[pi]->lpData, ((char*)pPlayBuffer), DATABLOCK_SIZE);
	whp[pi]->dwBufferLength = DATABLOCK_SIZE;
	waveOutPrepareHeader(hwo, whp[pi], sizeof(WAVEHDR));
	waveOutWrite(hwo, whp[pi], sizeof(WAVEHDR));
	pi++;
	if (pi==MAX_BUFFERS) pi = 0;
}

//---------------------------------------------------------------------------
void CAudioIO::RecordUnprepare()
{
	waveInStop(hwi);
	waveInReset(hwi);
	for (int i = 0; i < MAX_BUFFERS; i++)
		waveInUnprepareHeader(hwi, whr[i], sizeof(WAVEHDR));//----Ҵ  freeֱ ݵ ȣؾѴ.
}

//---------------------------------------------------------------------------
void CAudioIO::Transmission()
{
	for ( int i = 0 ; i < LINE_COUNT ; i++)
		m_coder->Mem_Encoding( pRecBuffer + 240*i , m_line+24*i);

//	((CSCTPClientDlg *)Maindlg)->m_audsocket->SendTo(m_line, LINE_LENG, 
//		((CSCTPClientDlg *)Maindlg)->m_audsocket->nPort, ((CEncTestDlg *)m_pMain)->m_audsocket->Ip);
		  

		

}

//---------------------------------------------------------------------------
void CALLBACK waveInProc( HWAVEIN hwi, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 )
{
	switch (uMsg)
	{
	case WIM_DATA:
		AudioIO->Recoding(dwParam1);
		break;
	case WIM_OPEN:
	case WIM_CLOSE:
		break;
	}
}

