/******************************************************************************
	
	  H263Dec.cpp: implementation of the CH263Dec class.

******************************************************************************/

#include "stdafx.h"
#include <vfw.h>
#include "H263Dec.h"
#include "H263Table.h"
#include "Buffer.h"
 
#include <io.h>
#include <stdio.h>
#include <fcntl.h>
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define mmax(a, b)  	((a) > (b) ? (a) : (b))
#define mmin(a, b)  	((a) < (b) ? (a) : (b))
#define mnint(a)        ((a) < 0 ? (int)(a - 0.5) : (int)(a + 0.5))
#define sign(a)         ((a) < 0 ? -1 : 1)

#define START	0
#define READY   1
#define PLAYING 2
#define PAUSED  3
#define FINISH  4
#define STOP    5

#define MAX_BUFF 256

HPALETTE hpal, hPalPrev = NULL;

char *image = NULL;
unsigned short gusState = START;
BITMAPINFO *pbmi = NULL;

int ycbpb_tab[2] = {0, 1};
int uvcbpb_tab[2] = {0, 1};

HDC myDC;

CBuffer buffer;

HANDLE hFullData;
HANDLE hPlay;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CH263Dec::CH263Dec (CWnd *pParent)
{
	zerorun = 0;		
	verbose = 0;		
	outtype = T_X11;	
	myDC = GetDC(pParent->m_hWnd);
	
	nDitherType = 24;

	ld = &base;
	expand	= 0;
	refidct = 0;
	
	outputname = "";
	gusState = STOP;
	init_dib();

    hPlay = CreateEvent(NULL,TRUE,TRUE,"Play Event");
    hFullData = CreateEvent(NULL,FALSE,FALSE,"Full Event");
}
//-----------------------------------------------------------------------------
CH263Dec::~CH263Dec ()
{
	close_dib();
	CloseHandle(hFullData);
	CloseHandle(hPlay);
}
//-----------------------------------------------------------------------------
void CH263Dec::dither(unsigned char *src[])
{
	if (nDitherType==8)
		ditherframe(src);
	else
		ConvertYUVtoRGB(src[0], src[1], src[2], dithered_image, coded_picture_width, coded_picture_height); 

    display_image(dithered_image);  
}
//-----------------------------------------------------------------------------
void CH263Dec::ditherframe(unsigned char *src[])
{
  int i,j;
  int y,u,v;
  unsigned char *py,*pu,*pv,*dst;

  py = src[0];
  pu = src[1];
  pv = src[2];
  
 
#ifdef _WIN32
  dst = dithered_image+(coded_picture_height-1)*coded_picture_width;
#else
  dst = dithered_image;
#endif

  for (j=0; j<coded_picture_height; j+=4)
  {
    /* line j + 0 */
    for (i=0; i<coded_picture_width; i+=4)
    {
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y]|utab[u]|vtab[v]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+8]|utab[u+8]|vtab[v+8]];
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+2]|utab[u+2]|vtab[v+2]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+10]|utab[u+10]|vtab[v+10]];
    }
   
    pu -= chrom_width;
    pv -= chrom_width;
  
#ifdef _WIN32
    dst -= 2*coded_picture_width;
#endif

    /* line j + 1 */
    for (i=0; i<coded_picture_width; i+=4)
    {
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+12]|utab[u+12]|vtab[v+12]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+4]|utab[u+4]|vtab[v+4]];
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+14]|utab[u+14]|vtab[v+14]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+6]|utab[u+6]|vtab[v+6]];
    }

#ifdef _WIN32
    dst -= 2*coded_picture_width;
#endif

    /* line j + 2 */
    for (i=0; i<coded_picture_width; i+=4)
    {
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+3]|utab[u+3]|vtab[v+3]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+11]|utab[u+11]|vtab[v+11]];
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+1]|utab[u+1]|vtab[v+1]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+9]|utab[u+9]|vtab[v+9]];
    }

    pu -= chrom_width;
    pv -= chrom_width;
  
#ifdef _WIN32
    dst -= 2*coded_picture_width;
#endif

    /* line j + 3 */
    for (i=0; i<coded_picture_width; i+=4)
    {
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+15]|utab[u+15]|vtab[v+15]];
      y = *py++;
     *dst++ = pixel[ytab[y+7]|utab[u+7]|vtab[v+7]];
      y = *py++;
      u = *pu++ >> 1;
      v = *pv++ >> 1;
      *dst++ = pixel[ytab[y+13]|utab[u+13]|vtab[v+13]];
      y = *py++;
      
      *dst++ = pixel[ytab[y+5]|utab[u+5]|vtab[v+5]];
    }

#ifdef _WIN32
    dst -= 2*coded_picture_width;
#endif
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::init_dither(int bpp)
{
	int i, v;
		
	bpp/=8;
	if(!(dithered_image = (unsigned char *)malloc(bpp*coded_picture_width*
													coded_picture_height)))
		error("malloc failed");
	
	for (i=-8; i<256+8; i++)
	{
		v = i>>4;
		if (v<1)
			v = 1;
		else if (v>14)
			v = 14;
		ytab[i+8] = v<<4;
	}
	
	for (i=0; i<128+16; i++)
	{
		v = (i-40)>>4;
		if (v<0)
			v = 0;
		else if (v>3)
			v = 3;
		utab[i] = v<<2;
		vtab[i] = v;
	}
	
	for (i=0; i<256; i++)
		pixel[i]=i;
}

//-----------------------------------------------------------------------------
void CH263Dec::initbits()
{
  ld->incnt = 0;
  ld->rdptr = ld->rdbfr + MAX_BUFF;
  ld->bitcnt = 0;
}

//-----------------------------------------------------------------------------
/* return next n bits (right adjusted) without advancing */
unsigned int CH263Dec::showbits( int n)
{
	unsigned char *v;
	unsigned int b;
	int c;
	
	if (ld->incnt<n)
		fillbfr();
	
	v = ld->inbfr + ((96 - ld->incnt)>>3);
	b = (v[0]<<24) | (v[1]<<16) | (v[2]<<8) | v[3];
	c = ((ld->incnt-1) & 7) + 25;
	return (b>>(c-n)) & msk[n];
}

//-----------------------------------------------------------------------------
/* return next bit (could be made faster than getbits(1)) */
unsigned int CH263Dec::getbits1()
{
  return getbits(1);
}

//-----------------------------------------------------------------------------
/* advance by n bits */
void CH263Dec::flushbits(int n)
{
  ld->bitcnt+= n;
  ld->incnt-= n;
  if (ld->incnt < 0)
    fillbfr();
}

//-----------------------------------------------------------------------------
/* return next n bits (right adjusted) */
unsigned int CH263Dec::getbits(int n)
{
  unsigned int l;

  l = showbits(n);
  flushbits(n);

  return l;
}
//-----------------------------------------------------------------------------
void CH263Dec::getblock(int comp, int mode)
{
  int val, i, j, sign;
  unsigned int code;
  VLCtabI *tab;
  short *bp;
  int run, last, level, QP;
  short *qval;

  bp = ld->block[comp];   

  /* decode AC coefficients */
  for (i=(mode==0); ; i++) {
    code = showbits(12);
    if (code>=512)
      tab = (VLCtabI *)&DCT3Dtab0[(code>>5)-16];
    else if (code>=128)
      tab = (VLCtabI *)&DCT3Dtab1[(code>>2)-32];
    else if (code>=8)
      tab = (VLCtabI *)&DCT3Dtab2[(code>>0)-8];
    else {
      if (!quiet)
        fprintf(stderr,"invalid Huffman code in getblock()\n");
      fault = 1;
      return;
    }

    flushbits(tab->len);

    run = (tab->val >> 4) & 255;
    level = tab->val & 15;
    last = (tab->val >> 12) & 1;

   if (tab->val==ESCAPE) { /* escape */
      last = getbits1();
      i += run = getbits(6);
      level = getbits(8);

      if ((sign = (level>=128)))
        val = 256 - level;
      else 
	val = level;
    }
    else {
      i+= run;
      val = level;
      sign = getbits(1);
    }

    if (i >= 64)
    {
      if (!quiet)
        fprintf(stderr,"DCT coeff index (i) out of bounds\n");
      fault = 1;
      return;
    }

    j = zig_zag_scan[i];
    qval = &bp[j];
    if (comp >= 6)
      QP = mmax (1, mmin( 31, ( bquant_tab[bquant] * quant ) >> 2 ));
    else 
      QP = quant;
      
    /* TMN3 dequantization */
    if ((QP % 2) == 1)
      *qval = ( sign ? -(QP * (2* val+1))  : QP * (2* val+1) );
    else
      *qval = ( sign ? -(QP * (2* val+1)-1): QP * (2* val+1)-1 );

    if (last) { /* That's it */
      // if (trace)        printf("last\n");
      return;
    }
  }
}



/*********************************************************************
 *
 * 	Name:		get_sac_block
 *
 *	Description:	Decodes blocks of Arithmetic Encoded DCT Coeffs.
 *			and performs Run Length Decoding and Coefficient       	
 *			Dequantisation.
 *
 *	Input:		Picture block type and number.
 *
 *	Returns:	Nothing.
 *
 *	Side Effects:	
 *
 *********************************************************************/
void CH263Dec::get_sac_block(int comp, int ptype)
{
  int position=0;
  int TCOEF_index, symbol_word;
  int last=0, QP, i, j;
  short *qval, *bp;
  RunCoef DCTcoef;

  bp = ld->block[comp];

  i = (ptype==0);

  while (!last) {	/* while there are DCT coefficients remaining */
    position++;	/* coefficient counter relates to Coeff. model */
    TCOEF_index = DecodeTCoef(position, !ptype);

    if (TCOEF_index == ESCAPE_INDEX) { 	/* ESCAPE code encountered */
      DCTcoef = Decode_Escape_Char(!ptype, &last);
    }
    else {
      symbol_word = tcoeftab[TCOEF_index];

      DCTcoef = vlc_word_decode(symbol_word,&last);
    }

    i += DCTcoef.run;

    j = zig_zag_scan[i];

    qval = &bp[j];

    i++;

    if (comp >= 6)
      QP = mmax (1, mmin( 31, ( bquant_tab[bquant] * quant ) >> 2 ));
    else 
      QP = quant;

    if ((QP % 2) == 1)
      *qval = ( (DCTcoef.sign) ? -(QP * (2* (DCTcoef.val)+1))  : 
		QP * (2* (DCTcoef.val)+1) );
    else
      *qval = ( (DCTcoef.sign) ? -(QP * (2* (DCTcoef.val)+1)-1): 
		QP * (2* (DCTcoef.val)+1)-1 );
	
  }	
  return;
}

/*********************************************************************
 *
 * 	Name:		vlc_word_decode
 *
 *	Description:	Fills Decoder FIFO after a fixed word length
 *			string has been detected.
 *
 *	Input:		Symbol to be decoded, last data flag.
 *
 *	Returns:	Decoded Symbol via the structure DCTcoeff.
 *
 *	Side Effects:	Updates last flag.
 *
 *
 *********************************************************************/

RunCoef CH263Dec::vlc_word_decode(int symbol_word, int *last)
{
  int sign_index;
  RunCoef DCTcoef;

  *last = (symbol_word >> 12) & 01;
 
  DCTcoef.run = (symbol_word >> 4) & 255; 

  DCTcoef.val = (symbol_word) & 15;

  sign_index = decode_a_symbol(cumf_SIGN);	

  DCTcoef.sign = signtab[sign_index];
	
  return (DCTcoef);
} 

/*********************************************************************
 *
 * 	Name:		Decode_Escape_Char
 *
 *	Description:	Decodes all components for a Symbol when an 
 *			ESCAPE character has been detected.
 *
 *	Input:		Picture Type and last data flag.
 *
 *	Returns:	Decoded Symbol via the structure DCTcoeff.
 *
 *	Side Effects:	Modifies last data flag.
 *
 *
 *********************************************************************/

RunCoef CH263Dec::Decode_Escape_Char(int intra, int *last)
{
  int last_index, run, run_index, level, level_index;
  RunCoef DCTcoef;

  if (intra) {
    last_index = decode_a_symbol(cumf_LAST_intra);
     *last = last_intratab[last_index];
  }
  else {
    last_index = decode_a_symbol(cumf_LAST);
    *last = lasttab[last_index];
  }

  if (intra) 
    run_index = decode_a_symbol(cumf_RUN_intra);
  else
    run_index = decode_a_symbol(cumf_RUN);

  run = runtab[run_index];

  /*$if (mrun) run|=64;$*/

  DCTcoef.run = run;

  if (intra)
    level_index = decode_a_symbol(cumf_LEVEL_intra);
  else
    level_index = decode_a_symbol(cumf_LEVEL);

  level = leveltab[level_index];

  if (level >128) 
    level -=256;

  if (level < 0) {
    DCTcoef.sign = 1;
    DCTcoef.val = abs(level);
  }

  else {	
    DCTcoef.sign = 0;
    DCTcoef.val = level;
  }

  return (DCTcoef);
	
}
/*********************************************************************
 *
 * 	Name:		DecodeTCoef
 *
 *	Description:	Decodes a.c DCT Coefficients using the 	
 *			relevant arithmetic decoding model.
 *
 *	Input:		DCT Coeff count and Picture Type.	
 *
 *	Returns:	Index to LUT
 *
 *	Side Effects:	None
 *
 *
 *********************************************************************/

int CH263Dec::DecodeTCoef(int position, int intra)
{
  int index;

  switch (position) {
  case 1:
    {
      if (intra) 
	index = decode_a_symbol(cumf_TCOEF1_intra);
      else 
	index = decode_a_symbol(cumf_TCOEF1); 
      break; 
    }
  case 2:
    {
      if (intra) 
	index = decode_a_symbol(cumf_TCOEF2_intra);
      else
	index = decode_a_symbol(cumf_TCOEF2);
      break; 
    }
  case 3:
    {
      if (intra) 
	index = decode_a_symbol(cumf_TCOEF3_intra);
      else
	index = decode_a_symbol(cumf_TCOEF3);
      break; 
    }
  default: 
    {
      if (intra) 
	index = decode_a_symbol(cumf_TCOEFr_intra);
      else
	index = decode_a_symbol(cumf_TCOEFr);
      break; 
    }
  }

  return (index);
}
//-----------------------------------------------------------------------------
int CH263Dec::getheader()
{
  unsigned int code, gob;

  /* look for startcode */
  startcode();
  code = getbits(PSC_LENGTH);
  gob = getbits(5);
  if (gob == SEC) 
    return 0;
  if (gob == 0) {
    getpicturehdr();
    if (syntax_arith_coding)		/* reset decoder after receiving */
      decoder_reset();	        /* fixed length PSC string */
  }
  return gob + 1;
}

//-----------------------------------------------------------------------------
/* align to start of next startcode */
void CH263Dec::startcode()
{
  /* search for new picture start code */
  while (showbits(PSC_LENGTH)!=1l) 
  	 flushbits(1);
}
//-----------------------------------------------------------------------------
/* decode picture header */
void CH263Dec::getpicturehdr()
{
  int pos, pei, tmp;
  static int prev_temp_ref;

  pos = ld->bitcnt;
  prev_temp_ref = temp_ref;
  temp_ref = getbits(8);
  trd = temp_ref - prev_temp_ref;
  if (trd < 0)
    trd += 256;
  tmp = getbits(1); /* always "1" */
  if (!tmp)
    if (!quiet)
      printf("warning: spare in picture header should be \"1\"\n");
  tmp = getbits(1); /* always "0" */
  if (tmp)
    if (!quiet)
      printf("warning: H.261 distinction bit should be \"0\"\n");
  tmp = getbits(1); /* split_screen_indicator */
  if (tmp) {
    if (!quiet)
      printf("error: split-screen not supported in this version\n");
    exit (-1);
  }
  tmp = getbits(1); /* document_camera_indicator */
  if (tmp)
    if (!quiet)
      printf("warning: document camera indicator not supported in this version\n");

  tmp = getbits(1); /* freeze_picture_release */
  if (tmp)
    if (!quiet)
      printf("warning: frozen picture not supported in this version\n");

  source_format = getbits(3);
  pict_type = getbits(1);
  mv_outside_frame = getbits(1);
  long_vectors = (mv_outside_frame ? 1 : 0);
  syntax_arith_coding = getbits(1);
  adv_pred_mode = getbits(1);
  mv_outside_frame = (adv_pred_mode ? 1 : mv_outside_frame);
  pb_frame = getbits(1);
  quant = getbits(5);
  tmp = getbits(1);
  if (tmp) {
    if (!quiet)
      printf("error: CPM not supported in this version\n");
    exit(-1);
  }
  if (pb_frame) {
    trb = getbits(3);
    bquant = getbits(2);
  }

  pei = getbits(1);
pspare:
  if (pei) {
     /* extra info for possible future backward compatible additions */
    getbits(8);  /* not used */
    pei = getbits(1);
    if (pei) goto pspare; /* keep on reading pspare until pei=0 */
  }


  if (verbose>0) {
    printf("picture header (byte %d)\n",(pos>>3)-4);
    if (verbose>1) {
      printf("  temp_ref=%d\n",temp_ref);
      printf("  pict_type=%d\n",pict_type);
      printf("  source_format=%d\n", source_format);
      printf("  quant=%d\n",quant);
      if (syntax_arith_coding) 
    	printf("  SAC coding mode used \n");
      if (mv_outside_frame)
	printf("  unrestricted motion vector mode used\n");
      if (adv_pred_mode)
	printf("  advanced prediction mode used\n");
      if (pb_frame) {
	printf("  pb-frames mode used\n");
	printf("  trb=%d\n",trb);
	printf("  bquant=%d\n", bquant);
      }
    }
  }
}

//-----------------------------------------------------------------------------
void CH263Dec::getpicture( int *framenum)
{
  int i;
  unsigned char *tmp;

  for (i=0; i<3; i++) {
    tmp				= oldrefframe[i];
    oldrefframe[i]	= refframe[i];
    refframe[i]		= tmp;

    newframe[i]		= refframe[i];
  }

  if (mv_outside_frame && *framenum > 0) {
    make_edge_image(oldrefframe[0],edgeframe[0],coded_picture_width,
		coded_picture_height,32);
    make_edge_image(oldrefframe[1],edgeframe[1],chrom_width, chrom_height,16);
    make_edge_image(oldrefframe[2],edgeframe[2],chrom_width, chrom_height,16);
  }
  getMBs(*framenum);

  if (pb_frame) {
    if (expand && outtype == T_X11) {
      interpolate_image(bframe[0], exnewframe[0],
			coded_picture_width, coded_picture_height);
      interpolate_image(bframe[1], exnewframe[1], chrom_width, chrom_height);
      interpolate_image(bframe[2], exnewframe[2], chrom_width, chrom_height);
      
      storeframe(exnewframe, *framenum);
    }
    else
      storeframe(bframe,*framenum);

  *framenum += pb_frame;

#ifdef USE_TIME
    if (framerate > 0)
      doframerate();
#endif
  }

  if (expand && outtype == T_X11) {
    interpolate_image(newframe[0], exnewframe[0],
		      coded_picture_width, coded_picture_height);
    interpolate_image(newframe[1], exnewframe[1], chrom_width, chrom_height);
    interpolate_image(newframe[2], exnewframe[2], chrom_width, chrom_height);

    storeframe(exnewframe, *framenum);
  }
  else
    storeframe(newframe,*framenum);
}


/* decode all macroblocks of the current picture */
//-----------------------------------------------------------------------------
void CH263Dec::getMBs(int framenum)
{
  int comp;
  int MBA, MBAmax;
  int bx, by;

  int COD=0,MCBPC, CBPY, CBP=0, CBPB=0, MODB=0, Mode=0, DQUANT;
  int COD_index, CBPY_index, MODB_index, DQUANT_index, MCBPC_index;
  int INTRADC_index, YCBPB_index, UVCBPB_index, mvdbx_index, mvdby_index;
  int mvx, mvy, mvy_index, mvx_index, pmv0, pmv1, xpos, ypos, gob, i,k;
  int mvdbx=0, mvdby=0, pmvdbx, pmvdby, gfid, YCBPB, UVCBPB, gobheader_read;
  int startmv,stopmv,offset,bsize,last_done=0,pCBP=0,pCBPB=0,pCOD=0;
  int DQ_tab[4] = {-1,-2,1,2};
  short *bp;

  /* number of macroblocks per picture */
  MBAmax = mb_width*mb_height;

  MBA = 0; /* macroblock address */
  newgob = 0;

  /* mark MV's above the picture */
  for (i = 1; i < m_MBC+1; i++) {
    for (k = 0; k < 5; k++) {
      MV[0][k][0][i] = NO_VEC;
      MV[1][k][0][i] = NO_VEC;
    }
    modemap[0][i] = MODE_INTRA;
  }
  /* zero MV's on the sides of the picture */
  for (i = 0; i < m_MBR+1; i++) {
    for (k = 0; k < 5; k++) {
      MV[0][k][i][0] = 0;
      MV[1][k][i][0] = 0;
      MV[0][k][i][m_MBC+1] = 0;
      MV[1][k][i][m_MBC+1] = 0;
    }
    modemap[i][0] = MODE_INTRA;
    modemap[i][m_MBC+1] = MODE_INTRA;
  }

  fault = 0;
  gobheader_read = 0;
  
  for (;;) {

    // if (trace)      printf("frame %d, MB %d\n",tr_framenum,MBA);

  resync:


    if (fault) {
      printf("Warning: A Fault Condition Has Occurred - Resyncing \n");
      startcode();  /* sync on new startcode */
      fault = 0;
    }

    if (!(showbits(22)>>6)) { /* startcode */

      startcode();  
      /* in case of byte aligned start code, ie. PSTUF, GSTUF or ESTUF
	 is used */
      
		if (showbits(22) == (32|SEC)) { /* end of sequence */
			if (!(syntax_arith_coding && MBA < MBAmax)) {
			return;
			}
		}
		else if ((showbits(22) == PSC<<5) ) { /* new picture */
			if (!(syntax_arith_coding && MBA < MBAmax)) {
				return;
			}
		}
		else {
			if (!(syntax_arith_coding && MBA%m_MBC)) {

				if (syntax_arith_coding) {   /* SAC hack to finish GOBs which   */
					gob = (showbits(22) & 31); /* end with MBs coded with no bits */
					if (gob * mb_width != MBA) 
						goto finish_gob;
				}

				gob = getheader() - 1;
				if (gob > m_MBR) {
					if (!quiet)
						printf("GN out of range\n");
				return;
				}
	  
		/* GFID is not allowed to change unless PTYPE in picture header 
			changes */
				gfid = getbits(2);
		/* NB: in error-prone environments the decoder can use this
	     value to determine whether a picture header where the PTYPE
	     has changed, has been lost */
	  
				quant = getbits(5);
				// if (trace) printf("GQUANT: %d\n", quant);
				xpos = 0;
				ypos = gob;
				MBA = ypos * mb_width;
	  
				newgob = 1;
				gobheader_read = 1;
				if (syntax_arith_coding) 
				decoder_reset();	/* init. arithmetic decoder buffer after gob */
			}
		}
    }

  finish_gob:  /* SAC specific label */

    if (!gobheader_read) {
      xpos = MBA%mb_width;
      ypos = MBA/mb_width;
      if (xpos == 0 && ypos > 0)
	newgob = 0;
    }
    else 
      gobheader_read = 0;

    if (MBA>=MBAmax) 
      return; /* all macroblocks decoded */

  read_cod:
    if (syntax_arith_coding) { 
      if (pict_type == PCT_INTER) {
	COD_index = decode_a_symbol(cumf_COD);
	COD = codtab[COD_index];
      }
      else
	COD = 0;  /* COD not used in I-pictures, set to zero */
    }
    else {
      if (pict_type == PCT_INTER) {
	COD = showbits(1);
      }
      else
	COD = 0; /* Intra picture -> not skipped */
    }


    if (!COD) {  /* COD == 0 --> not skipped */    

      if (syntax_arith_coding)  {
	if (pict_type == PCT_INTER) {
	  MCBPC_index = decode_a_symbol(cumf_MCBPC);
	  MCBPC = mcbpctab[MCBPC_index];
	}	
	else {
	  MCBPC_index = decode_a_symbol(cumf_MCBPC_intra);
	  MCBPC = mcbpc_intratab[MCBPC_index];
	}
      }

      else {
	if (pict_type == PCT_INTER)
	  flushbits(1); /* flush COD bit */
      	if (pict_type == PCT_INTRA) 
	  MCBPC = getMCBPCintra();
      	else
	  MCBPC = getMCBPC();
      }

      if (fault) goto resync;
      
      if (MCBPC == -1) { /* stuffing */
	goto read_cod;   /* read next COD without advancing MB count */
      }

      else {             /* normal MB data */

	Mode = MCBPC & 7;

	/* MODB and CBPB */
	if (pb_frame) {
	  CBPB = 0;
 	  if (syntax_arith_coding)  {
	    MODB_index = decode_a_symbol(cumf_MODB);
	    MODB = modb_tab[MODB_index];
	  }
	  else 
	    MODB = getMODB();
	  if (MODB == PBMODE_CBPB_MVDB) {
	    if (syntax_arith_coding)  {
	      for(i=0; i<4; i++) {
		YCBPB_index = decode_a_symbol(cumf_YCBPB);
		YCBPB = ycbpb_tab[YCBPB_index];
		CBPB |= (YCBPB << (6-1-i));
	      }
 
	      for(i=4; i<6; i++) {
		UVCBPB_index = decode_a_symbol(cumf_UVCBPB);
		UVCBPB = uvcbpb_tab[UVCBPB_index];
		CBPB |= (UVCBPB << (6-1-i));
	      }
	    }
	    else
	      CBPB = getbits(6);
	  }
	}

	if (syntax_arith_coding) {

	  if (Mode == MODE_INTRA || Mode == MODE_INTRA_Q) { /* Intra */
	    CBPY_index = decode_a_symbol(cumf_CBPY_intra);
	    CBPY = cbpy_intratab[CBPY_index];
	  }
	  else {
	    CBPY_index = decode_a_symbol(cumf_CBPY);
	    CBPY = cbpytab[CBPY_index];
	
	  }

	}
	else 
	  CBPY = getCBPY();
 
	/* Decode Mode and CBP */
	
	
	if (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)
	  {/* Intra */
	    if (!syntax_arith_coding)	
	      CBPY = CBPY^15;		/* needed in huffman coding only */
	  }

	CBP = (CBPY << 2) | (MCBPC >> 4);
      }

      if (Mode == MODE_INTER4V && !adv_pred_mode) 
	if (!quiet)
	  printf("8x8 vectors not allowed in normal prediction mode\n");
          /* Could set fault-flag and resync */


      if (Mode == MODE_INTER_Q || Mode == MODE_INTRA_Q) {
	/* Read DQUANT if necessary */

	if (syntax_arith_coding) {
	  DQUANT_index = decode_a_symbol(cumf_DQUANT);
	  DQUANT = dquanttab[DQUANT_index] - 2; 
	  quant +=DQUANT;
	}
	else {
	  DQUANT = getbits(2);
	  quant += DQ_tab[DQUANT];
	}

	if (quant > 31 || quant < 1) {
	  if (!quiet)
	    printf("Quantizer out of range: clipping\n");
	  quant = mmax(1,mmin(31,quant));
	  /* could set fault-flag and resync here */
	}
      }

      /* motion vectors */
      if (Mode == MODE_INTER || Mode == MODE_INTER_Q || 
	  Mode == MODE_INTER4V || pb_frame) {

	if (Mode == MODE_INTER4V) { startmv = 1; stopmv = 4;}
	else { startmv = 0; stopmv = 0;}

	for (k = startmv; k <= stopmv; k++) {
	  if (syntax_arith_coding) {
	    mvx_index = decode_a_symbol(cumf_MVD);
	    mvx = mvdtab[mvx_index];
	    mvy_index = decode_a_symbol(cumf_MVD);
	    mvy = mvdtab[mvy_index];
	  }
	  else {
	    mvx = getTMNMV();
	    mvy = getTMNMV();
	  }

	  pmv0 = find_pmv(xpos,ypos,k,0);
	  pmv1 = find_pmv(xpos,ypos,k,1);
	  mvx = motion_decode(mvx, pmv0);
	  mvy = motion_decode(mvy, pmv1);

	  /* Check mv's to prevent seg.faults when error rate is high */
	  if (!mv_outside_frame) {
	    bsize = k ? 8 : 16;
	    offset = k ? (((k-1)&1)<<3) : 0;
	    /* checking only integer component */
	    if ((xpos<<4) + (mvx/2) + offset < 0 ||
		(xpos<<4) + (mvx/2) + offset > (mb_width<<4) - bsize) {
	      if (!quiet)
		printf("mvx out of range: searching for sync\n");
	      fault = 1;
	    }
	    offset = k ? (((k-1)&2)<<2) : 0;
	    if ((ypos<<4) + (mvy/2) + offset < 0 ||
		(ypos<<4) + (mvy/2) + offset > (mb_height<<4) - bsize) {
	      if (!quiet)
		printf("mvy out of range: searching for sync\n");
	      fault = 1;
	    }
	  }
	  MV[0][k][ypos+1][xpos+1] = mvx;
	  MV[1][k][ypos+1][xpos+1] = mvy;
	}

	/* PB frame delta vectors */

	if (pb_frame) {
	  if (MODB == PBMODE_MVDB || MODB == PBMODE_CBPB_MVDB) {
	    if (syntax_arith_coding) {
	      mvdbx_index = decode_a_symbol(cumf_MVD);
	      mvdbx = mvdtab[mvdbx_index];
	      
	      mvdby_index = decode_a_symbol(cumf_MVD);
	      mvdby = mvdtab[mvdby_index];
	    }
	    else {
	      mvdbx = getTMNMV();
	      mvdby = getTMNMV();
	    }


	    mvdbx = motion_decode(mvdbx, 0);
	    mvdby = motion_decode(mvdby, 0);

	  }
	  else {
	    mvdbx = 0; 
	    mvdby = 0;
	  }
	}
      }

      if (fault) goto resync;

    }
    else { /* COD == 1 --> skipped MB */
      if (MBA>=MBAmax)
	return; /* all macroblocks decoded */
      if (!syntax_arith_coding)
	if (pict_type == PCT_INTER)
	  flushbits(1);

      Mode = MODE_INTER;
      
      /* Reset CBP */
      CBP = CBPB = 0;

      /* reset motion vectors */
      MV[0][0][ypos+1][xpos+1] = 0;
      MV[1][0][ypos+1][xpos+1] = 0;
      mvdbx = 0;
      mvdby = 0;
    }

    /* Store Mode*/
    modemap[ypos+1][xpos+1] = Mode;

    if (Mode == MODE_INTRA || Mode == MODE_INTRA_Q) 
      if (!pb_frame)
	MV[0][0][ypos+1][xpos+1]=MV[1][0][ypos+1][xpos+1] = 0;


  reconstruct_mb:

    /* pixel coordinates of top left corner of current macroblock */
    /* one delayed because of OBMC */
    if (xpos > 0) {
      bx = 16*(xpos-1);
      by = 16*ypos;
    }
    else {
      bx = coded_picture_width-16;
      by = 16*(ypos-1);
    }

    if (MBA > 0) {

      Mode = modemap[by/16+1][bx/16+1];

      /* forward motion compensation for B-frame */
      if (pb_frame)
	reconstruct(bx,by,0,pmvdbx,pmvdby);
      
      /* motion compensation for P-frame */
      if (Mode == MODE_INTER || Mode == MODE_INTER_Q || Mode == MODE_INTER4V)
	reconstruct(bx,by,1,0,0);

      /* copy or add block data into P-picture */
      for (comp=0; comp<blk_cnt; comp++) {
	/* inverse DCT */
	if (Mode == MODE_INTRA || Mode == MODE_INTRA_Q) {
	  if (refidct)
	    idctref(ld->block[comp]);
	  else
	    idct(ld->block[comp]);
	  addblock(comp,bx,by,0);
	}
	else if ( (pCBP & (1<<(blk_cnt-1-comp))) ) {
	  /* No need to to do this for blocks with no coeffs */
	  if (refidct)
	    idctref(ld->block[comp]);
	  else
	    idct(ld->block[comp]);
	  addblock(comp,bx,by,1);
	}
      }
      
      
      if (pb_frame) {
	/* add block data into B-picture */
	for (comp = 6; comp<blk_cnt+6; comp++) {
	  if (!pCOD || adv_pred_mode)
	    reconblock_b(comp-6,bx,by,Mode,pmvdbx,pmvdby);
	  if ( (pCBPB & (1<<(blk_cnt-1-comp%6))) ) {
	    if (refidct)
	      idctref(ld->block[comp]);
	    else
	      idct(ld->block[comp]);
	    addblock(comp,bx,by,1);
	  }
	}
      }
      
    } /* end if (MBA > 0) */

    if (!COD) {

      Mode = modemap[ypos+1][xpos+1];

      /* decode blocks */
      for (comp=0; comp<blk_cnt; comp++) {

        clearblock(comp);
	if (Mode == MODE_INTRA || Mode == MODE_INTRA_Q) { /* Intra */
	  bp = ld->block[comp];
	  if(syntax_arith_coding) {
	    INTRADC_index = decode_a_symbol(cumf_INTRADC);
	    bp[0] = intradctab[INTRADC_index];
	  }
	  else {
	    bp[0] = getbits(8);
	  }

	  if (bp[0] == 128)
	    if (!quiet)
	      fprintf(stderr,"Illegal DC-coeff: 1000000\n");
	  if (bp[0] == 255)  /* Spec. in H.26P, not in TMN4 */
	    bp[0] = 128;
	  bp[0] *= 8; /* Iquant */
	  if ( (CBP & (1<<(blk_cnt-1-comp))) ) {
	    if (!syntax_arith_coding)
	      getblock(comp,0);
	    else 
	      get_sac_block(comp,0);
	  }
	}
	else { /* Inter */
	  if ( (CBP & (1<<(blk_cnt-1-comp))) ) {
	    if (!syntax_arith_coding)
	      getblock(comp,1);
	    else
	      get_sac_block(comp,1);
	  }

	}
	if (fault) goto resync;
      }

      /* Decode B blocks */
      if (pb_frame) {
	for (comp=6; comp<blk_cnt+6; comp++) {
	  clearblock(comp);
	  if ( (CBPB & (1<<(blk_cnt-1-comp%6))) ) {
	    if (!syntax_arith_coding)
	      getblock(comp,1);
	    else
	      get_sac_block(comp,1);
	  }
	  if (fault) goto resync;
	}
      }
	  
    }

    /* advance to next macroblock */
    MBA++;

    pCBP = CBP; pCBPB = CBPB; pCOD = COD;
    pmvdbx = mvdbx; pmvdby = mvdby;
    fflush(stdout);

    if (MBA >= MBAmax && !last_done) {
      COD = 1;
      xpos = 0;
      ypos++;
      last_done = 1;
      goto reconstruct_mb;
    }

  }
}
//-----------------------------------------------------------------------------
/* set block to zero */
void CH263Dec::clearblock( int comp)
{
  int *bp;
  int i;

  bp = (int *)ld->block[comp];

  for (i=0; i<8; i++)
  {
    bp[0] = bp[1] = bp[2] = bp[3] = 0;
    bp += 4;
  }
}

//-----------------------------------------------------------------------------
/* move/add 8x8-Block from block[comp] to refframe or bframe */
void CH263Dec::addblock(int comp,int bx, int by, int addflag)
{
  int cc,i, iincr, P = 1;
  unsigned char *rfp;
  short *bp;

  bp = ld->block[comp];

  if (comp >= 6) {
    /* This is a component for B-frame forward prediction */
    P = 0;
    addflag = 1;
    comp -= 6;
  }

  cc = (comp<4) ? 0 : (comp&1)+1; /* color component index */

  if (cc==0) {
    /* luminance */
    
    /* frame DCT coding */
    if (P)
      rfp = newframe[0]
	+ coded_picture_width*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
    else
      rfp = bframe[0]
	+ coded_picture_width*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
    iincr = coded_picture_width;
  }
  else {
    /* chrominance */

    /* scale coordinates */
    bx >>= 1;
    by >>= 1;
    /* frame DCT coding */
    if (P)
      rfp = newframe[cc] + chrom_width*by + bx;
    else
      rfp = bframe[cc] + chrom_width*by + bx;
    iincr = chrom_width;
  }


  if (addflag) {
    for (i=0; i<8; i++) {
      rfp[0] = clp[bp[0]+rfp[0]];
      rfp[1] = clp[bp[1]+rfp[1]];
      rfp[2] = clp[bp[2]+rfp[2]];
      rfp[3] = clp[bp[3]+rfp[3]];
      rfp[4] = clp[bp[4]+rfp[4]];
      rfp[5] = clp[bp[5]+rfp[5]];
      rfp[6] = clp[bp[6]+rfp[6]];
      rfp[7] = clp[bp[7]+rfp[7]];
      bp += 8;
      rfp+= iincr;
    }
  }
  else  {
    for (i=0; i<8; i++) {
      rfp[0] = clp[bp[0]];
      rfp[1] = clp[bp[1]];
      rfp[2] = clp[bp[2]];
      rfp[3] = clp[bp[3]];
      rfp[4] = clp[bp[4]];
      rfp[5] = clp[bp[5]];
      rfp[6] = clp[bp[6]];
      rfp[7] = clp[bp[7]];
      bp += 8;
      rfp += iincr;
    }
  }
}

//-----------------------------------------------------------------------------
/* bidirectionally reconstruct 8x8-Block from block[comp] to bframe */
void CH263Dec::reconblock_b(int comp, int bx, int by, int mode, int bdx,int bdy)
{
  int cc,i,j,k, ii;
  unsigned char *bfr, *ffr;
  int BMVx, BMVy;
  int xa,xb,ya,yb,x,y,xvec,yvec,mvx,mvy;
  int xint,xhalf,yint,yhalf,pel;

  x = bx/16+1;y=by/16+1;

  if (mode == MODE_INTER4V) {
    if (comp < 4) {
      /* luma */
      mvx = MV[0][comp+1][y][x];
      mvy = MV[1][comp+1][y][x];
      BMVx = (bdx == 0 ? (trb-trd)* mvx/trd : trb * mvx/trd + bdx - mvx);
      BMVy = (bdy == 0 ? (trb-trd)* mvy/trd : trb * mvy/trd + bdy - mvy);
    }
    else {
      /* chroma */
      xvec = yvec = 0;
      for (k = 1; k <= 4; k++) {
	mvx = MV[0][k][y][x];
	mvy = MV[1][k][y][x];
	xvec += (bdx == 0 ? (trb-trd)* mvx/trd : trb * mvx/trd + bdx - mvx);
	yvec += (bdy == 0 ? (trb-trd)* mvy/trd : trb * mvy/trd + bdy - mvy);
      }
      
      /* chroma rounding (table 16/H.263) */
      BMVx = sign(xvec)*(roundtab[abs(xvec)%16] + (abs(xvec)/16)*2);
      BMVy = sign(yvec)*(roundtab[abs(yvec)%16] + (abs(yvec)/16)*2);
    }
  }
  else {
    if (comp < 4) {
      /* luma */
      mvx = MV[0][0][y][x];
      mvy = MV[1][0][y][x];
      BMVx = (bdx == 0 ? (trb-trd)* mvx/trd : trb * mvx/trd + bdx - mvx);
      BMVy = (bdy == 0 ? (trb-trd)* mvy/trd : trb * mvy/trd + bdy - mvy);
    }
    else {
      /* chroma */
      mvx = MV[0][0][y][x];
      mvy = MV[1][0][y][x];
      xvec = (bdx == 0 ? (trb-trd)* mvx/trd : trb * mvx/trd + bdx - mvx);
      yvec = (bdy == 0 ? (trb-trd)* mvy/trd : trb * mvy/trd + bdy - mvy);
      xvec *= 4;
      yvec *= 4;
      
      /* chroma rounding (table 16/H.263) */
      BMVx = sign(xvec)*(roundtab[abs(xvec)%16] + (abs(xvec)/16)*2);
      BMVy = sign(yvec)*(roundtab[abs(yvec)%16] + (abs(yvec)/16)*2);
    }
  }

  cc = (comp<4) ? 0 : (comp&1)+1; /* color component index */

  if (cc==0) {
    /* luminance */
    find_bidir_limits(BMVx,&xa,&xb,comp&1);
    find_bidir_limits(BMVy,&ya,&yb,(comp&2)>>1);
    bfr = bframe[0] +
      coded_picture_width*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
    ffr = newframe[0] +
      coded_picture_width*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
    ii = coded_picture_width;
  }
  else {
    /* chrominance */
    /* scale coordinates and vectors*/
    bx >>= 1;
    by >>= 1;

    find_bidir_chroma_limits(BMVx,&xa,&xb);
    find_bidir_chroma_limits(BMVy,&ya,&yb);

    bfr = bframe[cc]      + chrom_width*(by+((comp&2)<<2)) + bx + (comp&8);
    ffr = newframe[cc]    + chrom_width*(by+((comp&2)<<2)) + bx + (comp&8);
    ii = chrom_width;
  }

  xint = BMVx>>1;
  xhalf = BMVx - 2*xint;
  yint = BMVy>>1;
  yhalf = BMVy - 2*yint;

  ffr += xint + (yint+ya)*ii;
  bfr += ya*ii;
  
  if (!xhalf && !yhalf) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	pel = ffr[i];
	bfr[i] = ((unsigned int)(pel + bfr[i]))>>1;
      }
      bfr += ii;
      ffr += ii;
    }
  }
  else if (xhalf && !yhalf) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	pel = (ffr[i]+ffr[i+1]+1)>>1;
	bfr[i] = ((unsigned int)(pel + bfr[i]))>>1;
      }
      bfr += ii;
      ffr += ii;
    }
  }
  else if (!xhalf && yhalf) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	pel = (ffr[i]+ffr[ii+i]+1)>>1;
	bfr[i] = ((unsigned int)(pel + bfr[i]))>>1;
      }
      bfr += ii;
      ffr += ii;
    }
  }
  else { /* if (xhalf && yhalf) */
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	pel = ((unsigned int)(ffr[i]+ffr[i+1]+ffr[ii+i]+ffr[ii+i+1]+2))>>2;
	bfr[i] = (pel + bfr[i])>>1;
      }
      bfr += ii;
      ffr += ii;
    }
  }
  return;
}
//-----------------------------------------------------------------------------
int CH263Dec::motion_decode( int vec, int pmv)
{
  if (vec > 31) vec -= 64;
  vec += pmv;
  if (!long_vectors) {
    if (vec > 31)
      vec -= 64;
    if (vec < -32)
      vec += 64;
  }
  else {
    if (pmv < -31 && vec < -63)
      vec += 64;
    if (pmv > 32 && vec > 63)
      vec -= 64;
  }
  return vec;
}

//-----------------------------------------------------------------------------
int CH263Dec::find_pmv( int x,  int y, int block, int comp)
{
  int p1,p2,p3;
  int xin1,xin2,xin3;
  int yin1,yin2,yin3;
  int vec1,vec2,vec3;
  int l8,o8,or8;

  x++;y++;

  l8 = (modemap[y][x-1] == MODE_INTER4V ? 1 : 0);
  o8 =  (modemap[y-1][x] == MODE_INTER4V ? 1 : 0);
  or8 = (modemap[y-1][x+1] == MODE_INTER4V ? 1 : 0);

  switch (block) {
  case 0: 
    vec1 = (l8 ? 2 : 0) ; yin1 = y  ; xin1 = x-1;
    vec2 = (o8 ? 3 : 0) ; yin2 = y-1; xin2 = x;
    vec3 = (or8? 3 : 0) ; yin3 = y-1; xin3 = x+1;
    break;
  case 1:
    vec1 = (l8 ? 2 : 0) ; yin1 = y  ; xin1 = x-1;
    vec2 = (o8 ? 3 : 0) ; yin2 = y-1; xin2 = x;
    vec3 = (or8? 3 : 0) ; yin3 = y-1; xin3 = x+1;
    break;
  case 2:
    vec1 = 1            ; yin1 = y  ; xin1 = x;
    vec2 = (o8 ? 4 : 0) ; yin2 = y-1; xin2 = x;
    vec3 = (or8? 3 : 0) ; yin3 = y-1; xin3 = x+1;
    break;
  case 3:
    vec1 = (l8 ? 4 : 0) ; yin1 = y  ; xin1 = x-1;
    vec2 = 1            ; yin2 = y  ; xin2 = x;
    vec3 = 2            ; yin3 = y  ; xin3 = x;
    break;
  case 4:
    vec1 = 3            ; yin1 = y  ; xin1 = x;
    vec2 = 1            ; yin2 = y  ; xin2 = x;
    vec3 = 2            ; yin3 = y  ; xin3 = x;
    break;
  default:
    fprintf(stderr,"Illegal block number in find_pmv (getpic.c)\n");
    exit(1);
    break;
  }
  p1 = MV[comp][vec1][yin1][xin1];
  p2 = MV[comp][vec2][yin2][xin2];
  p3 = MV[comp][vec3][yin3][xin3];

  if (newgob && (block == 0 || block == 1 || block == 2))
    p2 = NO_VEC;

  if (p2 == NO_VEC) { p2 = p3 = p1; }

  return p1+p2+p3 - mmax(p1,mmax(p2,p3)) - mmin(p1,mmin(p2,p3));
}


//-----------------------------------------------------------------------------
void CH263Dec::find_bidir_limits( int vec, int *start, int *stop, int nhv)
{
  /* limits taken from C loop in section G5 in H.263 */
  *start = mmax(0,(-vec+1)/2 - nhv*8);
  *stop = mmin(7,15-(vec+1)/2 - nhv*8);

  (*stop)++; /* I use < and not <= in the loop */
}
//-----------------------------------------------------------------------------
void CH263Dec::find_bidir_chroma_limits( int vec, int *start, int *stop)
{

  /* limits taken from C loop in section G5 in H.263 */
  *start = mmax(0,(-vec+1)/2);
  *stop = mmin(7,7-(vec+1)/2);

  (*stop)++; /* I use < and not <= in the loop */
  return;
}
//-----------------------------------------------------------------------------
void CH263Dec::make_edge_image(unsigned char *src,unsigned char *dst,
							  int width, int height, int edge)
{
  int i,j;
  unsigned char *p1,*p2,*p3,*p4;
  unsigned char *o1,*o2,*o3,*o4;

  /* center image */
  p1 = dst;
  o1 = src;
  for (j = 0; j < height;j++) {
    for (i = 0; i < width; i++) {
      *(p1 + i) = *(o1 + i);
    }
    p1 += width + (edge<<1);
    o1 += width;
  }

  /* left and right edges */
  p1 = dst-1;
  o1 = src;
  for (j = 0; j < height;j++) {
    for (i = 0; i < edge; i++) {
      *(p1 - i) = *o1;
      *(p1 + width + i + 1) = *(o1 + width - 1);
    }
    p1 += width + (edge<<1);
    o1 += width;
  }    
    
  /* top and bottom edges */
  p1 = dst;
  p2 = dst + (width + (edge<<1))*(height-1);
  o1 = src;
  o2 = src + width*(height-1);
  for (j = 0; j < edge;j++) {
    p1 = p1 - (width + (edge<<1));
    p2 = p2 + (width + (edge<<1));
    for (i = 0; i < width; i++) {
      *(p1 + i) = *(o1 + i);
      *(p2 + i) = *(o2 + i);
    }
  }    

  /* corners */
  p1 = dst - (width+(edge<<1)) - 1;
  p2 = p1 + width + 1;
  p3 = dst + (width+(edge<<1))*(height)-1;
  p4 = p3 + width + 1;

  o1 = src;
  o2 = o1 + width - 1;
  o3 = src + width*(height-1);
  o4 = o3 + width - 1;
  for (j = 0; j < edge; j++) {
    for (i = 0; i < edge; i++) {
      *(p1 - i) = *o1;
      *(p2 + i) = *o2;
      *(p3 - i) = *o3;
      *(p4 + i) = *o4; 
    }
    p1 = p1 - (width + (edge<<1));
    p2 = p2 - (width + (edge<<1));
    p3 = p3 + width + (edge<<1);
    p4 = p4 + width + (edge<<1);
  }

}

//-----------------------------------------------------------------------------  
void CH263Dec::interpolate_image(unsigned char *in, unsigned char *out, int width, int height)
/* only used for displayed interpolated frames, not reconstructed ones */
{

  int x,xx,y,w2;

  unsigned char *pp,*ii;

  w2 = 2*width;

  /* Horizontally */
  pp = out;
  ii = in;
  for (y = 0; y < height-1; y++) {
    for (x = 0,xx=0; x < width-1; x++,xx+=2) {
      *(pp + xx) = *(ii + x);
      *(pp + xx+1) = ((unsigned int)(*(ii + x)  + *(ii + x + 1)))>>1;
      *(pp + w2 + xx) = ((unsigned int)(*(ii + x) + *(ii + x + width)))>>1;
      *(pp + w2 + xx+1) = ((unsigned int)(*(ii + x) + *(ii + x + 1) + 
			   *(ii + x + width) + *(ii + x + width + 1)))>>2;
      
    }
    *(pp + w2 - 2) = *(ii + width - 1);
    *(pp + w2 - 1) = *(ii + width - 1);
    *(pp + w2 + w2 - 2) = *(ii + width + width - 1);
    *(pp + w2 + w2 - 1) = *(ii + width + width - 1);
    pp += w2<<1;
    ii += width;
  }

  /* last lines */
  for (x = 0,xx=0; x < width-1; x++,xx+=2) {
    *(pp+ xx) = *(ii + x);    
    *(pp+ xx+1) = ((unsigned int)(*(ii + x) + *(ii + x + 1) + 1))>>1;
    *(pp+ w2+ xx) = *(ii + x);    
    *(pp+ w2+ xx+1) = ((unsigned int)(*(ii + x) + *(ii + x + 1) + 1))>>1;
  }
  
  /* bottom right corner pels */
  *(pp + (width<<1) - 2) = *(ii + width -1);
  *(pp + (width<<1) - 1) = *(ii + width -1);
  *(pp + (width<<2) - 2) = *(ii + width -1);
  *(pp + (width<<2) - 1) = *(ii + width -1);

  return;
}
//-----------------------------------------------------------------------------    
int CH263Dec::getTMNMV()
{
  int code;

  if (getbits1())
  {
    return 0;
  }

  if ((code = showbits(12))>=512)
  {
    code = (code>>8) - 2;
    flushbits(TMNMVtab0[code].len);

    return TMNMVtab0[code].val;
  }

  if (code>=128)
  {
    code = (code>>2) -32;
    flushbits(TMNMVtab1[code].len);

    return TMNMVtab1[code].val;
  }

  if ((code-=5)<0)
  {
    if (!quiet)
      fprintf(stderr,"Invalid motion_vector code\n");
    fault=1;
    return 0;
  }

  flushbits(TMNMVtab2[code].len);

  return TMNMVtab2[code].val;
}

//-----------------------------------------------------------------------------
int CH263Dec::getMCBPC()
{
  int code;

  code = showbits(9);

  if (code == 1) {
    flushbits(9);
    return -1;
  }

  if (code == 0) {
    if (!quiet) 
      fprintf(stderr,"Invalid MCBPC code\n");
    fault = 1;
    return 0;
  }
    
  if (code>=256)
  {
    flushbits(1);
    return 0;
  }
    
  flushbits(MCBPCtab[code].len);

  return MCBPCtab[code].val;
}
//-----------------------------------------------------------------------------
int CH263Dec::getMODB()
{
  int code;
  int MODB;

  code = showbits(2);

  if (code < 2) {
    MODB = 0;
    flushbits(1);
  }
  else if (code == 2) {
    MODB = 1;
    flushbits(2);
  }
  else { /* code == 3 */
    MODB = 2;
    flushbits(2);
  }
  return MODB;
}

//-----------------------------------------------------------------------------
int CH263Dec::getMCBPCintra()
{
  int code;

  code = showbits(9);

  if (code == 1) {
    flushbits(9);
    return -1;
  }

  if (code < 8) {
    if (!quiet) 
      fprintf(stderr,"Invalid MCBPCintra code\n");
    fault = 1;
    return 0;
  }

  code >>= 3;
    
  if (code>=32)
  {
    flushbits(1);
    return 3;
  }

  flushbits(MCBPCtabintra[code].len);

  return MCBPCtabintra[code].val;
}
//-----------------------------------------------------------------------------
int CH263Dec::getCBPY()
{
  int code;

  code = showbits(6);
  if (code < 2) {
    if (!quiet) 
      fprintf(stderr,"Invalid CBPY code\n");
    fault = 1;
    return -1;
  }
    
  if (code>=48)
  {
    flushbits(2);
    return 0;
  }

  flushbits(CBPYtab[code].len);

  return CBPYtab[code].val;
}

//-----------------------------------------------------------------------------
void CH263Dec::init_display(void)
{
  pbmi    = (PBITMAPINFO)malloc(sizeof(BITMAPINFOHEADER) + 240 * sizeof(RGBQUAD));
  pbmi->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
  pbmi->bmiHeader.biPlanes = 1;
  pbmi->bmiHeader.biCompression = 0l;
  pbmi->bmiHeader.biSizeImage = 0l;
  pbmi->bmiHeader.biXPelsPerMeter = 0l;
  pbmi->bmiHeader.biYPelsPerMeter = 0l;
  pbmi->bmiHeader.biClrUsed = 240;
  pbmi->bmiHeader.biClrImportant = 240;
  pbmi->bmiHeader.biBitCount = nDitherType;
  pbmi->bmiHeader.biWidth = coded_picture_width ;
  pbmi->bmiHeader.biHeight= coded_picture_height;
  if ( pbmi->bmiHeader.biBitCount==8 )
  {
     // for 8 BPP, build the color palette

     LOGPALETTE *plgpl;
     short      *pPalIndex;
     int         crv, cbu, cgu, cgv;
     int         y, u, v;
     int         i;

     plgpl = (LOGPALETTE*) malloc(sizeof(LOGPALETTE) + 240 * sizeof(PALETTEENTRY));
     plgpl->palNumEntries = 240;
     plgpl->palVersion = 0x300;
     pPalIndex=(short *)pbmi->bmiColors;

     /* matrix coefficients */
     crv = convmat[matrix_coefficients][0];
     cbu = convmat[matrix_coefficients][1];
     cgu = convmat[matrix_coefficients][2];
     cgv = convmat[matrix_coefficients][3];

     for (i=16; i<240; i++)
     {
       /* color space conversion */
       y = 16*((i>>4)&15) + 8;
       u = 32*((i>>2)&3)  - 48;
       v = 32*(i&3)       - 48;

       y = 76309 * (y - 16); /* (255/219)*65536 */

       plgpl->palPalEntry[i].peRed   = clp[(y + crv*v + 32768)>>16];
       plgpl->palPalEntry[i].peGreen = clp[(y - cgu*u -cgv*v + 32768)>>16];
       plgpl->palPalEntry[i].peBlue  = clp[(y + cbu*u + 32786)>>16];
       pPalIndex[i]=i;
     }
     hpal = CreatePalette(plgpl);
     free(plgpl);
     hPalPrev=SelectPalette(myDC,hpal,FALSE);
     RealizePalette(myDC);
  }
}
//-----------------------------------------------------------------------------
/* display clean-up */
void CH263Dec::exit_display()
{
   if (pbmi)
   {
      free(pbmi);
      pbmi=NULL;
   }
   if (hPalPrev)
   {
       SelectPalette(myDC,hPalPrev,FALSE);
       DeleteObject(hpal);
       hPalPrev=NULL;
   }
}


//-----------------------------------------------------------------------------
/* VFWԼ ̿Ͽ ȭ鿡 Ѵ. */
void CH263Dec::init_dib()
{
	/* Ʈ  Ѵ. */
	biHeader.biWidth		 = 176;
	biHeader.biHeight		 = 144;
	biHeader.biSize			 = sizeof(BITMAPINFOHEADER);
	biHeader.biCompression   = BI_RGB;
	biHeader.biPlanes		 = 1;
	biHeader.biBitCount		 = nDitherType;

	biHeader.biSizeImage	 = 0;
	biHeader.biXPelsPerMeter = 0;
	biHeader.biYPelsPerMeter = 0;
	biHeader.biClrUsed		 = 0;
	biHeader.biClrImportant  = 0;

	/* VFW  ʱ⼳ ش. */
	hDrawDib = DrawDibOpen();
	
	VERIFY(hDrawDib);

	DrawDibBegin(
		hDrawDib,
		myDC,
		176,
		144,
		&biHeader,
		176,
		144,
		0);
}
//-----------------------------------------------------------------------------
void CH263Dec::close_dib()
{
	DrawDibEnd(hDrawDib);
	DrawDibClose(hDrawDib);
}
//-----------------------------------------------------------------------------
void CH263Dec::display_image(unsigned char *dithered_image)
{
  /* display dithered image */
	if ( nDitherType == 8)
	{
		StretchDIBits ( myDC, m_x, m_y, m_dx, m_dy,
					0, 0, coded_picture_width, coded_picture_height, 
					dithered_image, pbmi,DIB_PAL_COLORS, SRCCOPY);
		return;
	}
	DrawDibDraw(
		hDrawDib,
		myDC,
		m_x, m_y, m_dx,	m_dy,
		&biHeader,
		dithered_image,
		0, 0, 176,	144,
		DDF_SAME_DRAW
		);
}

//-----------------------------------------------------------------------------
/* decode a file and store and/or display it */
int CH263Dec::play_movie(void)
{
	gusState = PLAYING;
	ld = &base;
    initbits();

    return 0;
} 


/////////////////////////////////////////////////////////////////////
//
// MPEG related functions
//
// (portable)
//
//-----------------------------------------------------------------------------
void CH263Dec::initdecoder()
{
  
  int i, cc, size;
  FILE *cleared;

  /* clip table */
  if (!(clp=(unsigned char *)malloc(1024)))
    error("malloc failed\n");

  clp += 384;

  for (i=-384; i<640; i++)
    clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* MPEG-1 = TMN parameters */
  matrix_coefficients = 5;

  if (source_format == SF_CIF) {  // Input stream is CIF
    m_MBC = 22;
    m_MBR = 18;
  }
  else if (source_format == SF_QCIF) {   // Input stream is QCIF
    m_MBC = 11;
    m_MBR = 9;
  }
  else if (source_format == SF_SQCIF) {  // Input stream is SQCIF
	m_MBC = 8;
    m_MBR = 6;
  }
  else
    exit(1);

  horizontal_size = m_MBC*16;
  vertical_size = m_MBR*16;

  mb_width		= horizontal_size/16;
  mb_height		= vertical_size/16;
  coded_picture_width = 16*mb_width;
  coded_picture_height = 16*mb_height;
  chrom_width	=  coded_picture_width>>1;
  chrom_height	= coded_picture_height>>1;
  blk_cnt = 6;

  for (cc=0; cc<3; cc++) {
    if (cc==0)
      size = coded_picture_width*coded_picture_height;
    else
      size = chrom_width*chrom_height;

    if (!(refframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");

    if (!(oldrefframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");

    if (!(bframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");
  }

  for (cc=0; cc<3; cc++) {
    if (cc==0) {
      size = (coded_picture_width+64)*(coded_picture_height+64);
      if (!(edgeframeorig[cc] = (unsigned char *)malloc(size)))
	error("malloc failed\n");
      edgeframe[cc] = edgeframeorig[cc] + (coded_picture_width+64) * 32 + 32;
    }
    else {
      size = (chrom_width+32)*(chrom_height+32);
      if (!(edgeframeorig[cc] = (unsigned char *)malloc(size)))
	error("malloc failed\n");
      edgeframe[cc] = edgeframeorig[cc] + (chrom_width+32) * 16 + 16;
    }
  }

  if (expand) {
    for (cc=0; cc<3; cc++) {
      if (cc==0)
	size = coded_picture_width*coded_picture_height*4;
      else
	size = chrom_width*chrom_height*4;
      
      if (!(exnewframe[cc] = (unsigned char *)malloc(size)))
	error("malloc failed\n");
    }
  }
  /* Clear output file for concatenated storing */
  if (outtype == T_YUV_CONC) {
    if ((cleared = fopen(outputname,"wb")) == NULL) 

      error("couldn't clear outputfile\n");
    else
      fclose(cleared);
  }
  /* IDCT */
  if (refidct)
    init_idctref();
  else
    init_idct();

}

//-----------------------------------------------------------------------------
void CH263Dec::error( char *text)
{
	::MessageBox(NULL,  text,"H263 DEC ERROR", MB_OK | MB_ICONERROR );
	exit(1);
}

/* compliance warning messages to user, but don't exit() */

/* trace output */
//-----------------------------------------------------------------------------
void CH263Dec::printbits(int code, int bits, int len)
{
  int i;
  for (i=0; i<len; i++)
    printf("%d",(code>>(bits-1-i))&1);
}

//-----------------------------------------------------------------------------
int CH263Dec::getval(char *argv[])
{
  int val;

  if (sscanf(argv[1]+2,"%d",&val)!=1)
    return 0;

  while (isdigit(argv[1][2]))
    argv[1]++;

  return val;
}

//-----------------------------------------------------------------------------
void CH263Dec::idctrow( short *blk)
{
  int x0, x1, x2, x3, x4, x5, x6, x7, x8;

  /* shortcut */
  if (!((x1 = blk[4]<<11) | (x2 = blk[6]) | (x3 = blk[2]) |
        (x4 = blk[1]) | (x5 = blk[7]) | (x6 = blk[5]) | (x7 = blk[3])))
  {
    blk[0]=blk[1]=blk[2]=blk[3]=blk[4]=blk[5]=blk[6]=blk[7]=blk[0]<<3;
    return;
  }

  x0 = (blk[0]<<11) + 128; /* for proper rounding in the fourth stage */

  /* first stage */
  x8 = W7*(x4+x5);
  x4 = x8 + (W1-W7)*x4;
  x5 = x8 - (W1+W7)*x5;
  x8 = W3*(x6+x7);
  x6 = x8 - (W3-W5)*x6;
  x7 = x8 - (W3+W5)*x7;
  
  /* second stage */
  x8 = x0 + x1;
  x0 -= x1;
  x1 = W6*(x3+x2);
  x2 = x1 - (W2+W6)*x2;
  x3 = x1 + (W2-W6)*x3;
  x1 = x4 + x6;
  x4 -= x6;
  x6 = x5 + x7;
  x5 -= x7;
  
  /* third stage */
  x7 = x8 + x3;
  x8 -= x3;
  x3 = x0 + x2;
  x0 -= x2;
  x2 = (181*(x4+x5)+128)>>8;
  x4 = (181*(x4-x5)+128)>>8;
  
  /* fourth stage */
  blk[0] = (x7+x1)>>8;
  blk[1] = (x3+x2)>>8;
  blk[2] = (x0+x4)>>8;
  blk[3] = (x8+x6)>>8;
  blk[4] = (x8-x6)>>8;
  blk[5] = (x0-x4)>>8;
  blk[6] = (x3-x2)>>8;
  blk[7] = (x7-x1)>>8;
}

/* column (vertical) IDCT
 *
 *             7                         pi         1
 * dst[8*k] = sum c[l] * src[8*l] * cos( -- * ( k + - ) * l )
 *            l=0                        8          2
 *
 * where: c[0]    = 1/1024
 *        c[1..7] = (1/1024)*sqrt(2)
 */
//-----------------------------------------------------------------------------
void CH263Dec::idctcol( short *blk)
{
  int x0, x1, x2, x3, x4, x5, x6, x7, x8;

  /* shortcut */
  if (!((x1 = (blk[8*4]<<8)) | (x2 = blk[8*6]) | (x3 = blk[8*2]) |
        (x4 = blk[8*1]) | (x5 = blk[8*7]) | (x6 = blk[8*5]) | (x7 = blk[8*3])))
  {
    blk[8*0]=blk[8*1]=blk[8*2]=blk[8*3]=blk[8*4]=blk[8*5]=blk[8*6]=blk[8*7]=
      iclp[(blk[8*0]+32)>>6];
    return;
  }

  x0 = (blk[8*0]<<8) + 8192;

  /* first stage */
  x8 = W7*(x4+x5) + 4;
  x4 = (x8+(W1-W7)*x4)>>3;
  x5 = (x8-(W1+W7)*x5)>>3;
  x8 = W3*(x6+x7) + 4;
  x6 = (x8-(W3-W5)*x6)>>3;
  x7 = (x8-(W3+W5)*x7)>>3;
  
  /* second stage */
  x8 = x0 + x1;
  x0 -= x1;
  x1 = W6*(x3+x2) + 4;
  x2 = (x1-(W2+W6)*x2)>>3;
  x3 = (x1+(W2-W6)*x3)>>3;
  x1 = x4 + x6;
  x4 -= x6;
  x6 = x5 + x7;
  x5 -= x7;
  
  /* third stage */
  x7 = x8 + x3;
  x8 -= x3;
  x3 = x0 + x2;
  x0 -= x2;
  x2 = (181*(x4+x5)+128)>>8;
  x4 = (181*(x4-x5)+128)>>8;
  
  /* fourth stage */
  blk[8*0] = iclp[(x7+x1)>>14];
  blk[8*1] = iclp[(x3+x2)>>14];
  blk[8*2] = iclp[(x0+x4)>>14];
  blk[8*3] = iclp[(x8+x6)>>14];
  blk[8*4] = iclp[(x8-x6)>>14];
  blk[8*5] = iclp[(x0-x4)>>14];
  blk[8*6] = iclp[(x3-x2)>>14];
  blk[8*7] = iclp[(x7-x1)>>14];
}

//-----------------------------------------------------------------------------
/* two dimensional inverse discrete cosine transform */
void CH263Dec::idct(short *block)
{
  int i;

  for (i=0; i<8; i++)
    idctrow(block+8*i);

  for (i=0; i<8; i++)
    idctcol(block+i);
}

//-----------------------------------------------------------------------------
void CH263Dec::init_idct()
{
  int i;

  iclp = iclip+512;
  for (i= -512; i<512; i++)
    iclp[i] = (i<-256) ? -256 : ((i>255) ? 255 : i);
}

//-----------------------------------------------------------------------------
void CH263Dec::init_idctref()
{
  int freq, time;
  double scale;

  for (freq=0; freq < 8; freq++)
  {
    scale = (freq == 0) ? sqrt(0.125) : 0.5;
    for (time=0; time<8; time++)
      c[freq][time] = scale*cos((PI/8.0)*freq*(time + 0.5));
  }
}

//-----------------------------------------------------------------------------
/* perform IDCT matrix multiply for 8x8 coefficient block */
void CH263Dec::idctref(short *block)
{
  int i, j, k, v;
  double partial_product;
  double tmp[64];

  for (i=0; i<8; i++)
    for (j=0; j<8; j++)
    {
      partial_product = 0.0;

      for (k=0; k<8; k++)
        partial_product+= c[k][j]*block[8*i+k];

      tmp[8*i+j] = partial_product;
    }

  /* Transpose operation is integrated into address mapping by switching 
     loop order of i and j */

  for (j=0; j<8; j++)
    for (i=0; i<8; i++)
    {
      partial_product = 0.0;

      for (k=0; k<8; k++)
        partial_product+= c[k][i]*tmp[8*k+j];

      v = (int)floor(partial_product+0.5);
      block[8*i+j] = (v<-256) ? -256 : ((v>255) ? 255 : v);
    }
}
//-----------------------------------------------------------------------------
void CH263Dec::reconstruct(int bx, int by, int P, int bdx, int bdy)
{
  int w,h,lx,lx2,dx,dy,xp,yp,comp,sum;
  int x,y,mode,xvec,yvec;
  unsigned char *src[3];

  x = bx/16+1; y = by/16+1;
  lx = coded_picture_width;

  if (mv_outside_frame) {
    lx2 = coded_picture_width + 64;
    src[0] = edgeframe[0];
    src[1] = edgeframe[1];
    src[2] = edgeframe[2];
  }
  else {
    lx2 = coded_picture_width;
    src[0] = oldrefframe[0];
    src[1] = oldrefframe[1];
    src[2] = oldrefframe[2];
  }

  mode = modemap[y][x];

  if (P) {
    /* P prediction */
    if (adv_pred_mode) {
      w = 8; h = 8;
      /* Y*/
      for (comp = 0; comp < 4; comp++) {
	xp = bx + ((comp&1)<<3);
	yp = by + ((comp&2)<<2);
	recon_comp_obmc(src[0],newframe[0], lx,lx2,comp,w,h,xp,yp);
      }
      if (mode == MODE_INTER4V) {

	sum = MV[0][1][y][x]+MV[0][2][y][x]+ MV[0][3][y][x]+MV[0][4][y][x];
	dx = sign(sum)*(roundtab[abs(sum)%16] + (abs(sum)/16)*2);

	sum = MV[1][1][y][x]+MV[1][2][y][x]+ MV[1][3][y][x]+MV[1][4][y][x];
	dy = sign(sum)*(roundtab[abs(sum)%16] + (abs(sum)/16)*2);

      }
      else {
	dx = MV[0][0][y][x];
	dy = MV[1][0][y][x];
	/* chroma rounding */
	dx = ( dx % 4 == 0 ? dx >> 1 : (dx>>1)|1 );
	dy = ( dy % 4 == 0 ? dy >> 1 : (dy>>1)|1 );
      }
      lx>>=1;bx>>=1; lx2>>=1; 
      by>>=1;
      /* Chroma */
      recon_comp(src[1],newframe[1], lx,lx2,w,h,bx,by,dx,dy,1);
      recon_comp(src[2],newframe[2], lx,lx2,w,h,bx,by,dx,dy,2);
    }
    else { /* normal prediction mode */
      /* P prediction */
      w = 16; h = 16;
      dx = MV[0][0][y][x];
      dy = MV[1][0][y][x];
      
      /* Y */
      recon_comp(src[0],newframe[0], lx,lx2,w,h,bx,by,dx,dy,0);
      
      lx>>=1; w>>=1; bx>>=1; lx2>>=1; 
      h>>=1; by>>=1;  
      /* chroma rounding */
      dx = ( dx % 4 == 0 ? dx >> 1 : (dx>>1)|1 );
      dy = ( dy % 4 == 0 ? dy >> 1 : (dy>>1)|1 );

      /* Chroma */
      recon_comp(src[1],newframe[1], lx,lx2,w,h,bx,by,dx,dy,1);
      recon_comp(src[2],newframe[2], lx,lx2,w,h,bx,by,dx,dy,2);
    }
  }

  else {
    /* B forward prediction */
    if (adv_pred_mode) {
      if (mode == MODE_INTER4V) {
	w = 8; h = 8;
	/* Y*/
	xvec = yvec = 0;
	for (comp = 0; comp < 4; comp++) {
	  xvec += dx = (trb)*MV[0][comp+1][y][x]/trd + bdx;
	  yvec += dy = (trb)*MV[1][comp+1][y][x]/trd + bdy;
	  xp = bx + ((comp&1)<<3);
	  yp = by + ((comp&2)<<2);
	  recon_comp(src[0],bframe[0], lx,lx2,w,h,xp,yp,dx,dy,0);
	}

	/* chroma rounding (table 16/H.263) */
	dx = sign(xvec)*(roundtab[abs(xvec)%16] + (abs(xvec)/16)*2);
	dy = sign(yvec)*(roundtab[abs(yvec)%16] + (abs(yvec)/16)*2);

	lx>>=1;bx>>=1; lx2>>=1;
	by>>=1;
	/* Chroma */
	recon_comp(src[1],bframe[1], lx,lx2,w,h,bx,by,dx,dy,1);
	recon_comp(src[2],bframe[2], lx,lx2,w,h,bx,by,dx,dy,2);
      }
      else {  /* adv_pred_mode but 16x16 vector */
	w = 16; h = 16;

	dx = (trb)*MV[0][0][y][x]/trd + bdx;
	dy = (trb)*MV[1][0][y][x]/trd + bdy;
	/* Y */
	recon_comp(src[0],bframe[0], lx,lx2,w,h,bx,by,dx,dy,0);
	
	lx>>=1; w>>=1; bx>>=1; lx2>>=1;
	h>>=1; by>>=1;  

	xvec = 4*dx;
	yvec = 4*dy;

	/* chroma rounding (table 16/H.263) */
	dx = sign(xvec)*(roundtab[abs(xvec)%16] + (abs(xvec)/16)*2);
	dy = sign(yvec)*(roundtab[abs(yvec)%16] + (abs(yvec)/16)*2);

	/* Chroma */
	recon_comp(src[1],bframe[1], lx,lx2,w,h,bx,by,dx,dy,1);
	recon_comp(src[2],bframe[2], lx,lx2,w,h,bx,by,dx,dy,2);
      }
    }
    else { /* normal B forward prediction */

      w = 16; h = 16;
      dx = (trb)*MV[0][0][y][x]/trd + bdx;
      dy = (trb)*MV[1][0][y][x]/trd + bdy;
      /* Y */
      recon_comp(src[0],bframe[0], lx,lx2,w,h,bx,by,dx,dy,0);

      lx>>=1; w>>=1; bx>>=1; lx2>>=1;
      h>>=1; by>>=1;  

      xvec = 4*dx;
      yvec = 4*dy;

      /* chroma rounding (table 16/H.263) */ 
      dx = sign(xvec)*(roundtab[abs(xvec)%16] + (abs(xvec)/16)*2);
      dy = sign(yvec)*(roundtab[abs(yvec)%16] + (abs(yvec)/16)*2);

      /* Chroma */
      recon_comp(src[1],bframe[1], lx,lx2,w,h,bx,by,dx,dy,1);
      recon_comp(src[2],bframe[2], lx,lx2,w,h,bx,by,dx,dy,2);
    }
  }
}



//-----------------------------------------------------------------------------
void CH263Dec::recon_comp( unsigned char *src,unsigned char *dst,int lx, int lx2,
						 int w, int h, int x, int y, int dx, int dy, int chroma)
{
  int xint, xh, yint, yh;
  unsigned char *s, *d;

  xint = dx>>1;
  xh = dx & 1;
  yint = dy>>1;
  yh = dy & 1;

  /* origins */
  s = src + lx2*(y+yint) + x + xint;
  d = dst + lx*y + x;

  if (!xh && !yh)
    if (w!=8)
      rec(s,d,lx,lx2,h);
    else
      recc(s,d,lx,lx2,h);
  else if (!xh && yh)
    if (w!=8)
      recve(s,d,lx,lx2,h);
    else 
      recvc(s,d,lx,lx2,h);
  else if (xh && !yh)
    if (w!=8)
      rech(s,d,lx,lx2,h);
    else
      rechc(s,d,lx,lx2,h);
  else /* if (xh && yh) */
    if (w!=8)
      rec4(s,d,lx,lx2,h);
    else
      rec4c(s,d,lx,lx2,h);
}
//-----------------------------------------------------------------------------
void rec(unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  int j;

  for (j=0; j<h; j++)
  {
    d[0] = s[0];
    d[1] = s[1];
    d[2] = s[2];
    d[3] = s[3];
    d[4] = s[4];
    d[5] = s[5];
    d[6] = s[6];
    d[7] = s[7];
    d[8] = s[8];
    d[9] = s[9];
    d[10] = s[10];
    d[11] = s[11];
    d[12] = s[12];
    d[13] = s[13];
    d[14] = s[14];
    d[15] = s[15];
    s+= lx2;
    d+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recc( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  int j;

  for (j=0; j<h; j++)
  {
    d[0] = s[0];
    d[1] = s[1];
    d[2] = s[2];
    d[3] = s[3];
    d[4] = s[4];
    d[5] = s[5];
    d[6] = s[6];
    d[7] = s[7];
    s+= lx2;
    d+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::reco(unsigned char *s, int *d, int lx, int lx2,
				   int addflag, int c, int xa, int xb, int ya, int yb)
{
  int i,j;
  int *om;

  om = &OM[c][ya][0];

  if (!addflag) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	d[i] = s[i] * om[i];
      }
      s+= lx2;
      d+= lx;
      om+= 8;
    }
  }
  else {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	d[i] += s[i] * om[i];
      }
      s+= lx2;
      d+= lx;
      om+= 8;
    }
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::rech(unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp;
  int j;
  unsigned int s1,s2;

  sp = s;
  dp = d;
  for (j=0; j<h; j++)
  {
    s1=sp[0];
    dp[0] = (unsigned int)(s1+(s2=sp[1])+1)>>1;
    dp[1] = (unsigned int)(s2+(s1=sp[2])+1)>>1;
    dp[2] = (unsigned int)(s1+(s2=sp[3])+1)>>1;
    dp[3] = (unsigned int)(s2+(s1=sp[4])+1)>>1;
    dp[4] = (unsigned int)(s1+(s2=sp[5])+1)>>1;
    dp[5] = (unsigned int)(s2+(s1=sp[6])+1)>>1;
    dp[6] = (unsigned int)(s1+(s2=sp[7])+1)>>1;
    dp[7] = (unsigned int)(s2+(s1=sp[8])+1)>>1;
    dp[8] = (unsigned int)(s1+(s2=sp[9])+1)>>1;
    dp[9] = (unsigned int)(s2+(s1=sp[10])+1)>>1;
    dp[10] = (unsigned int)(s1+(s2=sp[11])+1)>>1;
    dp[11] = (unsigned int)(s2+(s1=sp[12])+1)>>1;
    dp[12] = (unsigned int)(s1+(s2=sp[13])+1)>>1;
    dp[13] = (unsigned int)(s2+(s1=sp[14])+1)>>1;
    dp[14] = (unsigned int)(s1+(s2=sp[15])+1)>>1;
    dp[15] = (unsigned int)(s2+sp[16]+1)>>1;
    sp+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::rechc( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp;
  int j;
  unsigned int s1,s2;

  sp = s;
  dp = d;
  for (j=0; j<h; j++)
  {
    s1=sp[0];
    dp[0] = (unsigned int)(s1+(s2=sp[1])+1)>>1;
    dp[1] = (unsigned int)(s2+(s1=sp[2])+1)>>1;
    dp[2] = (unsigned int)(s1+(s2=sp[3])+1)>>1;
    dp[3] = (unsigned int)(s2+(s1=sp[4])+1)>>1;
    dp[4] = (unsigned int)(s1+(s2=sp[5])+1)>>1;
    dp[5] = (unsigned int)(s2+(s1=sp[6])+1)>>1;
    dp[6] = (unsigned int)(s1+(s2=sp[7])+1)>>1;
    dp[7] = (unsigned int)(s2+sp[8]+1)>>1;
    sp+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recho( unsigned char *s, int *d, int lx, int lx2,
					int addflag, int c, int xa, int xb, int ya, int yb)
{
  int *dp,*om;
  unsigned char *sp;
  int i,j;

  sp = s;
  dp = d;
  om = &OM[c][ya][0];

  if (!addflag) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] = (((unsigned int)(sp[i] + sp[i+1]+1))>>1)*om[i];
      }
      sp+= lx2;
      dp+= lx;
      om+= 8;
    }
  }
  else {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] += (((unsigned int)(sp[i] + sp[i+1]+1))>>1)*OM[c][j][i];
      }
      sp+= lx2;
      dp+= lx;
      om+= 8;
    }
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recve( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp,*sp2;
  int j;

  sp = s;
  sp2 = s+lx2;
  dp = d;
  for (j=0; j<h; j++)
  {
    dp[0] = (unsigned int)(sp[0]+sp2[0]+1)>>1;
    dp[1] = (unsigned int)(sp[1]+sp2[1]+1)>>1;
    dp[2] = (unsigned int)(sp[2]+sp2[2]+1)>>1;
    dp[3] = (unsigned int)(sp[3]+sp2[3]+1)>>1;
    dp[4] = (unsigned int)(sp[4]+sp2[4]+1)>>1;
    dp[5] = (unsigned int)(sp[5]+sp2[5]+1)>>1;
    dp[6] = (unsigned int)(sp[6]+sp2[6]+1)>>1;
    dp[7] = (unsigned int)(sp[7]+sp2[7]+1)>>1;
    dp[8] = (unsigned int)(sp[8]+sp2[8]+1)>>1;
    dp[9] = (unsigned int)(sp[9]+sp2[9]+1)>>1;
    dp[10] = (unsigned int)(sp[10]+sp2[10]+1)>>1;
    dp[11] = (unsigned int)(sp[11]+sp2[11]+1)>>1;
    dp[12] = (unsigned int)(sp[12]+sp2[12]+1)>>1;
    dp[13] = (unsigned int)(sp[13]+sp2[13]+1)>>1;
    dp[14] = (unsigned int)(sp[14]+sp2[14]+1)>>1;
    dp[15] = (unsigned int)(sp[15]+sp2[15]+1)>>1;
    sp+= lx2;
    sp2+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recvc( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp,*sp2;
  int j;

  sp = s;
  sp2 = s+lx2;
  dp = d;

  for (j=0; j<h; j++)
  {
    dp[0] = (unsigned int)(sp[0]+sp2[0]+1)>>1;
    dp[1] = (unsigned int)(sp[1]+sp2[1]+1)>>1;
    dp[2] = (unsigned int)(sp[2]+sp2[2]+1)>>1;
    dp[3] = (unsigned int)(sp[3]+sp2[3]+1)>>1;
    dp[4] = (unsigned int)(sp[4]+sp2[4]+1)>>1;
    dp[5] = (unsigned int)(sp[5]+sp2[5]+1)>>1;
    dp[6] = (unsigned int)(sp[6]+sp2[6]+1)>>1;
    dp[7] = (unsigned int)(sp[7]+sp2[7]+1)>>1;
    sp+= lx2;
    sp2+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recvo(unsigned char *s, int *d, int lx, int lx2, int addflag,
					int c, int xa, int xb, int ya, int yb)
{
  int *dp,*om;
  unsigned char *sp,*sp2;
  int i,j;

  sp = s;
  sp2 = s+lx2;
  dp = d;
  om = &OM[c][ya][0];

  if (!addflag) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] = (((unsigned int)(sp[i] + sp2[i]+1))>>1)*om[i];
      }
      sp+= lx2;
      sp2+= lx2;
      dp+= lx;
      om+= 8;
    }
  }
  else {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] += (((unsigned int)(sp[i] + sp2[i]+1))>>1)*om[i];
      }
      sp+= lx2;
      sp2+= lx2;
      dp+= lx;
      om+= 8;
    }
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::rec4( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp,*sp2;
  int j;
  unsigned int s1,s2,s3,s4;

  sp = s;
  sp2 = s+lx2;
  dp = d;
  for (j=0; j<h; j++)
  {
    s1=sp[0]; s3=sp2[0];
    dp[0] = (unsigned int)(s1+(s2=sp[1])+s3+(s4=sp2[1])+2)>>2;
    dp[1] = (unsigned int)(s2+(s1=sp[2])+s4+(s3=sp2[2])+2)>>2;
    dp[2] = (unsigned int)(s1+(s2=sp[3])+s3+(s4=sp2[3])+2)>>2;
    dp[3] = (unsigned int)(s2+(s1=sp[4])+s4+(s3=sp2[4])+2)>>2;
    dp[4] = (unsigned int)(s1+(s2=sp[5])+s3+(s4=sp2[5])+2)>>2;
    dp[5] = (unsigned int)(s2+(s1=sp[6])+s4+(s3=sp2[6])+2)>>2;
    dp[6] = (unsigned int)(s1+(s2=sp[7])+s3+(s4=sp2[7])+2)>>2;
    dp[7] = (unsigned int)(s2+(s1=sp[8])+s4+(s3=sp2[8])+2)>>2;
    dp[8] = (unsigned int)(s1+(s2=sp[9])+s3+(s4=sp2[9])+2)>>2;
    dp[9] = (unsigned int)(s2+(s1=sp[10])+s4+(s3=sp2[10])+2)>>2;
    dp[10] = (unsigned int)(s1+(s2=sp[11])+s3+(s4=sp2[11])+2)>>2;
    dp[11] = (unsigned int)(s2+(s1=sp[12])+s4+(s3=sp2[12])+2)>>2;
    dp[12] = (unsigned int)(s1+(s2=sp[13])+s3+(s4=sp2[13])+2)>>2;
    dp[13] = (unsigned int)(s2+(s1=sp[14])+s4+(s3=sp2[14])+2)>>2;
    dp[14] = (unsigned int)(s1+(s2=sp[15])+s3+(s4=sp2[15])+2)>>2;
    dp[15] = (unsigned int)(s2+sp[16]+s4+sp2[16]+2)>>2;
    sp+= lx2;
    sp2+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::rec4c( unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
  unsigned char *dp,*sp,*sp2;
  int j;
  unsigned int s1,s2,s3,s4;

  sp = s;
  sp2 = s+lx2;
  dp = d;
  for (j=0; j<h; j++)
  {
    s1=sp[0]; s3=sp2[0];
    dp[0] = (unsigned int)(s1+(s2=sp[1])+s3+(s4=sp2[1])+2)>>2;
    dp[1] = (unsigned int)(s2+(s1=sp[2])+s4+(s3=sp2[2])+2)>>2;
    dp[2] = (unsigned int)(s1+(s2=sp[3])+s3+(s4=sp2[3])+2)>>2;
    dp[3] = (unsigned int)(s2+(s1=sp[4])+s4+(s3=sp2[4])+2)>>2;
    dp[4] = (unsigned int)(s1+(s2=sp[5])+s3+(s4=sp2[5])+2)>>2;
    dp[5] = (unsigned int)(s2+(s1=sp[6])+s4+(s3=sp2[6])+2)>>2;
    dp[6] = (unsigned int)(s1+(s2=sp[7])+s3+(s4=sp2[7])+2)>>2;
    dp[7] = (unsigned int)(s2+sp[8]+s4+sp2[8]+2)>>2;
    sp+= lx2;
    sp2+= lx2;
    dp+= lx;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::rec4o(unsigned char *s, int *d, int lx, int lx2, 
					int addflag, int c, int xa , int xb,int ya, int yb)
{
  int *dp,*om;
  unsigned char *sp,*sp2;
  int i,j;

  sp = s;
  sp2 = s+lx2;
  dp = d;
  om = &OM[c][ya][0];

  if (!addflag) {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] = (((unsigned int)(sp[i]+sp2[i]+sp[i+1]+sp2[i+1]+2))>>2)*om[i];
      }
      sp+= lx2;
      sp2+= lx2;
      dp+= lx;
      om+= 8;
    }
  }
  else {
    for (j = ya; j < yb; j++) {
      for (i = xa; i < xb; i++) {
	dp[i] += (((unsigned int)(sp[i]+sp2[i]+sp[i+1]+sp2[i+1]+2))>>2)*om[i];
      }
      sp+= lx2;
      sp2+= lx2;
      dp+= lx;
      om += 8;
    }
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::recon_comp_obmc( unsigned char *src, unsigned char *dst,
							  int lx, int lx2, int comp, int w, int h,
							  int x, int y)
{
  int j,k;
  int xmb,ymb;
  int c8,t8,l8,r8;
  int xit,xib,xir,xil;
  int yit,yib,yir,yil;
  int vect,vecb,vecr,vecl;
  int nx[5],ny[5],xint[5],yint[5],xh[5],yh[5];
  int p[64],*pd;
  unsigned char *d,*s[5];

  xmb = (x>>4)+1;
  ymb = (y>>4)+1;

  c8 = (modemap[ymb][xmb] == MODE_INTER4V ? 1 : 0);
  t8 = (modemap[ymb-1][xmb] == MODE_INTER4V ? 1 : 0);
  l8 = (modemap[ymb][xmb-1] == MODE_INTER4V ? 1 : 0);
  r8 = (modemap[ymb][xmb+1] == MODE_INTER4V ? 1 : 0);

  switch (comp+1) {
  case 1:
    vect = (t8 ? 3 : 0) ; yit = ymb-1 ; xit = xmb;
    vecb = (c8 ? 3 : 0) ; yib = ymb   ; xib = xmb;
    vecl = (l8 ? 2 : 0) ; yil = ymb   ; xil = xmb-1;
    vecr = (c8 ? 2 : 0) ; yir = ymb   ; xir = xmb;
    if (ymb == 1) {
      yit = ymb;
      vect = (c8 ? 1 : 0);
    }
    if (xmb == 1) {
      xil = xmb;
      vecl = (c8 ? 1 : 0);
    }
    break;
  case 2:
    vect = (t8 ? 4 : 0) ; yit = ymb-1 ; xit = xmb;
    vecb = (c8 ? 4 : 0) ; yib = ymb   ; xib = xmb;
    vecl = (c8 ? 1 : 0) ; yil = ymb   ; xil = xmb;
    vecr = (r8 ? 1 : 0) ; yir = ymb   ; xir = xmb+1;
    if (ymb == 1) {
      yit = ymb;
      vect = (c8 ? 2 : 0);
    }
    if (xmb == mb_width) {
      xir = xmb;
      vecr = (c8 ? 2 : 0);
    }
    break;
  case 3:
    vect = (c8 ? 1 : 0) ; yit = ymb  ; xit = xmb;
    vecb = (c8 ? 3 : 0) ; yib = ymb  ; xib = xmb;
    vecl = (l8 ? 4 : 0) ; yil = ymb  ; xil = xmb-1;
    vecr = (c8 ? 4 : 0) ; yir = ymb  ; xir = xmb;
    if (xmb == 1) {
      xil = xmb;
      vecl = (c8 ? 3 : 0);
    }
    break;
  case 4:
    vect = (c8 ? 2 : 0) ; yit = ymb  ; xit = xmb;
    vecb = (c8 ? 4 : 0) ; yib = ymb  ; xib = xmb;
    vecl = (c8 ? 3 : 0) ; yil = ymb  ; xil = xmb;
    vecr = (r8 ? 3 : 0) ; yir = ymb  ; xir = xmb+1;
    if (xmb == mb_width) {
      xir = xmb;
      vecr = (c8 ? 4 : 0);
    }
    break;
  default:
    fprintf(stderr,"Illegal block number in recon_comp_obmc (recon.c)\n");
    exit(1);
    break;
  }

  nx[0] = MV[0][c8 ? comp + 1 : 0][ymb][xmb];
  ny[0] = MV[1][c8 ? comp + 1 : 0][ymb][xmb];
  
  nx[1] = MV[0][vect][yit][xit]; ny[1] = MV[1][vect][yit][xit];
  nx[2] = MV[0][vecb][yib][xib]; ny[2] = MV[1][vecb][yib][xib];
  nx[3] = MV[0][vecr][yir][xir]; ny[3] = MV[1][vecr][yir][xir];
  nx[4] = MV[0][vecl][yil][xil]; ny[4] = MV[1][vecl][yil][xil];

  for (k = 0; k < 5; k++) {
    xint[k] = nx[k]>>1;
    xh[k] = nx[k] & 1;
    yint[k] = ny[k]>>1;
    yh[k] = ny[k] & 1;
    s[k] = src + lx2*(y+yint[k]) + x + xint[k];
  }
  
  d = dst + lx*y + x;

  if (!xh[0] && !yh[0])
    reco(s[0],&p[0],8,lx2,0,0,0,8,0,8);
  else if (!xh[0] && yh[0])
    recvo(s[0],&p[0],8,lx2,0,0,0,8,0,8);
  else if (xh[0] && !yh[0])
    recho(s[0],&p[0],8,lx2,0,0,0,8,0,8);
  else /* if (xh[] && yh[]) */
    rec4o(s[0],&p[0],8,lx2,0,0,0,8,0,8);

  if (!xh[1] && !yh[1])
    reco(s[1],&p[0],8,lx2,1,1,0,8,0,4);
  else if (!xh[1] && yh[1])
    recvo(s[1],&p[0],8,lx2,1,1,0,8,0,4);
  else if (xh[1] && !yh[1])
    recho(s[1],&p[0],8,lx2,1,1,0,8,0,4);
  else /* if (xh[] && yh[]) */
    rec4o(s[1],&p[0],8,lx2,1,1,0,8,0,4);

  if (!xh[2] && !yh[2])
    reco(s[2]+(lx2<<2),&p[32],8,lx2,1,2,0,8,4,8);
  else if (!xh[2] && yh[2])
    recvo(s[2]+(lx2<<2),&p[32],8,lx2,1,2,0,8,4,8);
  else if (xh[2] && !yh[2])
    recho(s[2]+(lx2<<2),&p[32],8,lx2,1,2,0,8,4,8);
  else /* if (xh[] && yh[]) */
    rec4o(s[2]+(lx2<<2),&p[32],8,lx2,1,2,0,8,4,8);

  if (!xh[3] && !yh[3])
    reco(s[3],&p[0],8,lx2,1,3,4,8,0,8);
  else if (!xh[3] && yh[3])
    recvo(s[3],&p[0],8,lx2,1,3,4,8,0,8);
  else if (xh[3] && !yh[3])
    recho(s[3],&p[0],8,lx2,1,3,4,8,0,8);
  else /* if (xh[] && yh[]) */
    rec4o(s[3],&p[0],8,lx2,1,3,4,8,0,8);

  if (!xh[4] && !yh[4])
    reco(s[4],&p[0],8,lx2,1,4,0,4,0,8);
  else if (!xh[4] && yh[4])
    recvo(s[4],&p[0],8,lx2,1,4,0,4,0,8);
  else if (xh[4] && !yh[4])
    recho(s[4],&p[0],8,lx2,1,4,0,4,0,8);
  else /* if (xh[] && yh[]) */
    rec4o(s[4],&p[0],8,lx2,1,4,0,4,0,8);

  pd = &p[0];
  for (j = 0; j < 8; j++) {
    d[0] = (pd[0] + 4 )>>3;	
    d[1] = (pd[1] + 4 )>>3;	
    d[2] = (pd[2] + 4 )>>3;	
    d[3] = (pd[3] + 4 )>>3;	
    d[4] = (pd[4] + 4 )>>3;	
    d[5] = (pd[5] + 4 )>>3;	
    d[6] = (pd[6] + 4 )>>3;	
    d[7] = (pd[7] + 4 )>>3;	
    d += lx;
    pd += 8;
  }

}
//-----------------------------------------------------------------------------
int CH263Dec::decode_a_symbol(int cumul_freq[ ])
{

  length = high - low + 1;
  cum = (-1 + (code_value - low + 1) * cumul_freq[0]) / length;
  for (sacindex = 1; cumul_freq[sacindex] > cum; sacindex++);
  high = low - 1 + (length * cumul_freq[sacindex-1]) / cumul_freq[0];
  low += (length * cumul_freq[sacindex]) / cumul_freq[0];

  for ( ; ; ) {  
    if (high < q2) ;
    else if (low >= q2) {
      code_value -= q2; 
      low -= q2; 
      high -= q2;
    }
    else if (low >= q1 && high < q3) {
      code_value -= q1; 
      low -= q1; 
      high -= q1;
    }
    else {
      break;
    }
 
    low *= 2; 
    high = 2*high + 1;
    bit_out_psc_layer(); 
    code_value = 2*code_value + bit;
  }

  return (sacindex-1);
}
 
/*********************************************************************
 *
 * 	Name:		decoder_reset
 *
 *	Description:	Fills Decoder FIFO after a fixed word length
 *			string has been detected.
 *
 *	Input:		None
 *
 *	Returns:	Nothing
 *
 *	Side Effects:	Fills Arithmetic Decoder FIFO
 *
 *
 *********************************************************************/

void CH263Dec::decoder_reset( )
{
	int i;
	zerorun = 0;			/* clear consecutive zero's counter */
	code_value = 0;
	low = 0;
	high = top;
	for (i = 1;   i <= 16;   i++) {
		bit_out_psc_layer(); 
		code_value = 2 * code_value + bit;
	}
	TRACE("Arithmetic Decoder Reset \n");
}

/*********************************************************************
 *
 * 	Name:		bit_out_psc_layer
 *
 *	Description:	Gets a bit from the Encoded Stream, Checks for
 *			and removes any PSC emulation prevention bits
 *			inserted at the decoder, provides 'zeros' to the
 *			Arithmetic Decoder FIFO to allow it to finish 
 *			data prior to the next PSC. (Garbage bits)
 *
 *	Input:		None
 *
 *	Returns:	Nothing
 *
 *	Side Effects:	Gets a bit from the Input Data Stream
 *
 *
 *********************************************************************/

void CH263Dec::bit_out_psc_layer()
{
	if (showbits(17)!=1) { /* check for startcode in Arithmetic Decoder FIFO */
		
		bit = getbits(1);
		
		if(zerorun > 13) {	/* if number of consecutive zeros = 14 */	 
			if (!bit) {
				
				TRACE("PSC/GBSC, Header Data, or Encoded Stream Error \n");
				
				zerorun = 1;		
			}
			else { /* if there is a 'stuffing bit present */
				
				TRACE("Removing Startcode Emulation Prevention bit \n");
				bit = getbits(1); 	/* overwrite the last bit */	
				zerorun = !bit;	  	/* zerorun=1 if bit is a '0' */
			}
		}
		
		else { /* if consecutive zero's not exceeded 14 */
			
			if (!bit)
				zerorun++;
			else
				zerorun = 0;
		}
		
	} /* end of if !(showbits(17)) */
	
	else {
		bit = 0;
		TRACE("Startcode Found:Finishing Arithmetic Decoding using 'Garbage bits'\n");
	}
	
	
}
//-----------------------------------------------------------------------------
void CH263Dec::storeframe( unsigned char *src[], int frame)
{
	char outname[32];
	
	/* progressive */
	sprintf(outname,outputname,frame,'f');
	store_one(outname,src,0,coded_picture_width,vertical_size);
}
//-----------------------------------------------------------------------------
/* store one frame or one field */
void CH263Dec::store_one( char *outname, unsigned char *src[], int offset, int incr, int height)
{
  switch (outtype)
  {
  case T_YUV:
    store_yuv(outname,src,offset,incr,height);
    break;
  case T_YUV_CONC:
    store_yuv_append(outname,src,offset,incr,height);
    break;
  case T_SIF:
    store_sif(outname,src,offset,incr,height);
    break;
  case T_TGA:
    store_ppm_tga(outname,src,offset,incr,height,1);
    break;
  case T_PPM:
    store_ppm_tga(outname,src,offset,incr,height,0);
    break;
#ifdef DISPLAY
  case T_X11:
    dither(src);
    break;
#endif
  default:
    break;
  }
}
//-----------------------------------------------------------------------------
/* separate headerless files for y, u and v */
void CH263Dec::store_yuv( char *outname, unsigned char *src[], int offset, int incr, int height)
{
  int hsize;
  char tmpname[32];

  hsize = horizontal_size;

  sprintf(tmpname,"%s.Y",outname);
  store_yuv1(tmpname,src[0],offset,incr,hsize,height,0);

  offset>>=1; incr>>=1; hsize>>=1;
  height>>=1;

  sprintf(tmpname,"%s.U",outname);
  store_yuv1(tmpname,src[1],offset,incr,hsize,height,0);

  sprintf(tmpname,"%s.V",outname);
  store_yuv1(tmpname,src[2],offset,incr,hsize,height,0);
}
//-----------------------------------------------------------------------------
/* concatenated headerless file for y, u and v */
void CH263Dec::store_yuv_append( char *outname, unsigned char *src[], 
							   int offset, int incr, int height)
{
  int hsize;
  hsize = horizontal_size;

  store_yuv1(outname,src[0],offset,incr,hsize,height,1);

  offset>>=1; incr>>=1; hsize>>=1;
  height>>=1;

  store_yuv1(outname,src[1],offset,incr,hsize,height,1);
  store_yuv1(outname,src[2],offset,incr,hsize,height,1);
}
//-----------------------------------------------------------------------------
/* auxiliary routine */
void CH263Dec::store_yuv1( char *name,unsigned char *src,int offset,
						 int incr, int width, int height, int append)
{
  int i, j;
  unsigned char *p;

  if (append) {
    if ((outfile =_open(name,O_APPEND|O_WRONLY|O_BINARY,0666))==-1) {
      sprintf(errortext,"Couldn't append to %s\n",name);
      error(errortext);
    }
  }
  else {
    if ((outfile = open(name,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1) {
      sprintf(errortext,"Couldn't create %s\n",name);
      error(errortext);
    }
  }
  if (!quiet)
    fprintf(stderr,"saving %s\n",name);

  optr=obfr;

  for (i=0; i<height; i++) {
    p = src + offset + incr*i;
    for (j=0; j<width; j++)
      putbyte(*p++);
  }

  if (optr!=obfr)
    _write(outfile,obfr,optr-obfr);

  _close(outfile);
}
//-----------------------------------------------------------------------------
 /* store as headerless file in U,Y,V,Y format */
void CH263Dec::store_sif ( char *outname,unsigned char *src[],int offset,
						 int incr, int height)
{
  int i,j;
  unsigned char *py, *pu, *pv;
  static unsigned char *u422, *v422;

  if (!u422) {
    if (!(u422 = (unsigned char *)malloc((coded_picture_width>>1)
					 *coded_picture_height)))
      error("malloc failed");
    if (!(v422 = (unsigned char *)malloc((coded_picture_width>>1)
					 *coded_picture_height)))
      error("malloc failed");
  }
  
  conv420to422(src[1],u422);
  conv420to422(src[2],v422);

  strcat(outname,".SIF");

  if (!quiet)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(errortext,"Couldn't create %s\n",outname);
    error(errortext);
  }

  optr = obfr;

  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u422 + (offset>>1) + (incr>>1)*i;
    pv = v422 + (offset>>1) + (incr>>1)*i;

    for (j=0; j<horizontal_size; j+=2)
    {
      putbyte(*pu++);
      putbyte(*py++);
      putbyte(*pv++);
      putbyte(*py++);
    }
  }

  if (optr!=obfr)
    write(outfile,obfr,optr-obfr);

  close(outfile);
}
//-----------------------------------------------------------------------------
/* store as PPM (PBMPLUS) or uncompressed Truevision TGA ('Targa') file */
void CH263Dec::store_ppm_tga( char *outname, unsigned char *src[],
							int offset,int incr, int height, int tgaflag)
{
  int i, j;
  int y, u, v, r, g, b;
  int crv, cbu, cgu, cgv;
  unsigned char *py, *pu, *pv;
  static unsigned char tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
  char header[32];
  static unsigned char *u422, *v422, *u444, *v444;

  if (!u444) {
    if (!(u422 = (unsigned char *)malloc((coded_picture_width>>1)
					 *coded_picture_height)))
      error("malloc failed");
    if (!(v422 = (unsigned char *)malloc((coded_picture_width>>1)
					 *coded_picture_height)))
      error("malloc failed");

    if (!(u444 = (unsigned char *)malloc(coded_picture_width
					 *coded_picture_height)))
      error("malloc failed");

    if (!(v444 = (unsigned char *)malloc(coded_picture_width
					 *coded_picture_height)))
      error("malloc failed");
  }

  conv420to422(src[1],u422);
  conv420to422(src[2],v422);
  conv422to444(u422,u444);
  conv422to444(v422,v444);

  strcat(outname,tgaflag ? ".tga" : ".ppm");

  if (!quiet)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
    {
      sprintf(errortext,"Couldn't create %s\n",outname);
      error(errortext);
    }

  optr = obfr;

  if (tgaflag) {
    /* TGA header */
    for (i=0; i<12; i++)
      putbyte(tga24[i]);

    putword(horizontal_size); putword(height);
    putbyte(tga24[12]); putbyte(tga24[13]);
  }
  else {
    /* PPM header */
    sprintf(header,"P6\n%d %d\n255\n",horizontal_size,height);

    for (i=0; header[i]!=0; i++)
      putbyte(header[i]);
  }

  /* matrix coefficients */
  crv = convmat[matrix_coefficients][0];
  cbu = convmat[matrix_coefficients][1];
  cgu = convmat[matrix_coefficients][2];
  cgv = convmat[matrix_coefficients][3];
  
  for (i=0; i<height; i++) {
    py = src[0] + offset + incr*i;
    pu = u444 + offset + incr*i;
    pv = v444 + offset + incr*i;

    for (j=0; j<horizontal_size; j++) {
      u = *pu++ - 128;
      v = *pv++ - 128;
      y = 76309 * (*py++ - 16); /* (255/219)*65536 */
      r = clp[(y + crv*v + 32768)>>16];
      g = clp[(y - cgu*u - cgv*v + 32768)>>16];
      b = clp[(y + cbu*u + 32786)>>16];

      if (tgaflag) {
	putbyte(b); putbyte(g); putbyte(r);
      }
      else {
	putbyte(r); putbyte(g); putbyte(b);
      }
    }
  }

  if (optr!=obfr)
    write(outfile,obfr,optr-obfr);

  close(outfile);
}
//-----------------------------------------------------------------------------
void CH263Dec::putbyte( int c)
{
  *optr++ = c;

  if (optr == obfr+OBFRSIZE)
  {
    write(outfile,obfr,OBFRSIZE);
    optr = obfr;
  }
}
//-----------------------------------------------------------------------------
void CH263Dec::putword(int w)
{
  putbyte(w); putbyte(w>>8);
}
//-----------------------------------------------------------------------------
/* horizontal 1:2 interpolation filter */
void CH263Dec::conv422to444(unsigned char *src, unsigned char *dst)
{
  int i, i2, w, j, im3, im2, im1, ip1, ip2, ip3;

  w = coded_picture_width>>1;

  for (j=0; j<coded_picture_height; j++) {
    for (i=0; i<w; i++) {

      i2 = i<<1;
      im3 = (i<3) ? 0 : i-3;
      im2 = (i<2) ? 0 : i-2;
      im1 = (i<1) ? 0 : i-1;
      ip1 = (i<w-1) ? i+1 : w-1;
      ip2 = (i<w-2) ? i+2 : w-1;
      ip3 = (i<w-3) ? i+3 : w-1;

      /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
      dst[i2] =   clp[(int)(  5*src[im3]
			      -21*src[im2]
			      +70*src[im1]
			      +228*src[i]
			      -37*src[ip1]
			      +11*src[ip2]+128)>>8];

      dst[i2+1] = clp[(int)(  5*src[ip3]
			      -21*src[ip2]
			      +70*src[ip1]
			      +228*src[i]
			      -37*src[im1]
			      +11*src[im2]+128)>>8];
    }
    src+= w;
    dst+= coded_picture_width;
  }
}
//-----------------------------------------------------------------------------
/* vertical 1:2 interpolation filter */
void CH263Dec::conv420to422(unsigned char *src, unsigned char *dst)
{
  int w, h, i, j, j2;
  int jm3, jm2, jm1, jp1, jp2, jp3;

  w = coded_picture_width>>1;
  h = coded_picture_height>>1;

  /* intra frame */
  for (i=0; i<w; i++) {
    for (j=0; j<h; j++) {
      j2 = j<<1;
      jm3 = (j<3) ? 0 : j-3;
      jm2 = (j<2) ? 0 : j-2;
      jm1 = (j<1) ? 0 : j-1;
      jp1 = (j<h-1) ? j+1 : h-1;
      jp2 = (j<h-2) ? j+2 : h-1;
      jp3 = (j<h-3) ? j+3 : h-1;
      
      /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
      /* New FIR filter coefficients (*256): 3 -16 67 227 -32 7 */
      dst[w*j2] =     clp[(int)(  3*src[w*jm3]
				  -16*src[w*jm2]
				  +67*src[w*jm1]
				  +227*src[w*j]
				  -32*src[w*jp1]
				  +7*src[w*jp2]+128)>>8];

      dst[w*(j2+1)] = clp[(int)(  3*src[w*jp3]
				  -16*src[w*jp2]
				  +67*src[w*jp1]
				  +227*src[w*j]
				  -32*src[w*jm1]
				  +7*src[w*jm2]+128)>>8];
    }
    src++;
    dst++;
  }
}

//-----------------------------------------------------------------------------
void CH263Dec::rec(unsigned char *s, unsigned char *d, int lx, int lx2, int h)
{
	int j;
	
	for (j=0; j<h; j++)
	{
		d[0]  = s[0];		d[1]  = s[1];
		d[2]  = s[2];		d[3]  = s[3];
		d[4]  = s[4];		d[5]  = s[5];
		d[6]  = s[6];		d[7]  = s[7];
		d[8]  = s[8];		d[9]  = s[9];
		d[10] = s[10];		d[11] = s[11];
		d[12] = s[12];		d[13] = s[13];
		d[14] = s[14];		d[15] = s[15];
		s+= lx2;		d+= lx;
	}
}

//-----------------------------------------------------------------------------
BOOL CH263Dec::Play263File()
{

	WaitForSingleObject(hPlay,INFINITE);

	if ( gusState  == PLAYING ) {

		if ( getheader())
		{
			if ( first == 1) {
				initdecoder();
                
				init_dither(nDitherType);
                if (nDitherType==8)
					init_display();
				else
				{
					init_dib();
					init_dither_tab();
				}
				first = 5;
			}
			getpicture ( &FrameNum);

			FrameNum ++;
			tr_framenum ++;
			return TRUE;
		}
	    else 
		{
			if ( FrameNum != 0)
			{
				dither(refframe);
			}
			return FALSE;
		}
	}
	return FALSE;
}
//-----------------------------------------------------------------------------
BOOL CH263Dec::PlayStart()
{
	if ( gusState == STOP) {
		play_movie();
		first = 1;
		FrameNum = tr_framenum = 0;
		return TRUE;
	}
	else 
		return FALSE;
}
//-----------------------------------------------------------------------------
void CH263Dec::fillbfr()
{
	int l;
		
	ld->inbfr[0] = ld->inbfr[8];
	ld->inbfr[1] = ld->inbfr[9];
	ld->inbfr[2] = ld->inbfr[10];
	ld->inbfr[3] = ld->inbfr[11];
	
	if (ld->rdptr >= ld->rdbfr + MAX_BUFF )
	{
		for (l=0; l<MAX_BUFF;)
		{
			unsigned char data;

			if(buffer.GetData(data))
			{
				*(ld->rdbfr + l) = data;
			}
			else
			{
				WaitForSingleObject(hFullData,1);
				continue;
			}

			l++;
		}
		
		ld->rdptr = ld->rdbfr;
/*
		if ( l < MAX_BUFF )
		{
			if (l<0)
				l=0;

			while ( l<MAX_BUFF )	// Add recogniable sequence end code
			{
				ld->rdbfr[l++] = 0;
				ld->rdbfr[l++] = 0;
				ld->rdbfr[l++] = (1<<7) | (SEC<<2);
			}
		}
*/
	}

	for (l=0; l<8; l++)
		ld->inbfr[l+4] = ld->rdptr[l];

	ld->rdptr+=8;
	ld->incnt+=64;
}						   
//-----------------------------------------------------------------------------

long int crv_tab[256];
long int cbu_tab[256];
long int cgu_tab[256];

long int cgv_tab[256];
long int tab_76309[256];
//-----------------------------------------------------------------------------

void CH263Dec::init_dither_tab()
{
  long int crv,cbu,cgu,cgv;
  int i;   
  
  crv = 104597; cbu = 132201;  /* fra matrise i global.h */
  cgu = 25675;  cgv = 53279;
  
  for (i = 0; i < 256; i++) {
    crv_tab[i] = (i-128) * crv;
    cbu_tab[i] = (i-128) * cbu;
    cgu_tab[i] = (i-128) * cgu;
    cgv_tab[i] = (i-128) * cgv;
    tab_76309[i] = 76309*(i-16);
  }
}

          
/**********************************************************************
 *
 *	Name:	         ConvertYUVtoRGB	
 *	Description:     Converts YUV image to RGB (packed mode)
 *	
 *	Input:	         pointer to source luma, Cr, Cb, destination,
 *                       image width and height
 *	Returns:       
 *	Side effects:
 *
 *
 ***********************************************************************/

void CH263Dec::ConvertYUVtoRGB(unsigned char *src0,unsigned char *src1,
							   unsigned char *src2,unsigned char *dst_ori,
							   int width,int height)
{       
	extern long int crv_tab[];
	extern long int cbu_tab[];
	extern long int cgu_tab[];
	
	extern long int cgv_tab[];
	extern long int tab_76309[];
	
	int y11,y21;
	int y12,y22;
	int y13,y23;
	int y14,y24;
	int u,v; 
	int i,j;
	int c11, c21, c31, c41;
	int c12, c22, c32, c42;
	unsigned int DW;
	unsigned int *id1, *id2;
	unsigned char *py1,*py2,*pu,*pv;
	unsigned char *d1, *d2;
	
	d1 = dst_ori;
	d1 += width*height*3 - width*3;
	d2 = d1 - width*3;
	
	py1 = src0; pu = src1; pv = src2;
	py2 = py1 + width;
	
	id1 = (unsigned int *)d1;
	id2 = (unsigned int *)d2;
	
	for (j = 0; j < height; j += 2) { 
		/* line j + 0 */
		for (i = 0; i < width; i += 4) {
			u = *pu++;
			v = *pv++;
			c11 = crv_tab[v];
			c21 = cgu_tab[u];
			c31 = cgv_tab[v];
			c41 = cbu_tab[u];
			u = *pu++;
			v = *pv++;
			c12 = crv_tab[v];
			c22 = cgu_tab[u];
			c32 = cgv_tab[v];
			c42 = cbu_tab[u];
			
			y11 = tab_76309[*py1++]; /* (255/219)*65536 */
			y12 = tab_76309[*py1++];
			y13 = tab_76309[*py1++]; /* (255/219)*65536 */
			y14 = tab_76309[*py1++];
			
			y21 = tab_76309[*py2++];
			y22 = tab_76309[*py2++];
			y23 = tab_76309[*py2++];
			y24 = tab_76309[*py2++];
			
			/* RGBR*/
			DW = ((clp[(y11 + c41)>>16])) |
				((clp[(y11 - c21 - c31)>>16])<<8) |
				((clp[(y11 + c11)>>16])<<16) |  
				((clp[(y12 + c41)>>16])<<24);
			*id1++ = DW;
			
			/* GBRG*/
			DW = ((clp[(y12 - c21 - c31)>>16])) |
				((clp[(y12 + c11)>>16])<<8) |  
				((clp[(y13 + c42)>>16])<<16) |
				((clp[(y13 - c22 - c32)>>16])<<24);
			*id1++ = DW;
			
			/* BRGB*/
			DW = ((clp[(y13 + c12)>>16])) |  
				((clp[(y14 + c42)>>16])<<8) |
				((clp[(y14 - c22 - c32)>>16])<<16) |
				((clp[(y14 + c12)>>16])<<24);  
			*id1++ = DW;
			
			/* RGBR*/
			DW = ((clp[(y21 + c41)>>16])) |
				((clp[(y21 - c21 - c31)>>16])<<8) |
				((clp[(y21 + c11)>>16])<<16) |  
				((clp[(y22 + c41)>>16])<<24);
			*id2++ = DW;
			
			/* GBRG*/
			DW = ((clp[(y22 - c21 - c31)>>16])) |
				((clp[(y22 + c11)>>16])<<8) |  
				((clp[(y23 + c42)>>16])<<16) |
				((clp[(y23 - c22 - c32)>>16])<<24);
			*id2++ = DW;
			
			/* BRGB*/
			DW = ((clp[(y23 + c12)>>16])) |  
				((clp[(y24 + c42)>>16])<<8) |
				((clp[(y24 - c22 - c32)>>16])<<16) |
				((clp[(y24 + c12)>>16])<<24);  
			*id2++ = DW;
		}
		id1 -= (9 * width)>>2;
		id2 -= (9 * width)>>2;
		py1 += width;
		py2 += width;
	}           
}