/******************************************************************************
	
	  H263Dec.h: interface for the CH263Dec class.

******************************************************************************/

#ifndef _H263DEC_H_
#define _H263DEC_H_

#include <vfw.h>
 
#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#define T_YUV      0
#define T_SIF      1
#define T_TGA      2
#define T_PPM      3
#define T_X11      4
#define T_YUV_CONC 5

#define W1 2841 /* 2048*sqrt(2)*cos(1*pi/16) */
#define W2 2676 /* 2048*sqrt(2)*cos(2*pi/16) */
#define W3 2408 /* 2048*sqrt(2)*cos(3*pi/16) */
#define W5 1609 /* 2048*sqrt(2)*cos(5*pi/16) */
#define W6 1108 /* 2048*sqrt(2)*cos(6*pi/16) */
#define W7 565  /* 2048*sqrt(2)*cos(7*pi/16) */

#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
#  define PI 3.14159265358979323846
# endif
#endif

#define   q1    16384
#define   q2    32768
#define   q3    49152
#define   top   65535
#define OBFRSIZE 4096


/* From sim.h */
#define PSC								1
#define PSC_LENGTH						17
#define SEC                             31

#define MODE_INTER                      0
#define MODE_INTER_Q                    1
#define MODE_INTER4V                    2
#define MODE_INTRA                      3
#define MODE_INTRA_Q                    4

#define PBMODE_NORMAL                   0
#define PBMODE_MVDB                     1
#define PBMODE_CBPB_MVDB                2

#define ESCAPE                          7167
#define ESCAPE_INDEX                    102

#define PCT_INTER                       1
#define PCT_INTRA                       0
#define ON                              1
#define OFF                             0

#define SF_SQCIF                        1  /* 001 */
#define SF_QCIF                         2  /* 010 */
#define SF_CIF                          3  /* 011 */
#define SF_4CIF                         4  /* 100 */
#define SF_16CIF                        5  /* 101 */

// Corresponds to CIF
#define MBC_MAX                         22
#define MBR_MAX                         18


#define NO_VEC                          999

struct VLCtabI{
  int val, len;
};

struct RunCoef{
  int val, run, sign;
};

class CH263;

class CH263Dec 
{
public :
	CH263 *m_H263;
	HANDLE hDec;
	CH263Dec (CWnd *pParent);
	virtual ~CH263Dec();

	void initbits (void);
	void fillbfr (void);
	unsigned int showbits (int n);
	unsigned int getbits1 (void);
	void flushbits (int n);
	unsigned int getbits (int n);
	
	void initdecoder ( void);
	void option      ( int *argcp, char **argvp[]);
	int play_movie ( void);
	
	void getblock (int comp, int mode);
	void get_sac_block (int comp, int mode);

	
	int getheader (void);
	void startcode (void);
	
	void getpicture (int *framenum);
	void putlast (int framenum);

	
	int getTMNMV (void);
	int getMCBPC (void);
	int getMODB (void);
	int getMCBPCintra (void);
	int getCBPY (void);

	
	void idct (short *block);
	void init_idct (void);

	
	void init_idctref (void);
	void idctref (short *block);

	
	void error (char *text);
	void printbits (int code, int bits, int len);
	void doframerate();

	
	void reconstruct (int bx, int by, int P, int bdx, int bdy);

	
	void storeframe (unsigned char *src[], int frame);

	
	int decode_a_symbol (int cumul_freq[ ]);
	void decoder_reset( );

	#define DISPLAY
	#ifdef DISPLAY
	
	void init_display ( void);
	void exit_display (void);
	void dither (unsigned char *src[]);
	void init_dither (int bpp);
	#endif


	void display_image ( unsigned char *dithered_image);
	void ditherframe (unsigned char *src[]);

	/* local data */

	unsigned char ytab[256+16];
	unsigned char utab[128+16];
	unsigned char vtab[128+16];
	unsigned char pixel[256];

	void (*dither_frame)( unsigned char **);
	RunCoef vlc_word_decode (int symbol_word, int *last); 
	RunCoef Decode_Escape_Char (int intra, int *last);
	int DecodeTCoef (int position, int intra);
	void getpicturehdr (void);
	void getMBs (int framenum);
	void clearblock (int comp);
	int motion_decode (int vec,int pmv);
	int find_pmv (int x, int y, int block, int comp);
	void addblock (int comp, int bx, int by,int addflag);
	void reconblock_b (int comp,int bx,int by,int mode,int bdx, int bdy);
	void find_bidir_limits (int vec, int *start, int*stop, int nhv);
	void find_bidir_chroma_limits (int vec, int *start, int*stop);
	void make_edge_image (unsigned char *src, unsigned char *dst, int width, int height, int edge);
	void interpolate_image (unsigned char *in, unsigned char *out, int width, int height);
	void options     (int *argcp, char **argvp[]);
	int  getval      (char *argv[]);
	void recon_comp (unsigned char *src, unsigned char *dst,
		int lx, int lx2, int w, int h, int x, int y, int dx, int dy, int flag);
	void recon_comp_obmc (unsigned char *src, unsigned char *dst,int lx,int lx2,int comp,int w,int h,int x,int y);
	void rec (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void recc (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void reco (unsigned char *s, int *d, int lx, int lx2, int addflag,int c, int xa, int xb, int ya, int yb);
	void rech (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void rechc (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void recho (unsigned char *s, int *d, int lx, int lx2, int addflag,int c, int xa, int xb, int ya, int yb);
	void recve (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void recvc (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void recvo (unsigned char *s, int *d, int lx, int lx2, int addflag,int c, int xa, int xb, int ya, int yb);
	void rec4 (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void rec4c (unsigned char *s, unsigned char *d, int lx, int lx2, int h);
	void rec4o (unsigned char *s, int *d, int lx, int lx2, int addflag,int c, int xa, int xb, int ya, int yb);
	
	/* private prototypes */
	void idctrow (short *blk);
	void idctcol (short *blk);
	void bit_out_psc_layer();

	void store_one (char *outname, unsigned char *src[],
		int offset, int incr, int height);
	void store_yuv (char *outname, unsigned char *src[],
		int offset, int incr, int height);
	void store_yuv_append (char *outname, unsigned char *src[],
		int offset, int incr, int height);
	void store_sif (char *outname, unsigned char *src[],
		int offset, int incr, int height);
	void store_ppm_tga (char *outname, unsigned char *src[],
		int offset, int incr, int height, int tgaflag);
	void store_yuv1 (char *name, unsigned char *src,
		int offset, int incr, int width, int height, int append);
	void putbyte (int c);
	void putword (int w);
	void conv422to444 (unsigned char *src, unsigned char *dst);
	void conv420to422 (unsigned char *src, unsigned char *dst);

//----------------------------------------------------
	void init_dib();
	void close_dib();
	void init_dither_tab();
	void ConvertYUVtoRGB(unsigned char *src0,unsigned char *src1,
						 unsigned char *src2,unsigned char *dst_ori,
						 int width,int height);

	/* global variables */

	int quiet;
	char errortext[256];
	unsigned char *refframe[3],*oldrefframe[3],*bframe[3],*newframe[3];
	unsigned char *edgeframe[3], *edgeframeorig[3], *exnewframe[3];
	int MV[2][5][MBR_MAX+1][MBC_MAX+2];
	int modemap[MBR_MAX+1][MBC_MAX+2];
	unsigned char *clp;
	int horizontal_size,vertical_size,mb_width,mb_height;
	int coded_picture_width, coded_picture_height;
	int chrom_width,chrom_height,blk_cnt;
	int pict_type,newgob;
	int mv_outside_frame,syntax_arith_coding,adv_pred_mode,pb_frame;
	int long_vectors;
	int fault,expand;
	int verbose;
	int refidct;
	int matrix_coefficients;
	int temp_ref, quant, source_format, tr_framenum;
	#ifdef USE_TIME
	int framerate;
	struct timeval tftarget;
	#endif
	int trd, trb, bscan, bquant;
	/* output */
	char *outputname;
	int outtype;

	struct ld {
		/* bit input */
		unsigned char rdbfr[2051];
		unsigned char *rdptr;
		unsigned char inbfr[16];
		int incnt;
		int bitcnt;
		/* block data */
		short block[12][64];
	} base,*ld;

	int m_MBC, m_MBR ;

	int nDitherType;
	HDC myDC;
	unsigned char *dithered_image;
	short iclip[1024]; /* clipping table */
	short *iclp;
	double c[8][8];
	long low, high, code_value, bit, length, sacindex, cum, zerorun;
	unsigned char obfr[OBFRSIZE];
	unsigned char *optr;
	int outfile;

	HDRAWDIB hDrawDib;
	BITMAPINFOHEADER biHeader;
public:
	int m_BuffIndex;
	unsigned char * m_InputBuff1;
	unsigned char * m_InputBuff2;
	
	BOOL PlayStart();
	BOOL Play263File();
	int first;
	int FrameNum;
	
	int m_x, m_y, m_dx, m_dy;
};
			  
#endif // _H263DEC_H_
