/*
 * H263Cont.cpp : Defines the initialization routines for CH263dec,CH263enc.
*/

#include "stdafx.h"
#include "H263Cont.h"

#include "H263Enc.h"
#include "H263Dec.h"
#include "SCTPServerDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


UINT H263_Decode(LPVOID vParam);

// Output
unsigned char outBuffStream[IMAGE_BUFFER_SIZE];
int BuffOffset = 0;
static HWND hMain;

PictImage *prev_image = NULL;
PictImage *curr_image = NULL;
PictImage *curr_recon = NULL;
PictImage *prev_recon = NULL;

// PB-frame specific 
PictImage *B_recon = NULL;
PictImage *B_image = NULL;

unsigned char *image;
FILE *cleared;

float mean_frame_rate,  ref_frame_rate;
float frame_rate, seconds;

int first_loop_finished=0;
int total_frames_passed, PPFlag = 0, targetrate;
int frames,bframes,pframes,wcopies,icopies, write_repeated,pdist=0,bdist=0;
int start, end, frame_no, writediff;  
int first_frameskip, chosen_frameskip, orig_frameskip, frameskip;
int QP,QPI;

char seqfilename[80];
char *diff_filename = DEF_DIFFILENAME;

#ifndef OFFLINE_RATE_CONTROL
	float DelayBetweenFramesInSeconds;
	int CommBacklog;
#else
	PictImage *stored_image = NULL;
	int start_rate_control;
#endif

extern int arith_used;

Pict	*pic;
Bits	*bits;
Bits	*total_bits;
Bits	*intra_bits;
Results *res;
Results *total_res;
Results *b_res;

//-----------------------------------------------------------------------------
H263Cont::H263Cont(CWnd *pWnd)
{
	initialCoder(pWnd->m_hWnd);
	InitDecoder(pWnd);
}
//-----------------------------------------------------------------------------
H263Cont::~H263Cont()
{
	CloseCoder();
	CloseDecoder();
}
//-----------------------------------------------------------------------------
void H263Cont::initialCoder(HWND hWnd)
{
	hMain = hWnd;

	pic			= new Pict;
	bits		= new Bits;
	total_bits	= new Bits;
	intra_bits	= new Bits;
	res			= new Results;
	total_res	= new Results;
	b_res		= new Results;

    headerlength = DEF_HEADERLENGTH;

#ifndef FASTIDCT
	init_idctref();
#endif

	// Default variable values 
	advanced					= DEF_ADV_MODE;
	syntax_arith_coding			= DEF_SAC_MODE;
	pic->unrestricted_mv_mode	= DEF_UMV_MODE;
	mv_outside_frame			= DEF_UMV_MODE || DEF_ADV_MODE;
	long_vectors				= DEF_UMV_MODE;
	pb_frames					= DEF_PBF_MODE;
	
	QP					= DEF_INTER_QUANT;
	QPI					= DEF_INTRA_QUANT; 
	pic->BQUANT			= DEF_BQUANT; 
	pic->source_format	= DEF_CODING_FORMAT;
	
	ref_frame_rate		= (float)DEF_REF_FRAME_RATE;
	chosen_frameskip	= DEF_FRAMESKIP + 1;
	orig_frameskip		= DEF_ORIG_SKIP + 1;

#ifdef OFFLINE_RATE_CONTROL
	start_rate_control	= DEF_START_RATE_CONTROL;
#else
	pic->target_frame_rate = (float)DEF_TARGET_FRAME_RATE;
#endif
	
	strcpy(seqfilename, "test.yuv");

	writediff		= DEF_WRITE_DIFF;
	write_repeated		= DEF_WRITE_REPEATED;
	pic->seek_dist		= DEF_SEEK_DIST;
	pic->use_gobsync	= DEF_INSERT_SYNC;
	
	start				= DEF_START_FRAME;
	end					= DEF_STOP_FRAME;
	
	targetrate = 0; 

	//* default is variable bit rate (fixed quantizer) will be used 
	frames	= 0;
	pframes	= 0;
	bframes	= 0;
	total_frames_passed = 0;
	
	wcopies	= icopies = 1;
	
	pic->PB	= 0;
	pic->TR	= 0;
	pic->QP_mean = (float)0.0;
	
	pels  = 176;
	lines = 144;
	cpels = pels/2;

#ifndef OFFLINE_RATE_CONTROL
	// rate control varialbes
	pic->bit_rate = targetrate;
 	pic->src_frame_rate = (int)(ref_frame_rate / orig_frameskip);
	DelayBetweenFramesInSeconds = (float)1.0/(float)pic->src_frame_rate;
	InitializeRateControl();
#endif

	// Initialize bitcounters
	initbits();

	frame_rate = ref_frame_rate / (float)(orig_frameskip * chosen_frameskip);

#ifndef OFFLINE_RATE_CONTROL
	if (pic->bit_rate == 0)
		printf("Encoding frame rate : %.2f\n", frame_rate);
	else
		printf("Encoding frame rate : variable");
#endif
}

//-----------------------------------------------------------------------------
void H263Cont::CloseCoder()
{
	// Free memory
	FreeImage(curr_recon);
	FreeImage(curr_image);

	delete pic;
	delete bits;
	delete total_bits;
	delete intra_bits;
	delete res;
	delete total_res;
	delete b_res;
}

extern unsigned char *yuvimage;
void H263Cont::StartEncoder(unsigned char *image)
{
	static int num=0;

	yuvimage = image;
	if ( num % 1 == 0)
 		IntraCoding(num);
	else
		InterCoding(num);

	num++;

}

//-----------------------------------------------------------------------------
void H263Cont::IntraCoding(int frame_num)
{
	// Intra image
	image = ReadImage(seqfilename, frame_num, headerlength);
	curr_image = FillImage(image);
	
	pic->picture_coding_type = PCT_INTRA;

	int position = 40;
	if( position < 20)
		QP = QPI = 2;
	else
		QP = QPI = position/10 +1;
	
	pic->QUANT = QPI;

	curr_recon = CodeOneIntra(curr_image, QPI, bits, pic);

	bits->header += alignbits();

	finishFrame();

	pic->picture_coding_type = PCT_INTER;
	pic->QUANT = QP;

	FreeImage(curr_image);//Ӹ  ޷θ  
	FreeImage(curr_recon);
}
//-----------------------------------------------------------------------------
void H263Cont::InterCoding(int frame_num)
{
	bdist = chosen_frameskip;

	prev_image = curr_image;
	prev_recon = curr_recon;

	// Set QP to pic->QUANT from previous encoded picture
	QP = pic->QUANT;

	image = ReadImage(seqfilename, frame_num, headerlength);
	curr_image = FillImage(image);

	pic->PB = 0;

	frames  += (pic->PB ? 2:1);
	bframes += (pic->PB? 1:0);
	pframes += 1;

	curr_recon = InitImage(pels*lines);
	
	CodeOneOrTwo( curr_image, B_image, prev_image, prev_recon,
		QP, (bdist+pdist)*orig_frameskip, bits, pic, B_recon, curr_recon);


	bits->header += alignbits();

	FreeImage(prev_image);
	FreeImage(prev_recon);

	finishFrame();
}
//-----------------------------------------------------------------------------
/*    */
void H263Cont::finishFrame()
{
	PostMessage(hMain, WM_FINISH_CODER, (WPARAM)outBuffStream, (LPARAM)BuffOffset);
	BuffOffset = 0;
}
//-----------------------------------------------------------------------------
/* ڴ ʱȭ Ѵ */
void H263Cont::InitDecoder(CWnd *pParent)
{
	h263Dec = new CH263Dec(pParent);
	h263Dec->PlayStart();
}
//-----------------------------------------------------------------------------
void H263Cont::PlayFrame()
{
	h263Dec->Play263File();
}
//-----------------------------------------------------------------------------
void H263Cont::SetDisplayWindow(int x, int y, int dx, int dy)
{
	h263Dec->m_x  = x;
	h263Dec->m_y  = y;
	h263Dec->m_dx = dx;
	h263Dec->m_dy = dy;
}
//-----------------------------------------------------------------------------
void H263Cont::CloseDecoder()
{
	delete h263Dec;
}
//-----------------------------------------------------------------------------
/* ڴ  ƾ */
bool bDoDecode = false;

UINT H263_Decode(LPVOID vParam)
{
	if (bDoDecode) return 0;
	bDoDecode = true;
	while (bDoDecode)
		((H263Cont *)vParam)->PlayFrame();

	return TRUE;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void H263Cont::StartDecoder()
{
	AfxBeginThread(H263_Decode, this, THREAD_PRIORITY_LOWEST);
}
//-----------------------------------------------------------------------------

#define _X 176
#define _Y 144
#define _B 3
#define IMAGESIZE (_X * _Y * _B)

unsigned char * H263Cont::MakeYUV(unsigned char *pimage, int sx, int sy)
{
	int i, j, x, y;
	unsigned char Buffer[IMAGESIZE];

	unsigned char buffer[ _X * _Y * 3/2];

	
	/* ̹    176 x 144 ,YUV */
	
	int offsetSrc = 0, offsetDes = 0;
	int xx, yy;
	
	for(y=0; y<_Y; y++) 
	{
		yy = (int)(y * sy / 144.0);
		
		int y320 = yy * sx;
		int yrev = (143-y)*176;

		for(x=0; x<_X; x++)
		{
			int R, G, B, Y, Cr, Cb;

			xx = (int)(x * sx / 176.0);
			offsetSrc = (xx + y320) * 3;

			/* YUV Format  ٲ۴. */
			
			B = *(pimage + offsetSrc++ );
			G = *(pimage + offsetSrc++ );
			R = *(pimage + offsetSrc );
			
			Y  = ( 299 * R + 587*G + 114*B)/1000;
			Cr = (-169 * R - 332*G + 500*B)/1000;
			Cb = ( 500 * R - 419*G -  81*B)/1000; 

			/*  ´.	*/
			offsetDes = ( x + yrev ) * 3;
			
			Buffer[offsetDes++ ] = Y;
			Buffer[offsetDes++ ] = Cr+128;
			Buffer[offsetDes   ] = Cb+128;
		}
	}

	int offset;

	/* 4:2:0 Format . */
	int temp1, temp2;
	for (y=0; y<144; y+=2) 
	{
		yy = (y*176)>>1;
		for(x=0; x<176; x+=2)
		{
			temp1 = temp2 = 0;
			for(i=0; i<2; i++) {
				yy = (y+i)*176;
				for(j=0; j<2; j++)
				{
					offset = ((x+j) + (y+i)*176) *3;
					/* Write Y */
					*( buffer + (x+j) + yy ) = Buffer[ offset++ ];
					temp1 += Buffer[ offset++ ];
					temp2 += Buffer[ offset ];
				}
			}
			offset = (( x + y*176/2 ) >>1 ) + (176*144);					

			/* Write U, V */
			*(buffer + offset )               = temp1 >>2 ;
			*(buffer + offset + (176*144/4) ) = temp2 >>2 ;
		}
	}
	return buffer;
}
//-----------------------------------------------------------------------------